/*
** ADM-XRC Series Driver
**
** (C) Copyright 2001-2004 Alpha Data
**
** cardinfo.c - Routines to read the EEPROM configuration information on a
**              hardware instance and store it in the device context.
**
** ----------------------------------------------------------------------------
**
** Exported functions:
**
**   ADMXRC_SetupCardInfo()  Reads configuration information from nonvolatile
**                           memory and saves it in the device context.
*/


#include "device.h"
#include "drvparam.h"
#include "flash.h"

static void
getSpaceInfoPci9080(
  DeviceContext* pDevCtx)
{
  os_uint32 data32;
  PLX9080REGS*  regs;

  regs = pDevCtx->hw.adapter.pci9080;

  data32 = os_read_register_le32(&regs->LAS0BA);
  if (!(data32 & 0x1U)) {
    /* Space not enabled - we don't expect this */
    pDevCtx->info.spaceInfo[0].LocalBase = 0x0U;
    pDevCtx->info.spaceInfo[0].LocalSize = 0x0U;
  } else {
    pDevCtx->info.spaceInfo[0].LocalBase = data32 & ~0xf;
    data32 = os_read_register_le32(&regs->LAS0RR);
    pDevCtx->info.spaceInfo[0].LocalSize = (~(data32 & ~0xf)) + 1;
  }
  pDevCtx->info.spaceInfo[0].PhysicalBase = (unsigned long) pDevCtx->resources.raw.pci.base[pDevCtx->privateInfo.numBar - 2];
  pDevCtx->info.spaceInfo[0].VirtualBase = NULL;
  pDevCtx->info.spaceInfo[0].VirtualSize = 0;
  
  data32 = os_read_register_le32(&regs->LAS1BA);
  if (!(data32 & 0x1)) {
    /* Space not enabled - we don't expect this */
    pDevCtx->info.spaceInfo[1].LocalBase = 0x0U;
    pDevCtx->info.spaceInfo[1].LocalSize = 0x0U;
  } else {
    pDevCtx->info.spaceInfo[1].LocalBase = data32 & ~0xf;
    data32 = os_read_register_le32(&regs->LAS1RR);
    pDevCtx->info.spaceInfo[1].LocalSize = (~(data32 & ~0xf)) + 1;
  }
  pDevCtx->info.spaceInfo[1].PhysicalBase = (unsigned long) pDevCtx->resources.raw.pci.base[pDevCtx->privateInfo.numBar - 1];
  pDevCtx->info.spaceInfo[1].VirtualBase = NULL;
  pDevCtx->info.spaceInfo[1].VirtualSize = 0;
  
  pDevCtx->info.cardInfo.NumSpace = 2;
}

static void
getSpaceInfoPci9656(
  DeviceContext* pDevCtx)
{
  os_uint32 data32;
  PLX9656REGS*  regs;

  regs = pDevCtx->hw.adapter.pci9656;

  data32 = os_read_register_le32(&regs->LAS0BA);
  if (!(data32 & 0x1U)) {
    /* Space not enabled - we don't expect this */
    pDevCtx->info.spaceInfo[0].LocalBase = 0x0U;
    pDevCtx->info.spaceInfo[0].LocalSize = 0x0U;
  } else {
    pDevCtx->info.spaceInfo[0].LocalBase = data32 & ~0xf;
    data32 = os_read_register_le32(&regs->LAS0RR);
    pDevCtx->info.spaceInfo[0].LocalSize = (~(data32 & ~0xf)) + 1;
  }
  pDevCtx->info.spaceInfo[0].PhysicalBase = (unsigned long) pDevCtx->resources.raw.pci.base[pDevCtx->privateInfo.numBar - 2];
  pDevCtx->info.spaceInfo[0].VirtualBase = NULL;
  pDevCtx->info.spaceInfo[0].VirtualSize = 0;
  
  data32 = os_read_register_le32(&regs->LAS1BA);
  if (!(data32 & 0x1)) {
    /* Space not enabled - we don't expect this */
    pDevCtx->info.spaceInfo[1].LocalBase = 0x0U;
    pDevCtx->info.spaceInfo[1].LocalSize = 0x0U;
  } else {
    pDevCtx->info.spaceInfo[1].LocalBase = data32 & ~0xf;
    data32 = os_read_register_le32(&regs->LAS1RR);
    pDevCtx->info.spaceInfo[1].LocalSize = (~(data32 & ~0xf)) + 1;
  }
  pDevCtx->info.spaceInfo[1].PhysicalBase = (unsigned long) pDevCtx->resources.raw.pci.base[pDevCtx->privateInfo.numBar - 1];
  pDevCtx->info.spaceInfo[1].VirtualBase = NULL;
  pDevCtx->info.spaceInfo[1].VirtualSize = 0;
  
  pDevCtx->info.cardInfo.NumSpace = 2;    
}

/*
** Read EEPROM info on an ADM-XRC / ADM-XRC-P / ADM-XRCII-Lite.
**
** There are some differences between the ADM-XRC, ADM-XRC-P
** and the ADM-XRCIILite in the encoding of the EEPROM information,
** but they have enough in common for this routine to handle all 3
** models.
*/
static os_boolean
getCardInfoPci9080(
  DeviceContext* pDevCtx)
{
  os_uint32     data16, data32;
  unsigned long i;
  PLX9080REGS*  regs;

  os_dbgprint(1, ("getCardInfoPci9080: entered\n"));

  regs = pDevCtx->hw.adapter.pci9080;

  pDevCtx->info.adapterLimit = 0x100;

  switch (pDevCtx->devDesc.id.pci.subDId) {
  case PCI_SUBDEVICE_ADATA_ADMXRC:
    {
      /*
      ** ADM-XRC or ADM-XRC-P
      **
      ** Location 0x36: board type
      */
      pDevCtx->methods.configMem(pDevCtx, FALSE, 0x36, &data16);
      if (data16 == ADMXRC2_BOARD_ADMXRC || data16 == ADMXRC2_BOARD_ADMXRC_P) {
	pDevCtx->info.cardInfo.BoardType = (ADMXRC2_BOARD_TYPE) data16;
	pDevCtx->info.cardInfo.NumClock = 2;
	pDevCtx->info.cardInfo.NumDMAChan = 2;
      } else {
	/*
	** Board type must be recorded wrongly in EEPROM,
	** assume ADM-XRC.
	*/
	pDevCtx->info.cardInfo.BoardType = ADMXRC2_BOARD_ADMXRC;
	pDevCtx->info.cardInfo.NumClock = 2;
	pDevCtx->info.cardInfo.NumDMAChan = 2;
      }
      break;
    }

  case PCI_SUBDEVICE_ADATA_ADMXRC2L:
    {
      /*
      ** ADM-XRC2-L
      **
      ** Don't use the board type from the EEPROM, as
      ** the PCI subsystem device ID for the ADM-XRC2-L
      ** is sufficient to correctly identify it.
      */
      pDevCtx->info.cardInfo.BoardType = ADMXRC2_BOARD_ADMXRC2_LITE;
      pDevCtx->info.cardInfo.NumClock = 2;
      pDevCtx->info.cardInfo.NumDMAChan = 2;
      break;
    }

  default:
    {
      os_assert(FALSE);
      return FALSE;
    }
  }

  /*
  ** Location 0x37: modifications word
  **
  ** The modifications word is interpreted in a board
  ** specific manner.
  */
  pDevCtx->info.modifications = 0x0;
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x37, &data16);
  switch (pDevCtx->info.cardInfo.BoardType) {
  case ADMXRC2_BOARD_ADMXRC:
  case ADMXRC2_BOARD_ADMXRC_P:
    {
      /*
      ** ADM-XRC/ADM-XRC-P:
      **
      ** Bit     Meaning
      **
      ** 0       1 => clock swap mod has been performed
      */
      if (data16 & 0x1) {
	pDevCtx->info.modifications |= ADMXRC_MOD_CLOCKSWAP;
      }
      break;
    }

  case ADMXRC2_BOARD_ADMXRC2_LITE:
    {
      /*
      ** ADM-XRC-II-L: no modifications currently defined
      */
      break;
    }

  default:
    os_assert(FALSE);
    return FALSE;
  }
  
  /* Location 0x38: clock oscillator type (ie. frequency) */
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x38, &data16);  
  pDevCtx->info.clockType = (ADMXRC_CLOCK_TYPE) data16;
  
  /* Location 0x39: flash type (ie. device) - IGNORED */
  
  /* Location 0x3B: card ID */
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x3B, &data16);
  pDevCtx->info.cardInfo.CardID = (ADMXRC2_CARDID) data16;
  
  /* Location 0x3C: fpga type (ie. device) */
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x3C, &data16);
  pDevCtx->info.cardInfo.FPGAType = (ADMXRC2_FPGA_TYPE) data16;
  
  /* Location 0x3D: serial number low 16 bits */
  /* Location 0x3E: serial number high 16 bits */
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x3D, &data16);
  data32 = data16;
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x3E, &data16);
  data32 |= (data16 << 16);
  pDevCtx->info.cardInfo.SerialNum = data32;
  
  /*
  ** Location 0x3F:
  **    low 8 bits = logic rev
  **    high 8 bits = board rev
  */
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x3F, &data16);
  pDevCtx->info.cardInfo.BoardRevision = (os_uint8) ((data16 & 0xff00U) >> 8);
  pDevCtx->info.cardInfo.LogicRevision = (os_uint8) ((data16 & 0x00ffU) >> 0);
  
  /* Location 0x3A: SRAM presence info */
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x3A, &data16);
  switch (pDevCtx->info.cardInfo.BoardType) {
  case ADMXRC2_BOARD_ADMXRC:
  case ADMXRC2_BOARD_ADMXRC_P:
    {
      /*
      ** ADM-XRC/ADM-XRC-P: get bank size/presence info, 4 banks
      **
      ** 4 banks of x36 SSRAM
      */
      os_boolean found = FALSE;
      
      pDevCtx->info.cardInfo.NumRAMBank = 4;
      if ((data16 & 0xf) == 0xfU) {
	pDevCtx->info.cardInfo.RAMBanksFitted = 0x0;
	for (i = 0; i < 4; i++) {
	  pDevCtx->info.bankInfo[i].Fitted = TRUE;
	  pDevCtx->info.bankInfo[i].Type = ADMXRC2_RAM_ZBTFT;
	  pDevCtx->info.bankInfo[i].Width = 36;
	  if (data16 & (1 << (i + 4))) {
	    /* 16Mbit */
	    pDevCtx->info.bankInfo[i].Size = 512*1024;
	  } else {
	    /* 8Mbit */
	    pDevCtx->info.bankInfo[i].Size = 256*1024;
	  }
	  pDevCtx->info.cardInfo.RAMBanksFitted |= (1 << i);
	}
	found = TRUE;
      }
      if (!found) {
	for (i = 0; i < 4; i++) {
	  pDevCtx->info.bankInfo[i].Fitted = FALSE;
	  pDevCtx->info.bankInfo[i].Type = 0;
	  pDevCtx->info.bankInfo[i].Width = 36;
	  pDevCtx->info.bankInfo[i].Size = 0;
	}
	pDevCtx->info.cardInfo.RAMBanksFitted = 0x0;
      }
      break;
    }

  case ADMXRC2_BOARD_ADMXRC2_LITE:
    {
      /*	
      ** ADM-XRC-IIL: get bank size/presence info
      **
      ** 4 banks of x18 SSRAM
      */
      os_uint16  bank;
      os_boolean fitted = FALSE; /* Shut the compiler up */
      
      pDevCtx->info.cardInfo.NumRAMBank = 4;
      for (i = 0; i < pDevCtx->info.cardInfo.NumRAMBank; i++) {
	bank = (data16 >> (i << 2)) & 0xf;
	switch (bank & 0x3) {
	case 0x0:
	  /* not fitted */
	  fitted = FALSE;
	  pDevCtx->info.bankInfo[i].Type = 0;
	  break;
	case 0x1:
	  /* ZBT flow-through */
	  fitted = TRUE;
	  pDevCtx->info.bankInfo[i].Type = ADMXRC2_RAM_ZBTFT;
	  break;
	case 0x2:
	  /* ZBT pipelined */
	  fitted = TRUE;
	  pDevCtx->info.bankInfo[i].Type = ADMXRC2_RAM_ZBTP;
	  break;
	case 0x3:
	  /* ZBT flow-through + pipelined (selectable) */
	  fitted = TRUE;
	  pDevCtx->info.bankInfo[i].Type = ADMXRC2_RAM_ZBTFT | ADMXRC2_RAM_ZBTP;
	  break;
	}
	pDevCtx->info.bankInfo[i].Width = 18;
	if (!fitted) {
	  pDevCtx->info.bankInfo[i].Size = 0;
	} else {
	  switch ((bank >> 2) & 0x3) {
	  case 0x0:
	    /* 8Mbit */
	    pDevCtx->info.bankInfo[i].Size = 512*1024;
	    break;
	  case 0x1:
	    /* 16Mbit */
	    pDevCtx->info.bankInfo[i].Size = 1024*1024;
	    break;
	  case 0x2:
	    /* 32Mbit */
	    pDevCtx->info.bankInfo[i].Size = 2048*1024;
	    break;
	  case 0x3:
	    /* reserved */
	    pDevCtx->info.bankInfo[i].Size = 0;
	    fitted = FALSE;
	    break;
	  }
	}
	pDevCtx->info.bankInfo[i].Fitted = fitted;
	if (fitted) {
	  pDevCtx->info.cardInfo.RAMBanksFitted |= (1 << i);
	}
      }
      break;
    }

  default:
    os_assert(FALSE);
    return FALSE;
  }

  getSpaceInfoPci9080(pDevCtx);

  return TRUE;
}

/*
** Read EEPROM info on an ADM-XRCII (PCI9656 bridge).
*/
static os_boolean
getCardInfoADMXRC2(
  DeviceContext* pDevCtx)
{
  os_uint32     data16, data32;
  unsigned long i;
  PLX9656REGS*  regs;

  os_dbgprint(1, ("getCardInfoADMXRC2: entered\n"));

  regs = pDevCtx->hw.adapter.pci9656;

  pDevCtx->info.adapterLimit = 0x100;

  pDevCtx->info.cardInfo.BoardType = ADMXRC2_BOARD_ADMXRC2;
  pDevCtx->info.cardInfo.NumClock = 2;
  pDevCtx->info.cardInfo.NumDMAChan = 2;

  /*
  ** Location 0x36: modifications word
  **
  ** The modifications word is interpreted in a board
  ** specific manner.
  */
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x36, &data16);
  pDevCtx->info.modifications = 0x0; /* No modifications defined yet */

  /* Location 0x37: clock oscillator type (ie. frequency) */
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x37, &data16);
  pDevCtx->info.clockType = (ADMXRC_CLOCK_TYPE) data16;
  
  /* Location 0x38: flash type (ie. device) - IGNORED */

  /* Location 0x3B: card ID */
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x3B, &data16);
  pDevCtx->info.cardInfo.CardID = (ADMXRC2_CARDID) data16;
  
  /* Location 0x3C: fpga type (ie. device) */
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x3C, &data16);
  pDevCtx->info.cardInfo.FPGAType = (ADMXRC2_FPGA_TYPE) data16;

  /* Location 0x3D: serial number low 16 bits */
  /* Location 0x3E: serial number high 16 bits */
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x3D, &data16);
  data32 = data16;
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x3E, &data16);
  data32 |= (data16 << 16);
  pDevCtx->info.cardInfo.SerialNum = data32;

  /*
  ** Location 0x3F:
  **    low 8 bits = logic rev
  **    high 8 bits = board rev
  */
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x3F, &data16);
  pDevCtx->info.cardInfo.BoardRevision = (os_uint8) ((data16 & 0xff00) >> 8);
  pDevCtx->info.cardInfo.LogicRevision = (os_uint8) ((data16 & 0x00ff) >> 0);
  
  /* Locations 0x39 & 0x3A: SRAM presence info */
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x39, &data16);
  data32 = data16;
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x3A, &data16);
  data32 |= data16 << 16U;
  {
    /*
    ** ADM-XRC2: get bank size/presence info, 6 banks
    */
    unsigned long bank;
    os_boolean    fitted = FALSE;
    
    pDevCtx->info.cardInfo.NumRAMBank = 6;
    pDevCtx->info.cardInfo.RAMBanksFitted = 0;
    for (i = 0; i < pDevCtx->info.cardInfo.NumRAMBank; i++) {
      bank = (data32 >> (i << 2)) & 0xf;
      switch (bank & 0x3) {
      case 0x0:
	/* not fitted */
	fitted = FALSE;
	pDevCtx->info.bankInfo[i].Type = 0;
	break;

      case 0x1:
	/* ZBT flow-through */
	fitted = TRUE;
	pDevCtx->info.bankInfo[i].Type = ADMXRC2_RAM_ZBTFT;
	break;

      case 0x2:
	/* ZBT pipelined */
	fitted = TRUE;
	pDevCtx->info.bankInfo[i].Type = ADMXRC2_RAM_ZBTP;
	break;

      case 0x3:
	/* ZBT flow-through + pipelined (selectable) */
	fitted = TRUE;
	pDevCtx->info.bankInfo[i].Type = ADMXRC2_RAM_ZBTFT | ADMXRC2_RAM_ZBTP;
	break;

      default:
	os_assert(FALSE);
	break;
      }
      pDevCtx->info.bankInfo[i].Width = (pDevCtx->info.cardInfo.FPGAType == ADMXRC2_FPGA_2V3000) ? 32 : 36;
      if (!fitted) {
	pDevCtx->info.bankInfo[i].Size = 0;
      } else {
	switch ((bank >> 2) & 0x3) {
	case 0x0:
	  /* 8Mbit */
	  pDevCtx->info.bankInfo[i].Size = 256*1024;
	  break;

	case 0x1:
	  /* 16Mbit */
	  pDevCtx->info.bankInfo[i].Size = 512*1024;
	  break;

	case 0x2:
	  /* 32Mbit */
	  pDevCtx->info.bankInfo[i].Size = 1024*1024;
	  break;

	case 0x3:
	  /* 64Mbit */
	  pDevCtx->info.bankInfo[i].Size = 2048*1024;
	  fitted = FALSE;
	  break;

	default:
	  os_assert(FALSE);
	  break;
	}
      }
      pDevCtx->info.bankInfo[i].Fitted = fitted;
      if (fitted) {
	pDevCtx->info.cardInfo.RAMBanksFitted |= (1 << i);
      }
    }
  }

  getSpaceInfoPci9656(pDevCtx);
  return TRUE;
}

/*
** Read configuration information on an ADM-XRCIIPro-Lite (actually read
** from a Flash memory attached to the PCI bridge).
*/
static os_boolean
getCardInfoXpl(
  DeviceContext* pDevCtx)
{
  XPLBRIDGEREGS* regs;
  XPLCONFIGREV0	 config;
  os_uint32      val32;
  unsigned int	 i;
  os_boolean     bConfigFound = FALSE;
  
  os_dbgprint(3, ("getCardInfoXpl: entered\n"));
  
  regs = pDevCtx->hw.adapter.xpl;
  
  pDevCtx->info.adapterLimit = 0x100;
  
  if (!driverParams.model.admxrc2pl.unlockFlash) {
    os_uint16 version;

    if (readFlash(pDevCtx, 0, sizeof(config), &config) == MethodResult_Success) {
      version = os_le16_to_cpu(config.version);
      switch (version) {
      case 0:
	bConfigFound = TRUE;
	break;
	
      default:
	/* Unrecognised data structure */
	break;
      }
    }
  }

  if (!bConfigFound) {
    os_dbgprint(0, ("*** getCardInfoXpl: no configuration information found\n"));
  }

  if (bConfigFound) {
    pDevCtx->info.cardInfo.BoardType = ADMXRC2_BOARD_ADMXRC2_PRO_LITE;
    pDevCtx->info.cardInfo.NumClock = 1;
    pDevCtx->info.cardInfo.CardID = os_le32_to_cpu(config.cardId);
    pDevCtx->info.cardInfo.FPGAType = (ADMXRC2_FPGA_TYPE) os_le16_to_cpu(config.fpgaType);
    pDevCtx->info.cardInfo.SerialNum = os_le32_to_cpu(config.serialNumber);
    pDevCtx->info.cardInfo.BoardRevision = os_le8_to_cpu(config.pcbRev);
    pDevCtx->info.cardInfo.LogicRevision = pDevCtx->devDesc.id.pci.revId + 0x10U;
    pDevCtx->info.cardInfo.NumDMAChan = (pDevCtx->devDesc.id.pci.revId >= 0x1U) ? 2 : 0;
    pDevCtx->info.cardInfo.NumRAMBank = 2;
    pDevCtx->info.cardInfo.RAMBanksFitted = 0x0;
    pDevCtx->info.cardInfo.NumSpace = 2;
    pDevCtx->info.modifications = 0x0;
    pDevCtx->info.clockType = 0;
    for (i = 0; i < 2; i++) {
      os_uint16 typeMask;
      os_uint8 width, order;

      typeMask = os_le16_to_cpu(config.bank[i].typeMask);
      width = os_le8_to_cpu(config.bank[i].width);
      order = os_le8_to_cpu(config.bank[i].order);
      pDevCtx->info.bankInfo[i].Type = typeMask;
      pDevCtx->info.bankInfo[i].Width = width;
      if (order && typeMask && width) {
	pDevCtx->info.cardInfo.RAMBanksFitted |= (0x1U << i);
	pDevCtx->info.bankInfo[i].Fitted = TRUE;
	pDevCtx->info.bankInfo[i].Size = 1 << order;
      } else {
	pDevCtx->info.bankInfo[i].Fitted = FALSE;
	pDevCtx->info.bankInfo[i].Size = 0;
      }
    }
  } else {
    pDevCtx->info.cardInfo.BoardType = ADMXRC2_BOARD_ADMXRC2_PRO_LITE;
    pDevCtx->info.cardInfo.NumClock = 1;
    pDevCtx->info.cardInfo.NumDMAChan = 0;
    pDevCtx->info.cardInfo.CardID = 0xdeadU;
    pDevCtx->info.cardInfo.FPGAType = (ADMXRC2_FPGA_TYPE) ADMXRC2_FPGA_2VP7;
    pDevCtx->info.cardInfo.SerialNum = 0xcafefaceU;
    pDevCtx->info.cardInfo.BoardRevision = 0x10;
    pDevCtx->info.cardInfo.LogicRevision = 0x10;
    pDevCtx->info.cardInfo.NumRAMBank = 2;
    pDevCtx->info.cardInfo.RAMBanksFitted = 0x0;
    pDevCtx->info.cardInfo.NumSpace = 2;
    pDevCtx->info.modifications = 0x0;
    pDevCtx->info.clockType = 0;
    pDevCtx->info.bankInfo[0].Fitted = FALSE;
    pDevCtx->info.bankInfo[0].Type = ADMXRC2_RAM_ZBTP;
    pDevCtx->info.bankInfo[0].Size = 0;
    pDevCtx->info.bankInfo[0].Width = 64;
    pDevCtx->info.bankInfo[1].Fitted = FALSE;
    pDevCtx->info.bankInfo[1].Type = ADMXRC2_RAM_SDRAM_DDR;
    pDevCtx->info.bankInfo[1].Size = 0;
    pDevCtx->info.bankInfo[1].Width = 64;
  }
  
  pDevCtx->info.spaceInfo[0].PhysicalBase = (unsigned long) pDevCtx->resources.raw.pci.base[pDevCtx->privateInfo.numBar - 2];
  val32 = os_read_register_le32(&regs->LS0B) & 0xffff0000;
  pDevCtx->info.spaceInfo[0].LocalBase = val32;
  val32 = os_read_register_le32(&regs->LS0R) & 0xffff0000;
  pDevCtx->info.spaceInfo[0].LocalSize = ~val32 + 1;
  pDevCtx->info.spaceInfo[0].VirtualBase = NULL;
  pDevCtx->info.spaceInfo[0].VirtualSize = 0;

  pDevCtx->info.spaceInfo[1].PhysicalBase = (unsigned long) pDevCtx->resources.raw.pci.base[pDevCtx->privateInfo.numBar - 1];
  pDevCtx->info.spaceInfo[1].LocalBase = 0x800000U;
  pDevCtx->info.spaceInfo[1].LocalSize = 0x400000U;
  pDevCtx->info.spaceInfo[1].VirtualBase = NULL;
  pDevCtx->info.spaceInfo[1].VirtualSize = 0;

  return TRUE;
}

/*
** Read EEPROM info on an ADM-XRCII-N (PCI9656 bridge).
*/
static os_boolean
getCardInfoADPWRC2(
  DeviceContext* pDevCtx)
{
  os_uint32     data16, data32;
  unsigned long i;
  PLX9656REGS*  regs;

  os_dbgprint(1, ("getCardInfoADPWRC2: entered\n"));

  regs = pDevCtx->hw.adapter.pci9656;

  pDevCtx->info.adapterLimit = 0x100;

  pDevCtx->info.cardInfo.BoardType = ADMXRC2_BOARD_ADPWRC2;
  pDevCtx->info.cardInfo.NumClock = 2;
  pDevCtx->info.clockType = 0x0001U; /* 14.318MHz */
  pDevCtx->info.cardInfo.NumDMAChan = 2;
  pDevCtx->info.cardInfo.NumRAMBank = 2;
  pDevCtx->info.cardInfo.RAMBanksFitted = 0;

  /*
  ** Location 0x36: modifications word
  **
  ** The modifications word is interpreted in a board
  ** specific manner.
  */
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x36, &data16);
  pDevCtx->info.modifications = 0x0; /* No modifications defined yet */

  /* Location 0x37: clock oscillator type (ie. frequency) */
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x37, &data16);
  pDevCtx->info.clockType = (ADMXRC_CLOCK_TYPE) data16;
  
  /* Location 0x38: flash type (ie. device) - IGNORED */

  /* Location 0x3B: card ID */
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x3B, &data16);
  pDevCtx->info.cardInfo.CardID = (ADMXRC2_CARDID) data16;
  
  /* Location 0x3C: fpga type (ie. device) */
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x3C, &data16);
  pDevCtx->info.cardInfo.FPGAType = (ADMXRC2_FPGA_TYPE) data16;

  /* Location 0x3D: serial number low 16 bits */
  /* Location 0x3E: serial number high 16 bits */
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x3D, &data16);
  data32 = data16;
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x3E, &data16);
  data32 |= (data16 << 16);
  pDevCtx->info.cardInfo.SerialNum = data32;

  /*
  ** Location 0x3F:
  **    low 8 bits = logic rev
  **    high 8 bits = board rev
  */
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x3F, &data16);
  pDevCtx->info.cardInfo.BoardRevision = (os_uint8) ((data16 & 0xff00) >> 8);
  pDevCtx->info.cardInfo.LogicRevision = (os_uint8) ((data16 & 0x00ff) >> 0);
  
  /* Locations 0x39 & 0x3A: SRAM presence info */
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x39, &data16);
  data32 = data16;
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x3A, &data16);
  data32 |= data16 << 16U;

  /* Locations 0x39-0x3A: memory bank presence info */
  for (i = 0; i < 2; i++) {
    pDevCtx->methods.configMem(pDevCtx, FALSE, 0x39U + i, &data32);
    if (!(data32 & 0xffU)) {
      /* Bank not fitted */
      pDevCtx->info.bankInfo[i].Fitted = FALSE;
      pDevCtx->info.bankInfo[i].Size = 0;
    } else {
      pDevCtx->info.cardInfo.RAMBanksFitted |= (0x1U << i);
      pDevCtx->info.bankInfo[i].Fitted = TRUE;
      pDevCtx->info.bankInfo[i].Size = 0x1U << (data32 & 0xffU);
    }
    
    pDevCtx->info.bankInfo[i].Type = ADMXRC2_RAM_SDRAM_DDR;
    pDevCtx->info.bankInfo[i].Width = 512;
  }

  getSpaceInfoPci9656(pDevCtx);

  return TRUE;
}

/*
** Read EEPROM info on an ADM-XRCII-EMP (PCI9656 bridge).
*/
static os_boolean
getCardInfoADPDRC2(
  DeviceContext* pDevCtx)
{
  os_uint32     data16, data32;
  unsigned long i;
  PLX9656REGS*  regs;

  os_dbgprint(1, ("getCardInfoADPDRC2: entered\n"));

  regs = pDevCtx->hw.adapter.pci9656;

  pDevCtx->info.adapterLimit = 0x100;

  pDevCtx->info.cardInfo.BoardType = ADMXRC2_BOARD_ADPDRC2;
  pDevCtx->info.cardInfo.NumClock = 2;
  pDevCtx->info.cardInfo.NumDMAChan = 2;
  pDevCtx->info.clockType = 0x0001U; /* 14.318MHz */
  pDevCtx->info.cardInfo.NumRAMBank = 5;
  pDevCtx->info.cardInfo.RAMBanksFitted = 0;

  /*
  ** Location 0x35: modifications word
  **
  ** The modifications word is interpreted in a board
  ** specific manner.
  */
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x35, &data16);
  pDevCtx->info.modifications = 0x0; /* No modifications defined yet */

  /* Location 0x3B: card ID */
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x3B, &data16);
  pDevCtx->info.cardInfo.CardID = (ADMXRC2_CARDID) data16;
  
  /* Location 0x3C: fpga type (ie. device) */
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x3C, &data16);
  pDevCtx->info.cardInfo.FPGAType = (ADMXRC2_FPGA_TYPE) data16;

  /* Location 0x3D: serial number low 16 bits */
  /* Location 0x3E: serial number high 16 bits */
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x3D, &data16);
  data32 = data16;
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x3E, &data16);
  data32 |= (data16 << 16);
  pDevCtx->info.cardInfo.SerialNum = data32;

  /*
  ** Location 0x3F:
  **    low 8 bits = logic rev
  **    high 8 bits = board rev
  */
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x3F, &data16);
  pDevCtx->info.cardInfo.BoardRevision = (os_uint8) ((data16 & 0xff00) >> 8);
  pDevCtx->info.cardInfo.LogicRevision = (os_uint8) ((data16 & 0x00ff) >> 0);
  
  /* Locations 0x39 & 0x3A: SRAM presence info */
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x39, &data16);
  data32 = data16;
  pDevCtx->methods.configMem(pDevCtx, FALSE, 0x3A, &data16);
  data32 |= data16 << 16U;

  /* Locations 0x36-0x3A: memory bank presence info */
  for (i = 0; i < 4; i++) {
    os_uint32 type;
    
    pDevCtx->methods.configMem(pDevCtx, FALSE, 0x36U + i, &data32);
    if (!(data32 & 0xffU)) {
      /* Bank not fitted */
      pDevCtx->info.bankInfo[i].Fitted = FALSE;
      pDevCtx->info.bankInfo[i].Size = 0;
    } else {
      pDevCtx->info.bankInfo[i].Fitted = TRUE;
      pDevCtx->info.cardInfo.RAMBanksFitted |= (0x1U << i);
      pDevCtx->info.bankInfo[i].Size = 0x1U << (data32 & 0xffU);
    }
    type = (data32 & 0x7f00U) >> 8;
    pDevCtx->info.bankInfo[i].Type = ADMXRC2_RAM_SDRAM_DDR;
    pDevCtx->info.bankInfo[i].Width = 128;
  }
  
  for (i = 4; i < 5; i++) {
    pDevCtx->methods.configMem(pDevCtx, FALSE, 0x36U + i, &data32);
    if (!(data32 & 0xffU)) {
      /* Bank not fitted */
      pDevCtx->info.bankInfo[i].Fitted = FALSE;
      pDevCtx->info.bankInfo[i].Size = 0;
    } else {
      pDevCtx->info.bankInfo[i].Fitted = TRUE;
      pDevCtx->info.cardInfo.RAMBanksFitted |= (0x1U << i);
      pDevCtx->info.bankInfo[i].Size = 0x1U << (data32 & 0xffU);
    }
    if (data32 & 0x8000U) {
      pDevCtx->info.bankInfo[i].Width = 72;
    } else {
      pDevCtx->info.bankInfo[i].Width = 64;
    }
    pDevCtx->info.bankInfo[i].Type = ADMXRC2_RAM_SRAM_DDR2;
  }
  
  getSpaceInfoPci9656(pDevCtx);

  return TRUE;
}

/*
** Read configuration information on an ADM-XRCIIPro-Lite (actually read
** from a Flash memory attached to the PCI bridge).
*/
static os_boolean
getCardInfoXp(
  DeviceContext* pDevCtx)
{
  XPLBRIDGEREGS* regs;
  XPCONFIGREV0	 config;
  os_uint32      val32;
  unsigned int	 i;
  os_boolean     bConfigFound = FALSE;
  
  os_dbgprint(3, ("getCardInfoXp: entered\n"));
  
  regs = pDevCtx->hw.adapter.xpl;
  
  pDevCtx->info.adapterLimit = 0x100;
  
  if (!driverParams.model.admxrc2pl.unlockFlash) {
    os_uint16 version;

    if (readFlash(pDevCtx, 0, sizeof(config), &config) == MethodResult_Success) {
      version = os_le16_to_cpu(config.version);
      switch (version) {
      case 0:
	bConfigFound = TRUE;
	break;
	
      default:
	/* Unrecognised data structure */
	break;
      }
    }
  }

  if (!bConfigFound) {
    os_dbgprint(0, ("*** getCardInfoXpl: no configuration information found\n"));
  }

  if (bConfigFound) {
    pDevCtx->info.cardInfo.BoardType = ADMXRC2_BOARD_ADMXRC2_PRO;
    pDevCtx->info.cardInfo.NumClock = 1;
    pDevCtx->info.cardInfo.CardID = os_le32_to_cpu(config.cardId);
    pDevCtx->info.cardInfo.FPGAType = (ADMXRC2_FPGA_TYPE) os_le16_to_cpu(config.fpgaType);
    pDevCtx->info.cardInfo.SerialNum = os_le32_to_cpu(config.serialNumber);
    pDevCtx->info.cardInfo.BoardRevision = os_le8_to_cpu(config.pcbRev);
    pDevCtx->info.cardInfo.LogicRevision = pDevCtx->devDesc.id.pci.revId + 0x10U;
    pDevCtx->info.cardInfo.NumDMAChan = 2;
    pDevCtx->info.cardInfo.NumRAMBank = 6;
    pDevCtx->info.cardInfo.RAMBanksFitted = 0x0;
    pDevCtx->info.cardInfo.NumSpace = 2;
    pDevCtx->info.modifications = 0x0;
    pDevCtx->info.clockType = 0;
    for (i = 0; i < 4; i++) {
      pDevCtx->info.bankInfo[i].Type = ADMXRC2_RAM_SRAM_DDR2;
      pDevCtx->info.bankInfo[i].Width = 64;
      if (config.zbt[i].order) {
	pDevCtx->info.cardInfo.RAMBanksFitted |= (0x1U << i);
	pDevCtx->info.bankInfo[i].Fitted = TRUE;
	pDevCtx->info.bankInfo[i].Size = 1 << config.zbt[i].order;
      } else {
	pDevCtx->info.bankInfo[i].Fitted = FALSE;
	pDevCtx->info.bankInfo[i].Size = 0;
      }
    }
    for (i = 0; i < 2; i++) {
      pDevCtx->info.bankInfo[i + 4].Type = ADMXRC2_RAM_SDRAM_DDR;
      pDevCtx->info.bankInfo[i + 4].Width = 72;
      if (config.sdram[i].order) {
	pDevCtx->info.cardInfo.RAMBanksFitted |= (0x1U << (i + 4));
	pDevCtx->info.bankInfo[i + 4].Fitted = TRUE;
	pDevCtx->info.bankInfo[i + 4].Size = 1 << config.sdram[i].order;
      } else {
	pDevCtx->info.bankInfo[i + 4].Fitted = FALSE;
	pDevCtx->info.bankInfo[i + 4].Size = 0;
      }
    }
  } else {
    pDevCtx->info.cardInfo.BoardType = ADMXRC2_BOARD_ADMXRC2_PRO;
    pDevCtx->info.cardInfo.NumClock = 1;
    pDevCtx->info.cardInfo.NumDMAChan = 2;
    pDevCtx->info.cardInfo.CardID = 0xdeadU;
    pDevCtx->info.cardInfo.FPGAType = (ADMXRC2_FPGA_TYPE) ADMXRC2_FPGA_2VP70;
    pDevCtx->info.cardInfo.SerialNum = 0xcafefaceU;
    pDevCtx->info.cardInfo.BoardRevision = 0x10;
    pDevCtx->info.cardInfo.LogicRevision = 0x10;
    pDevCtx->info.cardInfo.NumRAMBank = 6;
    pDevCtx->info.cardInfo.RAMBanksFitted = 0x0;
    pDevCtx->info.cardInfo.NumSpace = 2;
    pDevCtx->info.modifications = 0x0;
    pDevCtx->info.clockType = 0;
    for (i = 0; i < 4; i++) {
      pDevCtx->info.bankInfo[i].Type = ADMXRC2_RAM_SRAM_DDR2;
      pDevCtx->info.bankInfo[i].Width = 64;
      pDevCtx->info.bankInfo[i].Fitted = FALSE;
      pDevCtx->info.bankInfo[i].Size = 0;
    }
    for (i = 0; i < 2; i++) {
      pDevCtx->info.bankInfo[i + 4].Type = ADMXRC2_RAM_SDRAM_DDR;
      pDevCtx->info.bankInfo[i + 4].Width = 72;
      pDevCtx->info.bankInfo[i + 4].Fitted = FALSE;
      pDevCtx->info.bankInfo[i + 4].Size = 0;
    }
  }
  
  pDevCtx->info.spaceInfo[0].PhysicalBase = (unsigned long) pDevCtx->resources.raw.pci.base[pDevCtx->privateInfo.numBar - 2];
  val32 = os_read_register_le32(&regs->LS0B) & 0xffff0000;
  pDevCtx->info.spaceInfo[0].LocalBase = val32;
  val32 = os_read_register_le32(&regs->LS0R) & 0xffff0000;
  pDevCtx->info.spaceInfo[0].LocalSize = ~val32 + 1;
  pDevCtx->info.spaceInfo[0].VirtualBase = NULL;
  pDevCtx->info.spaceInfo[0].VirtualSize = 0;

  pDevCtx->info.spaceInfo[1].PhysicalBase = (unsigned long) pDevCtx->resources.raw.pci.base[pDevCtx->privateInfo.numBar - 1];
  pDevCtx->info.spaceInfo[1].LocalBase = 0x800000U;
  pDevCtx->info.spaceInfo[1].LocalSize = 0x400000U;
  pDevCtx->info.spaceInfo[1].VirtualBase = NULL;
  pDevCtx->info.spaceInfo[1].VirtualSize = 0;

  return TRUE;
}

/*
** -------------------------------------------------------------------------
** Exported routines
** -------------------------------------------------------------------------
*/

os_boolean
ADMXRC_SetupCardInfo(
  DeviceContext* pDevCtx)
{
  os_boolean bFound = FALSE;

  os_dbgprint(1, ("ADMXRC_SetupCardInfo: entered\n"));
  
  switch (pDevCtx->devDesc.id.pci.subDId) {
  case PCI_SUBDEVICE_ADATA_ADMXRC:
  case PCI_SUBDEVICE_ADATA_ADMXRC2L:
    bFound = getCardInfoPci9080(pDevCtx);
    break;

  case PCI_SUBDEVICE_ADATA_ADMXRC2:
    bFound = getCardInfoADMXRC2(pDevCtx);
    break;

  case PCI_SUBDEVICE_ADATA_ADMXRC2PL_PCIX:
  case PCI_SUBDEVICE_ADATA_ADMXRC2PL_PCI:
    bFound = getCardInfoXpl(pDevCtx);
    break;

  case PCI_SUBDEVICE_ADATA_ADPWRC2:
    bFound = getCardInfoADPWRC2(pDevCtx);
    break;

  case PCI_SUBDEVICE_ADATA_ADPDRC2:
    bFound = getCardInfoADPDRC2(pDevCtx);
    break;

  case PCI_SUBDEVICE_ADATA_ADMXRC2P_PCIX:
  case PCI_SUBDEVICE_ADATA_ADMXRC2P_PCI:
    bFound = getCardInfoXp(pDevCtx);
    break;

  default:
    os_assert(FALSE);
    break;
  }
  
  if (bFound) {
    os_dbgprint(1, ("ADMXRC_SetupCardInfo: successful\n"));
  } else {
    os_dbgprint(0, ("*** ADMXRC_SetupCardInfo: ID not recognised\n"));
  }

  return bFound;
}
