#ifndef _ADATA_DEVICE_H
#define _ADATA_DEVICE_H

/*
** Device structure definition for the ADM-XRC series driver
*/

#include "platform.h"
#include "device_defs.h"
#include "plx.h"
#include "xplbridge.h"

#include "admxrc/struct.h"
#include "admxrc2/status.h"
#include "admxrc2/struct.h"

/*
** Declare some structures now so we can forward-reference them
*/
struct _ADMXRC;
typedef struct _ADMXRC ADMXRC, DeviceContext;
struct _ADMXRC_DMAQ;
typedef struct _ADMXRC_DMAQ ADMXRC_DMAQ;
struct _ADMXRC_DMANODE;
typedef struct _ADMXRC_DMANODE ADMXRC_DMANODE;
struct _ADMXRC_DMATGT;
typedef struct _ADMXRC_DMATGT ADMXRC_DMATGT;
struct _ADMXRC_INTEVENT;
typedef struct _ADMXRC_INTEVENT ADMXRC_INTEVENT;
struct _DmaAbortContext;
typedef struct _DmaAbortContext DmaAbortContext;
struct _DmaScatterNode;
typedef struct _DmaScatterNode DmaScatterNode;
struct _DmaBounceBuffer;
typedef struct _DmaBounceBuffer DmaBounceBuffer;

typedef enum _ADMXRC_FLASH_TYPE {
  ADMXRC_FLASH_SST    = 0,
  ADMXRC_FLASH_MICRON = 1,
  ADMXRC_FLASH_CFI    = 2
} ADMXRC_FLASH_TYPE;

/*
** A structure representing a region of memory that has been
** locked down as a target for DMA transfers.
*/
struct _ADMXRC_DMATGT {
  OS_USBD*      usbd;   /* description of mapped & locked region        */
  void*         buffer; /* user space address of mapped & locked region */
  unsigned long length;	/* size of mapped & locked region               */
  unsigned long	in_use; /* (in_use > 0) => cannot be freed, is in use   */
};

/*
** A DMA completion callback routine - called when a DMA transfer
** is completed.
*/
#define ADMXRC_DMACALLBACK_REASON_SUBMIT    (0)
#define ADMXRC_DMACALLBACK_REASON_COLLECT   (1)
#define ADMXRC_DMACALLBACK_REASON_COMPLETED (2)
typedef void (ADMXRC_DMACALLBACK) (DeviceContext* pDevCtx, ADMXRC_DMANODE* node);

/*
** A DMA queue, for queuing up DMA transfers
*/
struct _ADMXRC_DMAQ {
  int             length;
  ADMXRC_DMANODE* head;
  ADMXRC_DMANODE* tail;
};

/*
** A context structure used when aborting DMA transfers
*/
struct _DmaAbortContext {
  /* initialised before call to atomic function */
  DeviceContext* devCtx;
  unsigned char  channel;

  /* filled in by call to atomic function */
  os_boolean     success;
};

/*
** A node in a scatter-gather list
*/
struct _DmaScatterNode {
  os_uint32  busAddr;
  os_uint32  localAddr;
  os_uint32  length;
  os_boolean bWrite;
};

/*
** A node in an Event list
*/
struct _ADMXRC_INTEVENT {
  os_event event;
  ADMXRC_INTEVENT* next;
};

/*
** A structure for controlling a DMA transfer; usually part of
** a queue of DMA transfers.
*/
struct _ADMXRC_DMANODE {
  /* Client code must not touch these members */
  struct {
    ADMXRC_DMANODE*     prev;
    ADMXRC_DMANODE*     next;
    unsigned char       state;
    unsigned char       channel;
  } system;

  /* Client code must initialize these members */
  struct {
    void*               context;  /* Client code can store some context information here */
    ADMXRC_DMACALLBACK* callback; /* Called when client code must give attention */
    os_uint32           mode;
    unsigned long       length;   /* WARNING - decremented during DMA */
    unsigned long       pos;      /* WARNING - incremented during DMA */
    unsigned char       direction;
    unsigned char       queue;
  } client;

  /* Client code may inspect these members on completion */
  struct {
    int                 reason;   /* Reason for calling callback */
    OS_STATUS           status;
  } status;
};

/*
** A structure for controlling an FPGA configuration operation that
** uses DMA.
*/
typedef struct _ADMXRC_LOADFPGADMA {
  ADMXRC_DMANODE      dmanode;
  OS_EVENT            event;
} ADMXRC_LOADFPGADMA;

/*
** The driver parameters structure
*/
typedef struct _DriverParams {
  os_boolean        pciReadMultiple;
  os_boolean        pciWriteAndInvalidate;
  os_uint32         maxDmaDesc;
  os_uint32         maxPlxDesc;
  union {
    struct {
      os_boolean    pci9656ErrataFix;
    } admxrc2;
    struct {
      os_boolean    unlockFlash;
    } admxrc2pl;
  } model;
} DriverParams;

/*
** Definitions to do with hardware methods.
*/

typedef enum _MethodResult {
  MethodResult_Success,            /* Successful                               */
  MethodResult_Timeout,            /* Timeout occurred during operation        */
  MethodResult_VerifyFailed,       /* Error verifying operation                */
  MethodResult_InvalidLocation,    /* Invalid location to access               */
  MethodResult_NotPresent,         /* Hardware is not present                  */
  MethodResult_NotSupported,       /* Method isn't supported for this hardware */
  MethodResult_InvalidParameter,   /* An invalid parameter was passed          */
  MethodResult_GeneralFailure      /* All other failures                       */
} MethodResult;

typedef MethodResult (MethodFlash)(
  DeviceContext* pDevCtx,
  os_boolean     write,
  os_uint32      location,
  os_uint16*     value);

typedef MethodResult (MethodFlashReset)(
  DeviceContext* pDevCtx);

typedef MethodResult (MethodIntEnab)(
  DeviceContext* pDevCtx,
  os_boolean     enable);

typedef os_boolean (MethodIsr)(
  DeviceContext* pDevCtx);

typedef MethodResult (MethodConfigMem)(
  DeviceContext* pDevCtx,
  os_boolean     write,
  unsigned long  location,
  os_uint32*     data);

typedef enum _MethodFpgaOp {
  MethodFpgaOp_SetProg,
  MethodFpgaOp_GetInit,
  MethodFpgaOp_GetDone
} MethodFpgaOp;

typedef MethodResult (MethodFpga)(
  DeviceContext* pDevCtx,
  MethodFpgaOp   op,
  void*          operand);

typedef MethodResult (MethodSelectMap)(
  DeviceContext* pDevCtx,
  os_boolean     write,
  os_uint8*      data);

typedef MethodResult (MethodConfigSpace)(
  DeviceContext* pDevCtx,
  os_boolean     write,
  unsigned int   space,
  os_uint32*     value);

typedef MethodResult (MethodAddDmaBuffer)(
  DeviceContext*  pDevCtx,
  unsigned int    chan,
  unsigned int    index,
  DmaScatterNode* snode);

typedef MethodResult (MethodDoDma)(
  DeviceContext*  pDevCtx,
  unsigned int    chan,
  os_boolean      start,
  os_uint32       mode,
  unsigned int    numBuffer);

typedef MethodResult (MethodClockWord)(
  DeviceContext*  pDevCtx,
  unsigned int    clockIndex,
  os_uint32       refFreq,
  os_uint32       reqFreq,
  os_uint32*      actualFreq,
  os_uint32*      clockWord);

typedef MethodResult (MethodClockProg)(
  DeviceContext*  pDevCtx,
  os_uint32       actualFreq,
  os_uint32       clockWord,
  os_boolean      bBefore);

typedef enum _DmaCompletionCode {
  DmaCompletionCode_Normal = 0,
  DmaCompletionCode_Failed = 1,
  DmaCompletionCode_EotRequested = 2
} DmaCompletionCode;

/*
** The device context structure
*/

struct _ADMXRC {
  /* Saved from onCreateDevice */
  OS_DRIVER_OBJECT*     pDrvObj;
  OS_DEVICE_OBJECT*     pDevObj;
  OS_DEVICE_DESC        devDesc;

  /* Resources given in onStartDevice */
  OS_DEVICE_RESOURCES   resources;

  /* Mapped/attached resources */
  OS_ATTACHED_IRQ*      irqObj;
  void*                 bar[4];

  /* Hardware mapped into kernel space */
  struct {
    union {
      void*             untyped;
      PLX9080REGS*      pci9080;
      PLX9656REGS*      pci9656;
      XPLBRIDGEREGS*    xpl;
    } adapter;
    union {
      void*             untyped;
      CONTROLXRC*       plxbased;
      CONTROLWRC2*      wrc2;
      XPLSPACE1REGS*    xpl;
      XPSPACE1REGS*     xp;
    } space1;
    void*               fpga;
  } hw;

  /* Shadows of important registers */
  union {
    struct {
      os_uint32         intcsr;
      os_uint32         intcsr_mask;
      os_uint16         dmacsr;
      os_uint8          imask;
    } plxbased;
    struct {
      os_uint32         pictl;
      os_uint32         fictl;
      os_uint32         dxctl[NUM_XPLBRIDGE_DMA_CHANNEL];
    } xpl;
    struct {
      os_uint32         pictl;
      os_uint32         fictl;
      os_uint32         dxctl[NUM_XPLBRIDGE_DMA_CHANNEL];
      os_uint32         fldata;
    } xp;
  } shadow;

  struct {
    /* Are interrupts enabled or disabled? */
    os_boolean            enabled;
    OS_DPC*               fpgaintDpc;
    OS_DPC*               dmaintDpc[ADMXRC_NUM_DMA_CHANNEL_MAX];
  } interrupts;

  /* Tracks the open handle to this device that has 'modify access' */
  struct {
    OS_USER_CLIENT      handle;
    OS_SPINLOCK         lock;
  } owner;

  /* DMA resources and data structures */
  struct {
    OS_SPINLOCK         lock;
    OS_LCACHE*          ctlPool;
    struct {
      OS_SG_MAP*        sgMap;
      struct {
	void*           real;                                    /* actual kernel address of descriptors */
	os_uint64       realPci;
	unsigned long   realSize;
	union {
	  PLXDESC*      plxbased;
	  XPLDMADESC*   xpl;
	  void*         untyped;
	} aligned;                                               /* aligned kernel address of descriptors */
	os_uint32       pci[ADMXRC_MAX_PLXDESC_PER_CHANNEL];     /* aligned PCI addresses of descriptors */
      } descriptors;
      unsigned long     count;                                   /* Count of jobs on the DMA channel, should never be > 1 */ 
      ADMXRC_DMANODE*   waitsubmit;                              /* The job (if any) that is currently in the WAITBUFFERS state */
      ADMXRC_DMANODE*   active;                                  /* The job (if any) that is currently in the ACTIVE state */
      ADMXRC_DMANODE*   waitcollect;                             /* The job (if any) that is currently in the WAITCOLLECT state */
      unsigned long     chunk;
      os_boolean        cleanupPending;
      os_uint8          completionCode;                          /* Whether or not the DMA transfer was successful on the hardware */
    } channel[ADMXRC_NUM_DMA_CHANNEL_MAX];
    ADMXRC_DMAQ         queue[ADMXRC_NUM_DMA_CHANNEL_MAX + 1];
  } dma;

  /* Memory regions mapped into user space */
  struct {
    OS_MAPPABLE_REGION* regions[6];
    int                 count;
  } userMapped;

  /* Card information, which user-mode threads can read */
  struct {
    ADMXRC2_CARD_INFO   cardInfo;
    ADMXRC2_SPACE_INFO  spaceInfo[ADMXRC_NUM_SPACE_MAX];
    ADMXRC2_BANK_INFO   bankInfo[ADMXRC_NUM_BANK_MAX];
    os_uint32           adapterLimit;
    ADMXRC_CLOCK_TYPE   clockType;
    unsigned long       modifications; /* format is board-specific */
    struct {
      os_boolean        present;
      os_uint8          width;
      
      struct {
	char*           name;
	os_uint16       vendor, device;
      } identity;
      
      os_uint32         size;

      struct {
	unsigned int	numBootBlock;
	os_uint32       bootBlock[4];
	os_uint32       eraseBlock;
      } layout;
      
      os_uint8*         buffer;
      
      struct {
	os_uint32	start;
	os_uint32       length;
      } targetArea;
    } flash;

    os_uint32           spaceConfig[ADMXRC_NUM_SPACE_MAX];
  } info;

  struct {
    unsigned int        dmaDescSize;
    int                 numBar;
  } privateInfo;

  /* FPGA information and control */
  struct {
    OS_LCACHE*          ctlPool;
    OS_SPINLOCK         lock;
    os_boolean          busy;
  } fpga;

  /* Interrupt waiting */
  struct {
    os_boolean          signalled;
    struct {
      OS_IO_REQUEST*    req;
      os_boolean        completing;
    } waiting;
    struct {
      unsigned long     ticks;
      os_boolean        never;
    } expire;
    ADMXRC_INTEVENT*    eventList;
    OS_SPINLOCK         lock;
  } intwait;

  /* Clock generator programming */
  struct {
    OS_SPINLOCK         lock;
    os_boolean          busy;
  } setclock;

  /* DMA descriptors */
  struct {
    ADMXRC_DMATGT**     table;     /* Maps DMADESCs to ADMXRC_DMATGTs */
    int*                freestack; /* Stack of free table indices     */
    int                 top;       /* Stack pointer for 'freestack'   */
    OS_SPINLOCK         lock;
  } dmadesc;

  struct {
    MethodConfigMem*    configMem;
    MethodFlash*        flash;
    MethodFlashReset*   flashReset;
    MethodIntEnab*      intEnab;
    MethodIsr*          isr;
    MethodFpga*         fpga;
    MethodSelectMap*    selectMap;
    MethodConfigSpace*  configSpace;
    MethodAddDmaBuffer* addDmaBuffer;
    MethodDoDma*        doDma;
    MethodClockWord*    clockWord;
    MethodClockProg*    clockProg;
  } methods;
};

#endif
