/*
** ADM-XRC Series Driver
**
** (C) Copyright 2001-2004 Alpha Data
**
** dma.c - IOCTL handlers for the DMA API functions
**
** ----------------------------------------------------------------------------
*/

#include "platform.h"
#include "device.h"
#include "drvparam.h"
#include "ioctl.h"
#include "dma.h"
#include "dmadesc.h"
#include "dmacore.h"

/*
** This callback function is attached to the DMANODEs created by
** the DoDMA and DoDMAImmediate IOCTL handlers.
**
** May be called in an arbitrary context.
*/
static void
doDMACallback(
  DeviceContext*  devCtx,
  ADMXRC_DMANODE* node)
{
  OS_EVENT* event;

  event = (OS_EVENT*) node->client.context;
  os_signal_event(event);
}

/*
** Handle ADMXRC_IOCTL_DODMA
**
** Must be called in a thread context.
*/
OS_STATUS
ADMXRC_DoDMA(
  DeviceContext* devCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_DODMA* iptr;
  ADMXRC_DMANODE*      dmanode = NULL;
  ADMXRC_DMATGT*       dmatgt = NULL;
  OS_STATUS            status = OS_STATUS_SUCCESS;
  unsigned long        spanned;
  os_uint32            local;
  unsigned long        length;
  os_uint32            mode;
  unsigned long        flags;
  unsigned long        channel;
  unsigned long        dir;
  unsigned long        offset;
  unsigned char        queue;
  OS_EVENT             event;
  unsigned long        submit_len, submit_offs;
  os_boolean           must_start;

  iptr = (ADMXRC_IOCTLS_DODMA*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(*iptr)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  if (!devCtx->info.cardInfo.NumDMAChan) {
    return OS_MAKE_STATUS(ADMXRC_NOT_SUPPORTED);
  }

  /* Validate PLX DMA mode */
  mode = iptr->in.mode;
  if (mode & 0xffff2600U) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  /* Validate direction */
  dir = iptr->in.direction;
  if (dir != ADMXRC_PCI2LOCAL && dir != ADMXRC_LOCAL2PCI) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  /* Validate local bus address & length */
  local = iptr->in.local;
  length = iptr->in.length;
  if (length == 0) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  /* Validate flags parameter */
  flags = iptr->in.flags;
  if (flags & ~(ADMXRC_DMAFLAG_TIMEOUT | ADMXRC_DMAFLAG_DONOTQUEUE)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  flags &= ~ADMXRC_DMAFLAG_TIMEOUT;

  /* Validate channel parameter */
  channel = iptr->in.channel;
  if (channel != ADMXRC_DMACHAN_ANY && channel >= devCtx->info.cardInfo.NumDMAChan) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  dmanode = (ADMXRC_DMANODE*) os_alloc_lcache(devCtx->dma.ctlPool, 0);
  if (dmanode == NULL) {
    return OS_MAKE_STATUS(ADMXRC_NO_MEMORY);
  }

  /* Lock down the user buffer and obtain an MDL for it */
  dmatgt = ADMXRC_GetDMATarget(devCtx, iptr->in.dmadesc);
  if (dmatgt == NULL) {
    status = OS_MAKE_STATUS(ADMXRC_INVALID_DMADESC);
    goto cleanup;
  }

  /* Validate the offset and size parameters */
  offset = iptr->in.offset;
  if (offset >= dmatgt->length || (offset + length) > dmatgt->length) {
    status = OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
    goto cleanup;
  }

  /* Determine which dmaQueue to add this node to */
  queue = (channel == ADMXRC_DMACHAN_ANY) ? devCtx->info.cardInfo.NumDMAChan : (unsigned char) channel;

  /* Prepare the node for insertion into DMA queue */
  dmanode->client.context = (void*) &event;
  dmanode->client.callback = doDMACallback;
  dmanode->client.direction = (unsigned char) dir;
  dmanode->client.length = length;
  dmanode->client.pos = 0;
  dmanode->client.mode = mode;
  dmanode->client.queue = queue;

  os_dbgprint(6, ("ADMXRC_DoDMA: enqueuing DMANODE 0x%p dmadesc=%ld offset=%ld length=%ld loc=0x%lx dir=%ld chan=%ld mode=0x%lx flags=0x%lx q=%ld\n",
		  (void*) dmanode,
		  (unsigned long) iptr->in.dmadesc,
		  (unsigned long) offset,
		  (unsigned long) length,
		  (unsigned long) local,
		  (unsigned long) dir,
		  (unsigned long) channel,
		  (unsigned long) mode,
		  (unsigned long) flags,
		  (unsigned long) queue));

  submit_offs = offset;
  submit_len = length;
  status = OS_STATUS_SUCCESS;
  req->user.context = (void*) dmanode;
  os_init_event(&event);

  must_start = (flags & ADMXRC_DMAFLAG_DONOTQUEUE) != 0;
  if (ADMXRC_EnqueueDma(devCtx, dmanode, must_start)) {
    goto completed;
  }

  while (1) {
    os_boolean    signalled;

    signalled = os_wait_event(&event, TRUE);
    if (signalled) {
      if (ADMXRC_CancelDma(devCtx, dmanode)) {
	goto completed;
      }
      os_wait_event(&event, FALSE);
    }

    switch (dmanode->status.reason) {
    case ADMXRC_DMACALLBACK_REASON_SUBMIT:
      {
	ADMXRC_SubmitDmaUsbd(devCtx, dmanode, dmatgt->usbd, submit_offs, submit_len, local, &spanned);
	submit_offs += spanned;
	submit_len -= spanned;
	if (!(mode & ADMXRC_DMAMODE_FIXEDLOCAL)) {
	  local += spanned;
	}
	break;
      }

    case ADMXRC_DMACALLBACK_REASON_COLLECT:
      {
	ADMXRC_CollectDma(devCtx, dmanode);
	break;
      }

    case ADMXRC_DMACALLBACK_REASON_COMPLETED:
      goto completed;
      break;

    default:
      os_assert(FALSE);
      break;
    }

    os_reset_event(&event);
    if (ADMXRC_ContinueDma(devCtx, dmanode)) {
      goto completed;
    }
  }

 completed:
  status = dmanode->status.status;

 cleanup:
  if (dmatgt != NULL) {
    ADMXRC_ReleaseDMATarget(devCtx, dmatgt);
  }
  if (dmanode != NULL) {
    os_free_lcache(devCtx->dma.ctlPool, dmanode);
  }

  return status;
}

/*
** Handle ADMXRC_IOCTL_DODMA2D
**
** Must be called in a thread context.
*/
OS_STATUS
ADMXRC_DoDMA2D(
  DeviceContext* devCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_DODMA2D* iptr;
  ADMXRC_DMANODE*        dmanode = NULL;
  ADMXRC_DMATGT*         dmatgt = NULL;
  OS_STATUS              status = OS_STATUS_SUCCESS;
  unsigned long          spanned, local_spanned;
  os_uint32              local;
  unsigned long          width;
  unsigned long          skip, local_skip;
  unsigned long          length;
  os_uint32              mode;
  unsigned long          flags;
  unsigned long          channel;
  unsigned long          dir;
  unsigned long          offset;
  unsigned char          queue;
  OS_EVENT               event;
  unsigned long          submit_len, submit_offs;
  os_boolean             must_start;

  iptr = (ADMXRC_IOCTLS_DODMA2D*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(*iptr)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  if (!devCtx->info.cardInfo.NumDMAChan) {
    return OS_MAKE_STATUS(ADMXRC_NOT_SUPPORTED);
  }

  /* Validate PLX DMA mode */
  mode = iptr->in.mode;
  if (mode & 0xffff2600U) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  /* Validate direction */
  dir = iptr->in.direction;
  if (dir != ADMXRC_PCI2LOCAL && dir != ADMXRC_LOCAL2PCI) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  /* Validate local bus address & length */
  local = iptr->in.local;
  length = iptr->in.length;
  if (length == 0) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  /* Validate the 2D parameters */
  {
    unsigned long n_full_strip, n_skip;
    unsigned long partial_strip;

    width = iptr->in.width;
    skip = iptr->in.skip;
    local_skip = iptr->in.local_skip;

    n_full_strip = length / width;
    partial_strip = length % width;
    if (partial_strip) {
      n_skip = n_full_strip;
    } else {
      n_skip = n_full_strip - 1;
    }
    spanned = width * n_full_strip + partial_strip;
    local_spanned = spanned + n_skip * local_skip;
    spanned += n_skip * skip;
  }

  /* Validate flags parameter */
  flags = iptr->in.flags;
  if (flags & ~(ADMXRC_DMAFLAG_TIMEOUT | ADMXRC_DMAFLAG_DONOTQUEUE)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  flags &= ~ADMXRC_DMAFLAG_TIMEOUT;

  /* Validate channel parameter */
  channel = iptr->in.channel;
  if (channel != ADMXRC_DMACHAN_ANY && channel >= devCtx->info.cardInfo.NumDMAChan) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  dmanode = (ADMXRC_DMANODE*) os_alloc_lcache(devCtx->dma.ctlPool, 0);
  if (dmanode == NULL) {
    return OS_MAKE_STATUS(ADMXRC_NO_MEMORY);
  }

  /* Lock down the user buffer and obtain an MDL for it */
  dmatgt = ADMXRC_GetDMATarget(devCtx, iptr->in.dmadesc);
  if (dmatgt == NULL) {
    status = OS_MAKE_STATUS(ADMXRC_INVALID_DMADESC);
    goto cleanup;
  }

  /* Validate the offset and size parameters */
  offset = iptr->in.offset;
  if (offset >= dmatgt->length || (offset + spanned) > dmatgt->length) {
    status = OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
    goto cleanup;
  }

  /* Determine which dmaQueue to add this node to */
  queue = (channel == ADMXRC_DMACHAN_ANY) ? devCtx->info.cardInfo.NumDMAChan : (unsigned char) channel;

  /* Prepare the node for insertion into DMA queue */
  dmanode->client.context = (void*) &event;
  dmanode->client.callback = doDMACallback;
  dmanode->client.direction = (unsigned char) dir;
  dmanode->client.length = length;
  dmanode->client.pos = 0;
  dmanode->client.mode = mode;
  dmanode->client.queue = queue;

  os_dbgprint(6, ("ADMXRC_DoDMA: enqueuing DMANODE 0x%p dmadesc=%ld offset=%ld length=%ld loc=0x%lx dir=%ld chan=%ld mode=0x%lx flags=0x%lx q=%ld\n",
		  (void*) dmanode,
		  (unsigned long) iptr->in.dmadesc,
		  (unsigned long) offset,
		  (unsigned long) length,
		  (unsigned long) local,
		  (unsigned long) dir,
		  (unsigned long) channel,
		  (unsigned long) mode,
		  (unsigned long) flags,
		  (unsigned long) queue));

  submit_offs = offset;
  submit_len = length;
  status = OS_STATUS_SUCCESS;
  req->user.context = (void*) dmanode;
  os_init_event(&event);

  must_start = (flags & ADMXRC_DMAFLAG_DONOTQUEUE) != 0;
  if (ADMXRC_EnqueueDma(devCtx, dmanode, must_start)) {
    goto completed;
  }

  while (1) {
    os_boolean    signalled;
    unsigned long chunk;

    signalled = os_wait_event(&event, TRUE);
    if (signalled) {
      if (ADMXRC_CancelDma(devCtx, dmanode)) {
	goto completed;
      }
      os_wait_event(&event, FALSE);
    }

    switch (dmanode->status.reason) {
    case ADMXRC_DMACALLBACK_REASON_SUBMIT:
      {
	while (1) {
	  chunk = (submit_len > width) ? width : submit_len;

	  ADMXRC_SubmitDmaUsbd(devCtx, dmanode, dmatgt->usbd, submit_offs, chunk, local, &spanned);
	  if (!spanned) {
	    break;
	  }
	  submit_offs += (spanned + skip);
	  submit_len -= spanned;
	  if (!(mode & ADMXRC_DMAMODE_FIXEDLOCAL)) {
	    local += (spanned + local_skip);
	  }
	}
	break;
      }

    case ADMXRC_DMACALLBACK_REASON_COLLECT:
      {
	ADMXRC_CollectDma(devCtx, dmanode);
	break;
      }

    case ADMXRC_DMACALLBACK_REASON_COMPLETED:
      goto completed;
      break;

    default:
      os_assert(FALSE);
      break;
    }

    os_reset_event(&event);
    if (ADMXRC_ContinueDma(devCtx, dmanode)) {
      goto completed;
    }
  }
 completed:
  status = dmanode->status.status;

 cleanup:
  if (dmatgt != NULL) {
    ADMXRC_ReleaseDMATarget(devCtx, dmatgt);
  }
  if (dmanode != NULL) {
    os_free_lcache(devCtx->dma.ctlPool, dmanode);
  }

  return status;
}

/*
** Handle ADMXRC_IOCTL_DODMAIMMEDIATE
**
** Must be called in a thread context.
*/
OS_STATUS
ADMXRC_DoDMAImmediate(
  DeviceContext* devCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_DODMAIMMEDIATE*	iptr;
  ADMXRC_DMANODE*               dmanode;
  ADMXRC_DMATGT*                dmatgt = NULL;
  OS_STATUS                     status;
  unsigned long                 spanned;
  os_uint32                     local;
  unsigned long                 length;
  os_uint32                     mode;
  unsigned long                 flags;
  unsigned long                 channel;
  unsigned long                 dir;
  unsigned char                 queue;
  void*                         buffer;
  OS_EVENT                      event;
  unsigned long                 submit_len, submit_offs;
  os_boolean                    must_start;

  iptr = (ADMXRC_IOCTLS_DODMAIMMEDIATE*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(*iptr)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  if (!devCtx->info.cardInfo.NumDMAChan) {
    return OS_MAKE_STATUS(ADMXRC_NOT_SUPPORTED);
  }

  /* Validate PLX DMA mode */
  mode = iptr->in.mode;
  if (mode & 0xffff2600U) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  /* Validate direction */
  dir = iptr->in.direction;
  if (dir != ADMXRC_PCI2LOCAL && dir != ADMXRC_LOCAL2PCI) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  /* Validate local bus address & length */
  buffer = iptr->in.buffer;
  local = iptr->in.local;
  length = iptr->in.length;
  if (length == 0) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  /* Validate flags parameter */
  flags = iptr->in.flags;
  if (flags & ~(ADMXRC_DMAFLAG_TIMEOUT | ADMXRC_DMAFLAG_DONOTQUEUE)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  flags &= ~ADMXRC_DMAFLAG_TIMEOUT;

  /* Validate channel parameter */
  channel = iptr->in.channel;
  if (channel != ADMXRC_DMACHAN_ANY && channel >= devCtx->info.cardInfo.NumDMAChan) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  dmanode = (ADMXRC_DMANODE*) os_alloc_lcache(devCtx->dma.ctlPool, 0);
  if (dmanode == NULL) {
    status = OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
    goto cleanup;
  }

  /* Lock down the user buffer and obtain an MDL for it */
  status = ADMXRC_CreateDMATarget(devCtx, buffer, length, &dmatgt);
  if (!OS_SUCCESS(status)) {
    goto cleanup;
  }

  /* Determine which dmaQueue to add this node to */
  queue = (channel == ADMXRC_DMACHAN_ANY) ? devCtx->info.cardInfo.NumDMAChan : (unsigned char) channel;

  /* Prepare the node for insertion into DMA queue */
  dmanode->client.context = (void*) &event;
  dmanode->client.callback = doDMACallback;
  dmanode->client.direction = (unsigned char) dir;
  dmanode->client.length = length;
  dmanode->client.pos = 0;
  dmanode->client.mode = mode;
  dmanode->client.queue = queue;

  os_dbgprint(6, ("ADMXRC_DoDMAImmediate: enqueuing DMANODE 0x%p buffer=0x%p length=0x%ld loc=0x%lx dir=%ld chan=%ld mode=0x%lx flags=0x%lx q=%ld\n",
		  (void*) dmanode,
		  (void*) buffer,
		  (unsigned long) length,
		  (unsigned long) local,
		  (unsigned long) dir,
		  (unsigned long) channel,
		  (unsigned long) mode,
		  (unsigned long) flags,
		  (unsigned long) queue));

  submit_offs = 0;
  submit_len = length;
  status = OS_STATUS_SUCCESS;
  req->user.context = (void*) dmanode;
  os_init_event(&event);

  must_start = (flags & ADMXRC_DMAFLAG_DONOTQUEUE) != 0;
  if (ADMXRC_EnqueueDma(devCtx, dmanode, must_start)) {
    goto completed;
  }

  while (1) {
    os_boolean    signalled;

    signalled = os_wait_event(&event, TRUE);
    if (signalled) {
      if (ADMXRC_CancelDma(devCtx, dmanode)) {
	goto completed;
      }
      os_wait_event(&event, FALSE);
    }

    switch (dmanode->status.reason) {
    case ADMXRC_DMACALLBACK_REASON_SUBMIT:
      {
	ADMXRC_SubmitDmaUsbd(devCtx, dmanode, dmatgt->usbd, submit_offs, submit_len, local, &spanned);
	submit_offs += spanned;
	submit_len -= spanned;	
	if (!(mode & ADMXRC_DMAMODE_FIXEDLOCAL)) {
	  local += spanned;
	}
	break;
      }

    case ADMXRC_DMACALLBACK_REASON_COLLECT:
      {
	ADMXRC_CollectDma(devCtx, dmanode);
	break;
      }

    case ADMXRC_DMACALLBACK_REASON_COMPLETED:
      goto completed;
      break;

    default:
      os_assert(FALSE);
      break;
    }

    os_reset_event(&event);
    if (ADMXRC_ContinueDma(devCtx, dmanode)) {
      goto completed;
    }
  }
 completed:
  status = dmanode->status.status;

 cleanup:
  if (dmatgt != NULL) {
    ADMXRC_ReleaseDMATarget(devCtx, dmatgt);
  }
  if (dmanode != NULL) {
    os_free_lcache(devCtx->dma.ctlPool, dmanode);
  }

  return status;
}

/*
** Handle ADMXRC_IOCTL_DODMAIMMEDIATE2D
**
** Must be called in a thread context.
*/
OS_STATUS
ADMXRC_DoDMAImmediate2D(
  DeviceContext* devCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_DODMAIMMEDIATE2D* iptr;
  ADMXRC_DMANODE*                 dmanode;
  ADMXRC_DMATGT*                  dmatgt = NULL;
  OS_STATUS                       status;
  unsigned long                   spanned, local_spanned;
  os_uint32                       local;
  unsigned long                   width;
  unsigned long                   skip, local_skip;
  unsigned long                   length;
  os_uint32                       mode;
  unsigned long                   flags;
  unsigned long                   channel;
  unsigned long                   dir;
  unsigned char                   queue;
  void*                           buffer;
  OS_EVENT                        event;
  unsigned long                   submit_len, submit_offs;
  os_boolean                      must_start;

  iptr = (ADMXRC_IOCTLS_DODMAIMMEDIATE2D*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(*iptr)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  /* Validate PLX DMA mode */
  mode = iptr->in.mode;
  if (mode & 0xffff2600U) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  /* Validate direction */
  dir = iptr->in.direction;
  if (dir != ADMXRC_PCI2LOCAL && dir != ADMXRC_LOCAL2PCI) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  /* Validate local bus address & length */
  buffer = iptr->in.buffer;
  local = iptr->in.local;
  length = iptr->in.length;
  if (length == 0) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  /* Validate the 2D parameters */
  {
    unsigned long n_full_strip, n_skip;
    unsigned long partial_strip;

    width = iptr->in.width;
    skip = iptr->in.skip;
    local_skip = iptr->in.local_skip;

    n_full_strip = length / width;
    partial_strip = length % width;
    if (partial_strip) {
      n_skip = n_full_strip;
    } else {
      n_skip = n_full_strip - 1;
    }
    spanned = width * n_full_strip + partial_strip;
    local_spanned = spanned + n_skip * local_skip;
    spanned += n_skip * skip;
  }

  /* Validate flags parameter */
  flags = iptr->in.flags;
  if (flags & ~(ADMXRC_DMAFLAG_TIMEOUT | ADMXRC_DMAFLAG_DONOTQUEUE)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  flags &= ~ADMXRC_DMAFLAG_TIMEOUT;

  /* Validate channel parameter */
  channel = iptr->in.channel;
  if (channel != ADMXRC_DMACHAN_ANY && channel >= devCtx->info.cardInfo.NumDMAChan) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  dmanode = (ADMXRC_DMANODE*) os_alloc_lcache(devCtx->dma.ctlPool, 0);
  if (dmanode == NULL) {
    status = OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
    goto cleanup;
  }

  /* Lock down the user buffer and obtain an MDL for it */
  status = ADMXRC_CreateDMATarget(devCtx, buffer, spanned, &dmatgt);
  if (!OS_SUCCESS(status)) {
    goto cleanup;
  }

  /* Determine which dmaQueue to add this node to */
  queue = (channel == ADMXRC_DMACHAN_ANY) ? devCtx->info.cardInfo.NumDMAChan : (unsigned char) channel;

  /* Prepare the node for insertion into DMA queue */
  dmanode->client.context = (void*) &event;
  dmanode->client.callback = doDMACallback;
  dmanode->client.direction = (unsigned char) dir;
  dmanode->client.length = length;
  dmanode->client.pos = 0;
  dmanode->client.mode = mode;
  dmanode->client.queue = queue;

  os_dbgprint(6, ("ADMXRC_DoDMAImmediate: enqueuing DMANODE 0x%p buffer=0x%p length=0x%ld loc=0x%lx dir=%ld chan=%ld mode=0x%lx flags=0x%lx q=%ld\n",
		  (void*) dmanode,
		  (void*) buffer,
		  (unsigned long) length,
		  (unsigned long) local,
		  (unsigned long) dir,
		  (unsigned long) channel,
		  (unsigned long) mode,
		  (unsigned long) flags,
		  (unsigned long) queue));

  submit_offs = 0;
  submit_len = length;
  status = OS_STATUS_SUCCESS;
  req->user.context = (void*) dmanode;
  os_init_event(&event);

  must_start = (flags & ADMXRC_DMAFLAG_DONOTQUEUE) != 0;
  if (ADMXRC_EnqueueDma(devCtx, dmanode, must_start)) {
    goto completed;
  }

  while (1) {
    os_boolean    signalled;
    unsigned long chunk;

    signalled = os_wait_event(&event, TRUE);
    if (signalled) {
      if (ADMXRC_CancelDma(devCtx, dmanode)) {
	goto completed;
      }
      os_wait_event(&event, FALSE);
    }

    switch (dmanode->status.reason) {
    case ADMXRC_DMACALLBACK_REASON_SUBMIT:
      {
	while (1) {
	  chunk = (submit_len > width) ? width : submit_len;

	  ADMXRC_SubmitDmaUsbd(devCtx, dmanode, dmatgt->usbd, submit_offs, chunk, local, &spanned);
	  if (!spanned) {
	    break;
	  }
	  submit_offs += (spanned + skip);
	  submit_len -= spanned;	
	  if (!(mode & ADMXRC_DMAMODE_FIXEDLOCAL)) {
	    local += (spanned + local_skip);
	  }
	  break;
	}
      }

    case ADMXRC_DMACALLBACK_REASON_COLLECT:
      {
	ADMXRC_CollectDma(devCtx, dmanode);
	break;
      }

    case ADMXRC_DMACALLBACK_REASON_COMPLETED:
      goto completed;
      break;

    default:
      os_assert(FALSE);
      break;
    }

    os_reset_event(&event);
    if (ADMXRC_ContinueDma(devCtx, dmanode)) {
      goto completed;
    }
  }
 completed:
  status = dmanode->status.status;

 cleanup:
  if (dmatgt != NULL) {
    ADMXRC_ReleaseDMATarget(devCtx, dmatgt);
  }
  if (dmanode != NULL) {
    os_free_lcache(devCtx->dma.ctlPool, dmanode);
  }

  return status;
}
