/*
** ADM-XRC Series Driver
**
** (C) Copyright 2001-2004 Alpha Data 
**
** dmacore.c - Core functions to support DMA transfers.
**
** ----------------------------------------------------------------------------
**
** Exported functions are:
**
**   ADMXRC_EnqueueDma             Queues or begins a DMA transfer
**
**   ADMXRC_SubmitDmaUsbd          Submit region of a USBD as a target for DMA transfer
**
**   ADMXRC_SubmitDmaUser          Submit user-space buffer as a target for DMA transfer
**
**   ADMXRC_CollectDma             Must be called to ensure coherency of DMA targets
**
**   ADMXRC_ContinueDma            Directs DMA core to use the submitted buffers to
**                                 carry out the DMA transfer
**
**   ADMXRC_CancelDma              Cancels a DMA transfer
**
**   ADMXRC_CleanupDma             Called when shutting down the device; aborts all DMA
**                                 transfers and completes them with a failure status.
**
**
** Primary datatype is the ADMXRC_DMANODE, which is a control structure
** for a DMA operation.
**
** A DMA node's state is as follows:
**
**   QUEUED      => in a DMA queue
**   WAITBUFFERS => waiting for client to supply data buffers
**   ACTIVE      => active on the hardware
**   WAITCOLLECT => waiting for client to collect data buffers
**   COMPLETED   => completed by hardware
**   other       => not a valid state
**
** Each DMA node has an attention callback function associated with it,
** called when either:
**
**   (a) data buffers need to be submitted by the client
**   (b) data buffers need to be collected by the client
**   (c) the DMA transfer has been completed
**
** This callback function may be invoked by the DMA core in an arbitrary
** context.
*/

#include "platform.h"
#include "device.h"
#include "ioctl.h"
#include "dmacore.h"
#include "dmadesc.h"

#define ADMXRC_DMASTATE_QUEUED      (1)
#define ADMXRC_DMASTATE_WAITBUFFERS (2)
#define ADMXRC_DMASTATE_ACTIVE      (3)
#define ADMXRC_DMASTATE_WAITCOLLECT (4)
#define ADMXRC_DMASTATE_COMPLETED   (5)

/*
** Removes a DMANODE from (anywhere in) a queue.
**
** Must be called while holding the DMA spinlock
*/
static void
removeDMA(
  ADMXRC_DMAQ*    q,
  ADMXRC_DMANODE* node)
{
  ADMXRC_DMANODE* next;
  ADMXRC_DMANODE* prev;

  next = node->system.next;
  prev = node->system.prev;
  if (next != NULL) {
    next->system.prev = prev;
  } else {
    q->tail = prev;
  }
  if (prev != NULL) {
    prev->system.next = next;
  } else {
    q->head = next;
  }
  q->length--;
}

/*
** Adds a DMANODE to the tail of a queue.
**
** Must be called while holding the DMA spinlock
*/
static void
addDMA(
  ADMXRC_DMAQ*    q,
  ADMXRC_DMANODE* node)
{
  ADMXRC_DMANODE* head;
  ADMXRC_DMANODE* tail;
  
  head = q->head;
  tail = q->tail;
  if (head == NULL) {
    q->tail = q->head = node;
    node->system.next = NULL;
    node->system.prev = NULL;
  } else {
    tail->system.next = node;
    q->tail = node;
    node->system.prev = tail;
    node->system.next = NULL;
  }
  q->length++;
}

/*
** Critical section that aborts a DMA transfer on a hardware DMA
** channel. It must not run concurrently with the interrupt handler.
*/
static void*
abortDMAOnHardwareCritical(
  void* arg)
{
  DmaAbortContext* context = (DmaAbortContext*) arg;
  DeviceContext*   pDevCtx = context->devCtx;
  unsigned int     chan = context->channel;
  MethodResult     res;

  res = pDevCtx->methods.doDma(pDevCtx, chan, FALSE, 0, 0);

  context->success = (res == MethodResult_Success);

  return NULL;
}

/*
** Called to abort DMA transfer that is active on the hardware.
**
** Returns TRUE if DMA transfer was successfully aborted, otherwise FALSE.
*/
static os_boolean
abortDMAOnHardware(
  DeviceContext* pDevCtx,
  unsigned char  chan)
{
  DmaAbortContext context;

  context.devCtx = pDevCtx;
  context.channel = chan;
  os_invoke_atomic(pDevCtx->irqObj, abortDMAOnHardwareCritical, &context);

  return context.success;
}

/*
** Get & dequeue the next DMANODE, if it exists, on an idle DMA channel.
**
** Looks first for a DMANODE in the queue associated with the
** specified DMA channel, and if not found, looks in the 'any channel'
** queue. If a DMANODE is found, it is made active on the hardware.
**
** Must be called while holding the DMA spinlock.
*/
static ADMXRC_DMANODE*
getNextDMA(
  DeviceContext* pDevCtx,
  unsigned char  chan)
{
  ADMXRC_DMANODE* node;
  unsigned char   queue;

  os_dbgprint(6, ("getNextDMA: chan = %ld\n", (unsigned long) chan));
  
  /*
  ** The DMA channel is now idle - look for a candidate DMANODE
  ** to make active on the DMA channel.
  */
  node = pDevCtx->dma.queue[chan].head;
  if (node == NULL) {
    node = pDevCtx->dma.queue[pDevCtx->info.cardInfo.NumDMAChan].head;
    if (node != NULL) {
      /* Found one in the 'any' queue */
      queue = pDevCtx->info.cardInfo.NumDMAChan;
      goto found;
    }
  } else {
    /* Found one */
    queue = chan;
    goto found;
  }

  /* No DMANODE found - can't start a new DMA operation on the hardware */
  return NULL;

 found:
  removeDMA(&pDevCtx->dma.queue[queue], node);
  node->system.channel = chan;
  node->system.state = ADMXRC_DMASTATE_WAITBUFFERS;
  pDevCtx->dma.channel[chan].chunk = 0;
  pDevCtx->dma.channel[chan].waitsubmit = node;
  pDevCtx->dma.channel[chan].cleanupPending = FALSE;
  pDevCtx->dma.channel[chan].count++;
  if (pDevCtx->dma.channel[chan].count != 1) {
    os_dbgprint(0, ("*** getNextDMA: %ld job(s) on DMA channel %ld\n",
		    (unsigned long) pDevCtx->dma.channel[chan].count, (unsigned long) chan));
  }  
  
  os_dbgprint(6, ("getNextDMA: next node is 0x%p on DMA channel %ld\n",
		  (void*) node, (unsigned long) chan));
  return node;
}

/*
** Queue up a DMANODE
*/
os_boolean
ADMXRC_EnqueueDma(
  DeviceContext*  pDevCtx,
  ADMXRC_DMANODE* node,
  os_boolean      must_start)
{
  OS_SPINLOCK_FLAGS f;
  unsigned char     queue;
  unsigned char     chan;
  os_boolean        error = FALSE;
  os_boolean        canstart = FALSE;

  queue = node->client.queue;
  os_dbgprint(6, ("ADMXRC_EnqueueDMA: entered for DMA node 0x%p on queue %ld\n",
		  (void*) node, (unsigned long) queue));

  os_acquire_spinlock(&pDevCtx->dma.lock, &f);

  /*
  ** Attempt to bypass queueing mechanism - check if there is a DMA channel
  ** free on which to start the DMA transfer immediately
  */
  if (queue < pDevCtx->info.cardInfo.NumDMAChan) {
    /* Must use a specific DMA channel */
    chan = queue;
    if (!pDevCtx->dma.channel[chan].count && pDevCtx->dma.queue[chan].head == NULL) {
      canstart = TRUE;
    }
  } else {
    /* Can use any DMA channel */
    for (chan = 0; chan < pDevCtx->info.cardInfo.NumDMAChan; chan++) {
      if (!pDevCtx->dma.channel[chan].count && pDevCtx->dma.queue[chan].head == NULL) {
	canstart = TRUE;
	break;
      }
    }
  }

  if (canstart) {
    node->system.channel = chan;
    node->system.state = ADMXRC_DMASTATE_WAITBUFFERS;
    pDevCtx->dma.channel[chan].chunk = 0;
    pDevCtx->dma.channel[chan].waitsubmit = node;
    pDevCtx->dma.channel[chan].cleanupPending = FALSE;
    pDevCtx->dma.channel[chan].count++;
    if (pDevCtx->dma.channel[chan].count != 1) {
      os_dbgprint(0, ("*** ADMXRC_EnqueueDMA: %ld job(s) on DMA channel %ld\n",
		      (unsigned long) pDevCtx->dma.channel[chan].count, (unsigned long) chan));
    }
    os_release_spinlock(&pDevCtx->dma.lock, f);

    /* Request buffers from client */
    os_dbgprint(6, ("ADMXRC_EnqueueDMA: Calling DMA callback for node 0x%p, reason = SUBMIT\n", (void*) node));
    os_sg_init(pDevCtx->dma.channel[chan].sgMap,
	       node->client.direction == ADMXRC_LOCAL2PCI);
    node->status.reason = ADMXRC_DMACALLBACK_REASON_SUBMIT;
    node->client.callback(pDevCtx, node);
  } else {
    if (must_start) {
      os_dbgprint(6, ("ADMXRC_EnqueueDMA: must_start DMA node 0x%p cannot be started immediately\n", (void*) node));
      node->status.status = OS_MAKE_STATUS(ADMXRC_DEVICE_BUSY);
      error = TRUE;
    } else {
      /*
      ** No free DMA channel on which we can start this DMA operation
      ** immediately, so add the DMANODE to the tail of the DMA queue.
      */
      queue = node->client.queue;
      os_dbgprint(6, ("ADMXRC_EnqueueDMA: queueing DMA node 0x%p in queue %ld\n",
		      (void*) node, (unsigned long) queue));
      addDMA(&pDevCtx->dma.queue[queue], node);
      node->system.state = ADMXRC_DMASTATE_QUEUED;      
    }
    os_release_spinlock(&pDevCtx->dma.lock, f);
  }

  return error;
}

os_boolean
ADMXRC_SubmitDmaUsbd(
  DeviceContext*  pDevCtx,
  ADMXRC_DMANODE* node,
  OS_USBD*        usbd,
  unsigned long   offset,
  unsigned long   max,
  os_uint32       local,
  unsigned long*  spanned)
{
  unsigned char chan;
  unsigned long nbytes;
  os_boolean bFixedLocal;

  chan = node->system.channel;
  bFixedLocal = (node->client.mode & ADMXRC_DMAMODE_FIXEDLOCAL) != 0;
  os_dbgprint(7, ("ADMXRC_SubmitDmaUsbd: chan=%ld local=0x%lx node=0x%p usbd=0x%p offset=0x%lx max=0x%lx\n",
                  (unsigned long) chan, (unsigned long) local,
                  node, usbd,
                  (unsigned long) offset, (unsigned long) max));
  nbytes = os_sg_add_usbd_buffer(pDevCtx->dma.channel[chan].sgMap,
                                 local, bFixedLocal, usbd, offset, max);
  pDevCtx->dma.channel[chan].chunk += nbytes;
  *spanned = nbytes;

  return FALSE;
}

os_boolean
ADMXRC_SubmitDmaUser(
  DeviceContext*  pDevCtx,
  ADMXRC_DMANODE* node,
  void*           user,
  unsigned long   max,
  os_uint32       local,
  unsigned long*  spanned)
{
  unsigned char chan;
  unsigned long nbytes;
  os_boolean bFixedLocal;

  chan = node->system.channel;
  bFixedLocal = (node->client.mode & ADMXRC_DMAMODE_FIXEDLOCAL) != 0;
  os_dbgprint(7, ("ADMXRC_SubmitDmaUser: chan=%ld local=0x%lx node=0x%p user=0x%p max=0x%lx\n",
                  (unsigned long) chan, (unsigned long) local,
                  node, user,
                  (unsigned long) max));
  nbytes = os_sg_add_user_buffer(pDevCtx->dma.channel[chan].sgMap,
                                 local, bFixedLocal, user, max);
  pDevCtx->dma.channel[chan].chunk += nbytes;
  *spanned = nbytes;

  return FALSE;
}

void
ADMXRC_CollectDma(
  DeviceContext*  pDevCtx,
  ADMXRC_DMANODE* node)
{
  unsigned char    chan;

  chan = node->system.channel;
  os_dbgprint(7, ("ADMXRC_CollectDma: chan=%ld node=0x%p\n",
                  (unsigned long) chan, node));

  os_sg_unmap(pDevCtx->dma.channel[chan].sgMap);
}

os_boolean
ADMXRC_ContinueDma(
  DeviceContext*  pDevCtx,
  ADMXRC_DMANODE* node)
{
  OS_SPINLOCK_FLAGS f;
  unsigned char     chan;
  ADMXRC_DMANODE*   nextnode;
  unsigned long     length;
  os_boolean        bCleanup, bCompleted, bFailed, bWrite;
  unsigned int      i, nbuff;
  OS_SG_MAP*        map;
  os_uint8          completionCode;

  chan = node->system.channel;

  os_acquire_spinlock(&pDevCtx->dma.lock, &f);
  bCleanup = pDevCtx->dma.channel[chan].cleanupPending;
  switch (node->system.state) {
  case ADMXRC_DMASTATE_WAITBUFFERS:
    {
      if (bCleanup) {
	os_dbgprint(6, ("ADMXRC_ContinueDMA: cleanup pending in WAITSUBMIT state for node 0x%p on channel %ld\n",
			(void*) node,
			(unsigned long) chan));

	pDevCtx->dma.channel[chan].waitsubmit = NULL;
	pDevCtx->dma.channel[chan].count--;
	node->system.state = ADMXRC_DMASTATE_COMPLETED;
	nextnode = getNextDMA(pDevCtx, chan);	  
	os_release_spinlock(&pDevCtx->dma.lock, f);

	if (nextnode != NULL) {
	  os_dbgprint(6, ("ADMXRC_ContinueDMA: calling DMA callback for next node 0x%p, reason = SUBMIT (cleanup)\n",
			  (void*) nextnode));
	  os_sg_init(pDevCtx->dma.channel[chan].sgMap,
		     nextnode->client.direction == ADMXRC_LOCAL2PCI);
	  nextnode->status.reason = ADMXRC_DMACALLBACK_REASON_SUBMIT;
	  nextnode->client.callback(pDevCtx, nextnode);
	}

	node->status.status = OS_MAKE_STATUS(ADMXRC_CANCELLED);
	return TRUE;
      } else {
	os_dbgprint(6, ("ADMXRC_ContinueDMA: called in WAITSUBMIT state for node 0x%p on channel %ld\n",
			(void*) node, (unsigned long) chan));
	
	node->system.state = ADMXRC_DMASTATE_ACTIVE;
	if (pDevCtx->dma.channel[chan].active != NULL) {
	  os_dbgprint(0, ("*** ADMXRC_ContinueDMA: job 0x%p active on DMA channel %ld hardware\n",
			  (void*) pDevCtx->dma.channel[chan].active, (unsigned long) chan));
	}
	pDevCtx->dma.channel[chan].waitsubmit = NULL;
	pDevCtx->dma.channel[chan].active = node;

        bWrite = (node->client.direction == ADMXRC_PCI2LOCAL);
        map = pDevCtx->dma.channel[chan].sgMap;
        nbuff = os_sg_map(map);
        for (i = 0; i < nbuff; i++) {
          DmaScatterNode snode;
          os_uint32 local;
          os_uint64 logical;
          unsigned long length;

          os_sg_get_buffer(map, i, &local, &logical, &length);
          snode.localAddr = local;
          snode.busAddr = os_uint64_lo(logical);
          snode.length = (os_uint32) length;
          snode.bWrite = bWrite;
          pDevCtx->methods.addDmaBuffer(pDevCtx, chan, i, &snode);
        }

	pDevCtx->methods.doDma(pDevCtx, chan, TRUE, node->client.mode, nbuff);

	os_release_spinlock(&pDevCtx->dma.lock, f);
      }
      break;
    }

  case ADMXRC_DMASTATE_WAITCOLLECT:
    {
      length = node->client.length;

      os_dbgprint(6, ("ADMXRC_ContinueDMA: called in WAITCOLLECT state for node 0x%p length = %ld on channel %ld\n",
		      (void*) node,
		      (unsigned long) length,
		      (unsigned long) chan));

      bFailed = FALSE;
      bCompleted = !length;
      completionCode = pDevCtx->dma.channel[chan].completionCode;
      switch (completionCode) {
      case DmaCompletionCode_Normal:
	break;
      case DmaCompletionCode_EotRequested:
	bCompleted = TRUE;
	break;
      case DmaCompletionCode_Failed:
	bCompleted = TRUE;
	bFailed = TRUE;
	break;
      }

      if (bCleanup || bCompleted) {
	pDevCtx->dma.channel[chan].waitcollect = NULL;
	pDevCtx->dma.channel[chan].count--;
	node->system.state = ADMXRC_DMASTATE_COMPLETED;
	nextnode = getNextDMA(pDevCtx, chan);
	os_release_spinlock(&pDevCtx->dma.lock, f);
	
	os_dbgprint(6, ("ADMXRC_ContinueDMA: calling DMA callback for node 0x%p, reason = COMPLETED (success)\n",
			(void*) node));

	if (bCleanup) {
	  node->status.status = OS_MAKE_STATUS(ADMXRC_CANCELLED);
	} else {
	  node->status.status = bFailed ? OS_MAKE_STATUS(ADMXRC_FAILED) : OS_STATUS_SUCCESS;
	  node->status.reason = ADMXRC_DMACALLBACK_REASON_COMPLETED;
	  node->client.callback(pDevCtx, node);
	}
	
	if (nextnode != NULL) {
	  os_dbgprint(6, ("ADMXRC_ContinueDMA: calling DMA callback for next node 0x%p, reason = SUBMIT\n",
			  (void*) nextnode));
	  os_sg_init(pDevCtx->dma.channel[chan].sgMap,
		     nextnode->client.direction == ADMXRC_LOCAL2PCI);
	  nextnode->status.reason = ADMXRC_DMACALLBACK_REASON_SUBMIT;
	  nextnode->client.callback(pDevCtx, nextnode);
	}
      } else {
	/*
	** Do another chunk of DMA on the same node.
	**
	** Request more buffers from client.
	*/
	node->system.state = ADMXRC_DMASTATE_WAITBUFFERS;
	pDevCtx->dma.channel[chan].chunk = 0;
	pDevCtx->dma.channel[chan].waitsubmit = node;
	os_release_spinlock(&pDevCtx->dma.lock, f);
	
	os_dbgprint(6, ("ADMXRC_ContinueDMA: calling DMA callback for DMANODE 0x%p, reason = SUBMIT\n",
			(void*) node));
        os_sg_init(pDevCtx->dma.channel[chan].sgMap,
                   node->client.direction == ADMXRC_LOCAL2PCI);
	node->status.reason = ADMXRC_DMACALLBACK_REASON_SUBMIT;
	node->client.callback(pDevCtx, node);
      }
    }
    break;

  default:
    os_assert(FALSE);
    break;
  }

  return bCleanup;
}

/*
** This function is called to cancel a DMA transfer whose state
** is one of 'queued', 'active on the hardware', or 'completed'.
**
** If the DMA transfer was queued or active, TRUE is returned.
** If the DMA transfer was completed, FALSE is returned.
**
** After this function returns, the DMANODE is guaranteed not to
** be in any DMA queue and not active on the hardware.
*/
os_boolean
ADMXRC_CancelDma(
  DeviceContext*  pDevCtx,
  ADMXRC_DMANODE* node)
{
  os_boolean        success = TRUE;
  OS_SPINLOCK_FLAGS f;
  unsigned char     chan;
  int               state;
  ADMXRC_DMANODE*   nextnode = NULL;

  os_acquire_spinlock(&pDevCtx->dma.lock, &f);
  state = node->system.state;
  chan = node->system.channel;
  switch (state) {
  case ADMXRC_DMASTATE_QUEUED:
    {
      unsigned char queue;

      queue = node->client.queue;
      removeDMA(&pDevCtx->dma.queue[queue], node);
      node->status.status = OS_MAKE_STATUS(ADMXRC_CANCELLED);
      os_dbgprint(1, ("ADMXRC_CancelDMA: DMA node 0x%p in QUEUED state cancelled successfully\n", (void*) node));
      break;
    }

  case ADMXRC_DMASTATE_ACTIVE:
    {
      abortDMAOnHardware(pDevCtx, chan);
      pDevCtx->dma.channel[chan].count--;
      pDevCtx->dma.channel[chan].active = NULL;
      nextnode = getNextDMA(pDevCtx, chan);
      node->status.status = OS_MAKE_STATUS(ADMXRC_CANCELLED);
      os_dbgprint(1, ("ADMXRC_CancelDMA: DMA node 0x%p in ACTIVE state cancelled successfully\n", (void*) node));
      break;
    }

  case ADMXRC_DMASTATE_WAITBUFFERS:
  case ADMXRC_DMASTATE_WAITCOLLECT:
    {
      /*
      ** The ball is currently in the client's court. The DMA callback has been
      ** invoked, is being invoked, or is about to be invoked with WAITBUFFERS
      ** or WAITCOLLECT reason.
      **
      ** So don't call the DMA callback.
      */
      pDevCtx->dma.channel[chan].cleanupPending = TRUE;
      if (node->system.state == ADMXRC_DMASTATE_WAITBUFFERS) {
	os_dbgprint(1, ("ADMXRC_CancelDMA: cancellation of DMA node 0x%p in WAITBUFFERS state deferred\n", (void*) node));
      } else {
	os_dbgprint(1, ("ADMXRC_CancelDMA: cancellation of DMA node 0x%p in WAITCOLLECT state deferred\n", (void*) node));
      }
      success = FALSE;
      break;
    }

  case ADMXRC_DMASTATE_COMPLETED:
    {
      /*
      ** The ball is currently in the client's court. The DMA callback has been
      ** invoked, is being invoked, or is about to be invoked with COMPLETED reason.
      **
      ** So don't call the DMA callback.
      */
      os_dbgprint(1, ("ADMXRC_CancelDMA: cancellation of DMA node 0x%p in COMPLETED state not required\n", (void*) node));
      success = FALSE;
      break;
    }

  default: /* Should never happen! */
    {
      os_dbgprint(0, ("*** ADMXRC_CancelDMA: DMA node 0x%p has invalid state (%ld)\n",
		      (void*) node, (unsigned long) state));
      success = FALSE;
      break;
    }
  }
  
  os_release_spinlock(&pDevCtx->dma.lock, f);

  if (nextnode != NULL) {
    os_dbgprint(1, ("ADMXRC_CancelDMA: calling DMA callback for next DMA node 0x%p, reason = SUBMIT\n",
		    (void*) nextnode));
    os_sg_init(pDevCtx->dma.channel[chan].sgMap,
	       nextnode->client.direction == ADMXRC_LOCAL2PCI);
    nextnode->status.reason = ADMXRC_DMACALLBACK_REASON_SUBMIT;
    nextnode->client.callback(pDevCtx, nextnode);
  }

  return success;
}

/*
** This routine is called by the DMA interrupt DPC to service a DMA interrupt
** for a particular channel.
**
** May be called outside a thread context.
*/
void
ADMXRC_DmaCompleted(
 DeviceContext* pDevCtx,
 unsigned char  chan)
{
  OS_SPINLOCK_FLAGS f;
  ADMXRC_DMANODE*   node;
  unsigned long     chunk;

  os_acquire_spinlock(&pDevCtx->dma.lock, &f);
  node = pDevCtx->dma.channel[chan].active;
  if (node == NULL) {
    os_release_spinlock(&pDevCtx->dma.lock, f);
    
    os_dbgprint(0, ("*** ADMXRC_DMACompleted: no job on channel %ld!\n", (unsigned long) chan));
  } else {
    chunk = pDevCtx->dma.channel[chan].chunk;

    os_dbgprint(6, ("ADMXRC_DMACompleted: DMA node 0x%p completed on channel %ld, chunk = %ld\n",
		    (void*) node,
		    (unsigned long) chan,
		    (unsigned long) chunk));

    pDevCtx->dma.channel[chan].active = NULL;
    node->client.pos += chunk;
    node->client.length -= chunk;
    node->system.state = ADMXRC_DMASTATE_WAITCOLLECT;
    os_release_spinlock(&pDevCtx->dma.lock, f);
    
    os_dbgprint(6, ("ADMXRC_DMACompleted: calling DMA callback for node 0x%p, reason = COLLECT\n", (void*) node));
    node->status.reason = ADMXRC_DMACALLBACK_REASON_COLLECT;
    node->client.callback(pDevCtx, node);
  }
}

/*
** DMA cleanup function; called when card is about to be closed
*/
void
ADMXRC_CleanupDma(
  DeviceContext* pDevCtx)
{
  OS_SPINLOCK_FLAGS f;
  ADMXRC_DMANODE*   node;
  ADMXRC_DMANODE*   nextnode;
  unsigned char     i;

  if (os_dbglevel(1)) {
    for (i = 0; i <= pDevCtx->info.cardInfo.NumDMAChan; i++) {
      os_dbgprint(1, ("ADMXRC_CleanupDMA: queue %ld length = %ld\n",
		      (unsigned long) i, (unsigned long) pDevCtx->dma.queue[i].length));
    }
  }

  /* Cleanup any DMA transfers that are in the QUEUED state */
  for (i = 0; i <= pDevCtx->info.cardInfo.NumDMAChan; i++) {
    while (1) {
      os_acquire_spinlock(&pDevCtx->dma.lock, &f);
      node = pDevCtx->dma.queue[i].head;
      if (node == NULL) {
	os_release_spinlock(&pDevCtx->dma.lock, f);
	break;
      } else {
	os_dbgprint(3, ("ADMXRC_CleanupDMA: Cleaning up DMANODE 0x%p on DMA queue %ld\n",
			(void*) node, (unsigned long) i));
	
	/* Remove this DMANODE from the queue */
	removeDMA(&pDevCtx->dma.queue[i], node);
	
	/* Mark the DMANODE as finished, but with an error status */
	node->system.state = ADMXRC_DMASTATE_COMPLETED;
	node->status.status = OS_MAKE_STATUS(ADMXRC_CLOSED);
	
	os_release_spinlock(&pDevCtx->dma.lock, f);
	
	node->status.reason = ADMXRC_DMACALLBACK_REASON_COMPLETED;
	node->client.callback(pDevCtx, node);
      }
    }
  }

  /* Cleanup any DMA transfers that are currently in the WAITSUBMIT state */
  for (i = 0; i < pDevCtx->info.cardInfo.NumDMAChan; i++) {
    os_acquire_spinlock(&pDevCtx->dma.lock, &f);
    node = pDevCtx->dma.channel[i].waitsubmit;
    if (node != NULL) {
      os_dbgprint(1, ("ADMXRC_CleanupDMA: Cleaning up DMANODE 0x%p in WAITSUBMIT state on channel %ld\n",
		      (void*) node, (unsigned long) node->system.channel));
      pDevCtx->dma.channel[i].cleanupPending = TRUE;
    }
    os_release_spinlock(&pDevCtx->dma.lock, f);      
  }

  /* Cleanup any DMA transfers that are currently in the ACTIVE state */
  for (i = 0; i < pDevCtx->info.cardInfo.NumDMAChan; i++) {
    os_acquire_spinlock(&pDevCtx->dma.lock, &f);
    node = pDevCtx->dma.channel[i].active;
    if (node != NULL) {
      os_dbgprint(1, ("ADMXRC_CleanupDMA: Cleaning up DMANODE 0x%p in ACTIVE state on channel %ld\n",
		      (void*) node, (unsigned long) node->system.channel));
      
      node->system.state = ADMXRC_DMASTATE_COMPLETED;
      node->status.status = OS_MAKE_STATUS(ADMXRC_CLOSED);
      abortDMAOnHardware(pDevCtx, i);
      pDevCtx->dma.channel[i].active = NULL;
      pDevCtx->dma.channel[i].count--;
      nextnode = getNextDMA(pDevCtx, i);
      os_release_spinlock(&pDevCtx->dma.lock, f);
      
      os_dbgprint(6, ("ADMXRC_CleanupDMA: calling DMA callback for node 0x%p, reason = COMPLETED\n", (void*) node));
      node->status.reason = ADMXRC_DMACALLBACK_REASON_COMPLETED;
      node->client.callback(pDevCtx, node);

      if (nextnode != NULL) {
	os_dbgprint(6, ("ADMXRC_CleanupDMA: calling DMA callback for node 0x%p, reason = SUBMIT\n", (void*) nextnode));
        os_sg_init(pDevCtx->dma.channel[i].sgMap,
                   nextnode->client.direction == ADMXRC_LOCAL2PCI);
	nextnode->status.reason = ADMXRC_DMACALLBACK_REASON_SUBMIT;
	nextnode->client.callback(pDevCtx, nextnode);
      }
    } else {
      os_release_spinlock(&pDevCtx->dma.lock, f);
    }
  }
  
  /* Cleanup any DMA transfers that are currently in the WAITCOLLECT state */
  for (i = 0; i < pDevCtx->info.cardInfo.NumDMAChan; i++) {
    os_acquire_spinlock(&pDevCtx->dma.lock, &f);
    node = pDevCtx->dma.channel[i].waitcollect;
    if (node != NULL) {
      os_dbgprint(1, ("ADMXRC_CleanupDMA: Cleaning up DMANODE 0x%p in WAITCOLLECT state on channel %ld\n",
		      (void*) node, (unsigned long) node->system.channel));
      pDevCtx->dma.channel[i].cleanupPending = TRUE;
    }
    os_release_spinlock(&pDevCtx->dma.lock, f);      
  }

  os_dbgprint(2, ("ADMXRC_CleanupDMA: done\n"));
}
