/*
** ADM-XRC Series Driver
**
** (C) Copyright 2001-2004 Alpha Data
**
** dmadesc.c - Routines to create, destroy and manage DMA buffer descriptors.
**
** ----------------------------------------------------------------------------
*/

#include "platform.h"
#include "device.h"
#include "ioctl.h"
#include "dmadesc.h"

static void
destroyDMATarget(
  ADMXRC_DMATGT* tgt)
{
  os_destroy_usbd(tgt->usbd);
  os_kernel_free(tgt);
}

OS_STATUS
ADMXRC_CreateDMATarget(
  DeviceContext*  devCtx,
  void*           buffer,
  unsigned long   length,
  ADMXRC_DMATGT** dmatgt)
{
  ADMXRC_DMATGT* t;
  ADMXRC_STATUS  status;

  if (length == 0) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  t = (ADMXRC_DMATGT*) os_kernel_malloc(sizeof(ADMXRC_DMATGT));
  if (t == NULL) {
    os_dbgprint(0, ("ADMXRC_CreateDMATarget: failed to allocate an ADMXRC_DMATGT\n"));
    return OS_MAKE_STATUS(ADMXRC_NO_MEMORY);
  }
  status = os_create_usbd(buffer, length, &t->usbd);
  if (!OS_SUCCESS(status)) {
    switch (status) {
    case OS_STATUS_NOMEMORY:
      status = OS_MAKE_STATUS(ADMXRC_NO_MEMORY);
      break;
    case OS_STATUS_UNSUCCESSFUL:
      status = OS_MAKE_STATUS(ADMXRC_FAILED);
      break;
    default:
      break;
    }
    os_dbgprint(1, ("ADMXRC_CreateDMATarget: failed to allocate USBD\n"));
    os_kernel_free(t);
    return status;
  }
  t->in_use = 1;
  t->buffer = buffer;
  t->length = length;
	
  *dmatgt = t;
  
  return OS_STATUS_SUCCESS;
}

/*
** Decrements a DMA target's usage counter and destroys it if the
** usage counter drops to 0.
*/
void
ADMXRC_ReleaseDMATarget(
  DeviceContext* devCtx,
  ADMXRC_DMATGT* dmatgt)
{
  OS_SPINLOCK_FLAGS f;
  os_boolean        free_it;
  unsigned long     in_use;

  os_acquire_spinlock(&devCtx->dmadesc.lock, &f);
  in_use = dmatgt->in_use - 1;
  free_it = (in_use == 0);
  dmatgt->in_use = in_use;
  os_release_spinlock(&devCtx->dmadesc.lock, f);

  if (free_it) {
    os_dbgprint(3, ("ADMXRC_ReleaseDMATarget: destroying DMA target 0x%p\n", (void*) dmatgt));
    destroyDMATarget(dmatgt);
  }
}

/*
** Looks up a DMA target given its handle. If the DMA target exists,
** its usage counter is incremented.
*/
ADMXRC_DMATGT*
ADMXRC_GetDMATarget(
  DeviceContext* devCtx,
  unsigned long  handle)
{
  ADMXRC_DMATGT*    dmatgt;
  OS_SPINLOCK_FLAGS f;
  int               index = handle - 1;

  if (handle == 0 || handle > ADMXRC_MAX_DMADESC) {
    return NULL;
  }

  index = handle - 1;

  os_acquire_spinlock(&devCtx->dmadesc.lock, &f);
  dmatgt = devCtx->dmadesc.table[index];
  if (dmatgt != NULL) {
    dmatgt->in_use++;
  }
  os_release_spinlock(&devCtx->dmadesc.lock, f);

  return dmatgt;
}

void
ADMXRC_CleanupDMATargets(
  DeviceContext* devCtx)
{
  ADMXRC_DMATGT*    dmatgt;
  int               i;
  OS_SPINLOCK_FLAGS f;

  os_dbgprint(2, ("ADMXRC_CleanupDMATargets: entered\n"));

  for (i = 0; i < ADMXRC_MAX_DMADESC; i++) {
    os_acquire_spinlock(&devCtx->dmadesc.lock, &f);
    dmatgt = devCtx->dmadesc.table[i];
    if (dmatgt != NULL) {
      os_dbgprint(3, ("ADMXRC_CleanupDMATargets: cleaning up DMA descriptor %ld (dmatgt = 0x%p) busy = %ld\n",
		      (unsigned long) i + 1, (void*) dmatgt, (unsigned long) dmatgt->in_use));
      devCtx->dmadesc.table[i] = NULL;
      devCtx->dmadesc.freestack[devCtx->dmadesc.top++] = i;
      os_release_spinlock(&devCtx->dmadesc.lock, f);
      
      /* Now safe to release it */
      ADMXRC_ReleaseDMATarget(devCtx, dmatgt);
    } else {
      os_release_spinlock(&devCtx->dmadesc.lock, f);
    }
  }
}

/*
** Handle ADMXRC_IOCTL_SETUPDMA
*/
OS_STATUS
ADMXRC_SetupDMA(
  DeviceContext* devCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_SETUPDMA* iptr;
  ADMXRC_IOCTLS_SETUPDMA* optr;
  OS_STATUS               status;
  ADMXRC_DMATGT*          dmatgt = NULL;
  OS_SPINLOCK_FLAGS       f;
  int                     index;

  iptr = (ADMXRC_IOCTLS_SETUPDMA*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  optr = (ADMXRC_IOCTLS_SETUPDMA*) req->param.outBuf;
  if (optr == NULL || req->param.outSize != sizeof(optr->out)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  
  status = ADMXRC_CreateDMATarget(devCtx, iptr->in.buffer, iptr->in.size, &dmatgt);
  if (!OS_SUCCESS(status)) {
    goto failed;
  }

  os_acquire_spinlock(&devCtx->dmadesc.lock, &f);
  if (devCtx->dmadesc.top == 0) {
    os_release_spinlock(&devCtx->dmadesc.lock, f);
    status = OS_MAKE_STATUS(ADMXRC_NO_MEMORY);
    goto failed;
  } else {
    index = devCtx->dmadesc.freestack[--(devCtx->dmadesc.top)];
    devCtx->dmadesc.table[index] = dmatgt;
    os_release_spinlock(&devCtx->dmadesc.lock, f);

    optr->out.dmadesc = (unsigned long) (index + 1);
    req->status.length = sizeof(optr->out);
  }

  return status;

failed:
  if (dmatgt != NULL) {
    destroyDMATarget(dmatgt);
  }
  
  return status;
}

OS_STATUS
ADMXRC_UnsetupDMA(
  DeviceContext* devCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_UNSETUPDMA* iptr;
  unsigned long             handle;
  ADMXRC_DMATGT*            dmatgt;
  OS_SPINLOCK_FLAGS         f;
  int                       index;

  iptr = (ADMXRC_IOCTLS_UNSETUPDMA*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  
  handle = iptr->in.dmadesc;
  if (handle == 0 || handle > ADMXRC_MAX_DMADESC) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_DMADESC);
  }

  index = handle - 1;

  os_acquire_spinlock(&devCtx->dmadesc.lock, &f);
  dmatgt = devCtx->dmadesc.table[index];
  if (dmatgt == NULL) {
    os_release_spinlock(&devCtx->dmadesc.lock, f);

    return OS_MAKE_STATUS(ADMXRC_INVALID_DMADESC);
  } else {
    devCtx->dmadesc.table[index] = NULL;
    devCtx->dmadesc.freestack[devCtx->dmadesc.top++] = index;
    os_release_spinlock(&devCtx->dmadesc.lock, f);

    ADMXRC_ReleaseDMATarget(devCtx, dmatgt);

    return OS_STATUS_SUCCESS;
  }
}

/*
** x86 PC specific at the moment - how do we convert a PFN to
** a PCI address in a platform-independent manner?
*/
OS_STATUS
ADMXRC_MapDirectMaster(
  DeviceContext* devCtx,
  OS_IO_REQUEST* req)
{
  unsigned long                  dmadesc;
  ADMXRC_DMATGT*                 dmatgt = NULL;
  ADMXRC_IOCTLS_MAPDIRECTMASTER* iptr;
  ADMXRC_IOCTLS_MAPDIRECTMASTER* optr;
  ADMXRC_STATUS                  status = ADMXRC2_SUCCESS;
  unsigned long*                 pagebuff;
  unsigned long                  length, offset;
  unsigned long                  maxpage;
  unsigned long                  pf_offs;
  unsigned long                  chunk;
  unsigned long                  pagespan, bytespan;
  os_uint32                      busAddr;
  os_boolean                     bSuccess;

  iptr = (ADMXRC_IOCTLS_MAPDIRECTMASTER*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)){
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  optr = (ADMXRC_IOCTLS_MAPDIRECTMASTER*) req->param.outBuf;
  if (optr == NULL || req->param.outSize != sizeof(optr->out)){
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  dmadesc = iptr->in.dmadesc;
  offset = iptr->in.offset;
  length = iptr->in.length;
  maxpage = iptr->in.maxpage;

  pagebuff = (unsigned long*) req->param.massBuf;
  if (maxpage) {
    if (pagebuff == NULL || req->param.massSize != (sizeof(unsigned long) * maxpage)) {
      status = OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
      goto done;
    }
  }

  dmatgt = ADMXRC_GetDMATarget(devCtx, dmadesc);
  if (dmatgt == NULL) {
    status = OS_MAKE_STATUS(ADMXRC_INVALID_DMADESC);
    goto done;
  }

  if (offset >= dmatgt->length || (offset + length) > dmatgt->length) {
    status = OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
    goto done;
  }

  pagespan = 0;
  bytespan = 0;
  while (length && maxpage) {
    bSuccess = os_get_usbd_bus_addr(dmatgt->usbd, offset, &busAddr, &pf_offs);
    if (!bSuccess) {
      status = OS_MAKE_STATUS(ADMXRC_FAILED);
    }
    chunk = os_page_size - pf_offs;
    if (chunk > length) {
      chunk = length;
    }
    *pagebuff++ = (unsigned long) busAddr;
    length -= chunk;
    maxpage--;
    offset += chunk;
    bytespan += chunk;
    pagespan++;
  }

  optr->out.pagelen  = os_page_size;
  optr->out.pagebits = os_page_shift;
  optr->out.pagespan = pagespan;
  optr->out.bytespan = bytespan;
  optr->out.initoffs = os_get_usbd_page_offset(dmatgt->usbd);

  req->status.length = sizeof(optr->out);

  if (os_dbglevel(9)) {
    unsigned long  i;
    unsigned long* p = (unsigned long*) req->param.massBuf;

    os_dbgprint(0, ("ADMXRC_MapDirectMaster: pf_offs = 0x%ld pagespan = 0x%ld bytespan = 0x%ld\n",
		    (unsigned long) pf_offs, (unsigned long) pagespan, (unsigned long) bytespan));
    for (i = 0; i < pagespan; i++) {
      os_dbgprint(0, ("ADMXRC_MapDirectMaster: %4ld bus_addr = 0x%8.8lx\n",
		      (unsigned long) i, (unsigned long) p[i]));
    }
  }

done:
  if (dmatgt != NULL) {
    ADMXRC_ReleaseDMATarget(devCtx, dmatgt);
  }

  return status;
}

OS_STATUS
ADMXRC_SyncDirectMaster(
  DeviceContext* devCtx,
  OS_IO_REQUEST* req)
{
  unsigned long                   dmadesc;
  ADMXRC_DMATGT*                  dmatgt = NULL;
  ADMXRC_IOCTLS_SYNCDIRECTMASTER* iptr;
  int                             syncmode;
  ADMXRC_STATUS                   status = OS_STATUS_SUCCESS;
  unsigned long                   offset, length;

  iptr = (ADMXRC_IOCTLS_SYNCDIRECTMASTER*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  
  dmadesc = iptr->in.dmadesc;
  syncmode = iptr->in.syncmode;
  length = iptr->in.length;
  offset = iptr->in.offset;

  dmatgt = ADMXRC_GetDMATarget(devCtx, dmadesc);
  if (dmatgt == NULL) {
    status = OS_MAKE_STATUS(ADMXRC_INVALID_DMADESC);
    goto done;
  }

  if (offset >= dmatgt->length || (offset + length) > dmatgt->length) {
    status = OS_MAKE_STATUS(ADMXRC_INVALID_DMADESC);
    goto done;
  }

  switch (syncmode) {
  case ADMXRC_SYNC_CPUTOFPGA:
    os_flush_io_buffer(dmatgt->usbd, offset, length, FALSE);
    break;
  case ADMXRC_SYNC_FPGATOCPU:
    os_flush_io_buffer(dmatgt->usbd, offset, length, TRUE);
    break;
  default:
    status = OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
    goto done;
  }
  
done:
  if (dmatgt != NULL) {
    ADMXRC_ReleaseDMATarget(devCtx, dmatgt);
  }

  return status;
}
