/*
** ADM-XRC Series Driver
**
** (C) Copyright 2001-2004 Alpha Data
**
** flash.c - Flash memory access routines
**
** ----------------------------------------------------------------------------
*/

#include "device.h"
#include "flash.h"

static MethodResult
writeBlock(
  DeviceContext*  pDevCtx,
  os_uint32       start,
  os_uint32       length,
  const os_uint8* data)
{
  MethodResult	result = MethodResult_Success;
  os_uint32     venDev;
  os_uint32     i, limit;
  unsigned long timeout_us, remaining_us;
  os_boolean    bTimeout;
  os_uint16     val16;
  
  os_dbgprint(3, ("writeBlock: entered, start=0x%lx length=0x%lx data=0x%p\n",
		  (unsigned long) start, (unsigned long) length, data));

  if (!pDevCtx->info.flash.present) {
    return MethodResult_NotPresent;
  }

  limit = start + length;

  venDev = pDevCtx->info.flash.identity.vendor;
  venDev = (venDev << 16) | pDevCtx->info.flash.identity.device;
  switch (venDev) {
  case 0x890099: /* Micron 28F008B */
    {
      for (i = start; i < limit; i++) {
	/* Send WRITE-BYTE command */
	val16 = 0x40U;
	pDevCtx->methods.flash(pDevCtx, TRUE, 0x0, &val16);
	val16 = *data++;
	pDevCtx->methods.flash(pDevCtx, TRUE, i, &val16);

	/* Wait for completion */
	timeout_us = 10;
	bTimeout = FALSE;
	while (timeout_us) {
	  os_delay_busy_us(2);
	  /* Use data polling method */
	  pDevCtx->methods.flash(pDevCtx, FALSE, i, &val16);
	  if (val16 & 0x80U) {
	    break;
	  }
	  if (timeout_us < 2) {
	    bTimeout = TRUE;
	    break;
	  }
	  timeout_us -= 2;
	}
	if (bTimeout) {
	  result = MethodResult_Timeout;
	} else {
	  /* Status-check */
	  if (val16 & 0x0cU) {
	    os_dbgprint(0, ("*** writeBlock: Error detected after WRITE-BYTE: SR[7:0] = 0x%lx\n",
			    (unsigned long) val16));
	    result = MethodResult_VerifyFailed;
	  }
	}
	
	/* Return to READ-ARRAY mode */
	val16 = 0xffU;
	pDevCtx->methods.flash(pDevCtx, TRUE, 0x0, &val16);
      }
      break;
    }
    
  case 0x890016: /* Intel 28320J3A */
  case 0x890017: /* Intel 28640J3A */
  case 0x890018: /* Intel 28128J3A */
    {
      timeout_us = 2000;
      i = start;
      while (length) {
	os_uint32 j, n;
	
	/* Calculate number of bytes for this command */
	n = 32 - (i & 0x1fU);
	if (n > length) {
	  n = length;
	}
	
	os_dbgprint(3, ("writeBlock: doing WRITE-TO-BUFFER, start=0x%lx length=0x%lx i=0x%lx n=0x%lx data=0x%p\n",
			(unsigned long) start, (unsigned long) length, (unsigned long) i, (unsigned long) n, data));

	/* Send WRITE-TO-BUFFER command */
	val16 = 0xe8U;
	pDevCtx->methods.flash(pDevCtx, TRUE, start, &val16);
	
	/* Send the data to the write buffer */
	val16 = (os_uint8) (n - 1);
	pDevCtx->methods.flash(pDevCtx, TRUE, start, &val16);
	for (j = 0; j < n; j++) {
	  val16 = *data++;
	  pDevCtx->methods.flash(pDevCtx, TRUE, i + j, &val16);
	}
	
	/* Confirm WRITE-TO-BUFFER command */
	val16 = 0xd0U;
	pDevCtx->methods.flash(pDevCtx, TRUE, start, &val16);
	
	/* Wait for completion of WRITE-TO-BUFFER command */
	remaining_us = timeout_us;
	bTimeout = FALSE;
	while (1) {
	  /* Send Read-STATUS command */
	  val16 = 0x70U;
	  pDevCtx->methods.flash(pDevCtx, TRUE, start, &val16);
	  pDevCtx->methods.flash(pDevCtx, FALSE, start, &val16);
	  if (val16 & 0x80U) {
	    break;
	  }

	  os_delay_busy_us(10);
	  if (remaining_us < 10) {
	    os_dbgprint(0, ("*** writeBlock: timeout waiting for buffered write to complete\n"));
	    bTimeout = TRUE;
	    break;
	  }
	  remaining_us -= 10;
	}
	
	if (bTimeout) {
	  /* Send CLEAR-STATUS command */
	  val16 = 0x50U;
	  pDevCtx->methods.flash(pDevCtx, TRUE, start, &val16);
	  /* Send READ-ARRAY command */
	  val16 = 0xffU;
	  pDevCtx->methods.flash(pDevCtx, TRUE, start, &val16);
	  result = MethodResult_Timeout;
	  /* Return early */
	  break;
	}
	
	/* Perform status-check */
	if (val16 & 0x0cU) {
	  os_dbgprint(0, ("*** writeBlock: Error detected after WRITE-BYTE: SR[7:0] = 0x%lx\n",
			  (unsigned long) val16));
	  result = MethodResult_VerifyFailed;
	  /* Send CLEAR-STATUS command */
	  val16 = 0x50U;
	  pDevCtx->methods.flash(pDevCtx, TRUE, start, &val16);
	  /* Send READ-ARRAY command */
	  val16 = 0xffU;
	  pDevCtx->methods.flash(pDevCtx, TRUE, start, &val16);
	  /* Return early */
	  break;
	}
	
	/* Send READ-ARRAY command */
	val16 = 0xffU;
	pDevCtx->methods.flash(pDevCtx, TRUE, 0x0, &val16);

	i += n;
	length -= n;
      }
      break;
    }
    
  case 0x898803: /* Intel 28F256K3 */
    {
      /* Send UNLOCK-BLOCK command */
      val16 = 0x60U;
      pDevCtx->methods.flash(pDevCtx, TRUE, start, &val16);
      val16 = 0xd0U;
      pDevCtx->methods.flash(pDevCtx, TRUE, start, &val16);
      
      timeout_us = 2000;
      i = start;
      while (length) {
	os_uint32 j, n;
	
	/* Calculate number of bytes for this command */
	n = 64 - (i & 0x3eU);
	if (n > length) {
	  n = length;
	}
	
	os_dbgprint(3, ("writeBlock: doing WRITE-TO-BUFFER, start=0x%lx length=0x%lx i=0x%lx n=0x%lx data=0x%p\n",
			(unsigned long) start, (unsigned long) length, (unsigned long) i, (unsigned long) n, (void*) data));
	
	/* Send WRITE-TO-BUFFER command */
	val16 = 0xe8U;
	pDevCtx->methods.flash(pDevCtx, TRUE, start, &val16);
	
	/* Send the data to the write buffer */
	val16 = (os_uint8) ((n >> 1) - 1);
	pDevCtx->methods.flash(pDevCtx, TRUE, start, &val16);
	for (j = 0; j < n;  j += 2) {
	  val16 = *data++;
	  val16 |= (*data++ << 8);
	  pDevCtx->methods.flash(pDevCtx, TRUE, i + j, &val16);
	}
	
	/* Confirm WRITE-TO-BUFFER command */
	val16 = 0xd0U;
	pDevCtx->methods.flash(pDevCtx, TRUE, start, &val16);
	
	/* Wait for completion of WRITE-TO-BUFFER command */
	remaining_us = timeout_us;
	bTimeout = FALSE;
	while (1) {
	  /* Send Read-STATUS command */
	  val16 = 0x70U;
	  pDevCtx->methods.flash(pDevCtx, TRUE, start, &val16);
	  pDevCtx->methods.flash(pDevCtx, FALSE, start, &val16);
	  if (val16 & 0x80U) {
	    break;
	  }
	  
	  os_delay_busy_us(10);
	  if (remaining_us < 10) {
	    os_dbgprint(0, ("*** writeBlock: timeout waiting for buffered write to complete\n"));
	    bTimeout = TRUE;
	    break;
	  }
	  remaining_us -= 10;
	}
	
	if (bTimeout) {
	  /* Send CLEAR-STATUS command */
	  val16 = 0x50U;
	  pDevCtx->methods.flash(pDevCtx, TRUE, start, &val16);
	  /* Send READ-ARRAY command */
	  val16 = 0xffU;
	  pDevCtx->methods.flash(pDevCtx, TRUE, start, &val16);
	  result = MethodResult_Timeout;
	  /* Return early */
	  break;
	}
	
	/* Perform status-check */
	if (val16 & 0x0cU) {
	  os_dbgprint(0, ("*** writeBlock: Error detected after WRITE-TO-BUFFER: SR[7:0] = 0x%lx\n",
			  (unsigned long) val16));
	  result = MethodResult_VerifyFailed;
	  /* Send CLEAR-STATUS command */
	  val16 = 0x50U;
	  pDevCtx->methods.flash(pDevCtx, TRUE, start, &val16);
	  /* Send READ-ARRAY command */
	  val16 = 0xffU;
	  pDevCtx->methods.flash(pDevCtx, TRUE, start, &val16);
	  /* Return early */
	  break;
	}
	
	/* Check for program error */
	if (val16 & 0x10U) {
	  os_dbgprint(0, ("*** writeBlock: Error detected after WRITE-TO-BUFFER: SR[7:0] = 0x%lx\n",
			  (unsigned long) val16));
	  result = MethodResult_VerifyFailed;
	  /* Send CLEAR-STATUS command */
	  val16 = 0x50U;
	  pDevCtx->methods.flash(pDevCtx, TRUE, start, &val16);
	  /* Send READ-ARRAY command */
	  val16 = 0xffU;
	  pDevCtx->methods.flash(pDevCtx, TRUE, start, &val16);
	  /* Return early */
	  break;
	}
	
	i += n;
	length -= n;
      }
      
      /* Send LOCK-BLOCK command */
      val16 = 0x60U;
      pDevCtx->methods.flash(pDevCtx, TRUE, start, &val16);
      val16 = 0x01U;
      pDevCtx->methods.flash(pDevCtx, TRUE, start, &val16);
      
      /* Send READ-ARRAY command */
      val16 = 0xffU;
      pDevCtx->methods.flash(pDevCtx, TRUE, 0x0, &val16);
      
      break;
    }

  case 0xbf00d9: /* SST 39VF016Q */
    {
      os_uint16 byte;

      for (i = start; i < limit; i++) {
	/* Send WRITE-BYTE command */
	val16 = 0xaaU;
	pDevCtx->methods.flash(pDevCtx, TRUE, 0x5555, &val16);
	val16 = 0x55U;
	pDevCtx->methods.flash(pDevCtx, TRUE, 0x2aaa, &val16);
	val16 = 0xa0U;
	pDevCtx->methods.flash(pDevCtx, TRUE, 0x5555, &val16);
	byte = *data++;
	val16 = byte;
	pDevCtx->methods.flash(pDevCtx, TRUE, i, &val16);
	
	/* Wait for completion */
	timeout_us = 25;
	bTimeout = FALSE;
	while (timeout_us) {
	  os_delay_busy_us(5);
	  /* Use data polling method */
	  pDevCtx->methods.flash(pDevCtx, FALSE, i, &val16);
	  if ((val16 & 0x80U) == (byte & 0x80U)) {
	    break;
	  }
	  if (timeout_us < 5) {
	    bTimeout = TRUE;
	    break;
	  }
	  timeout_us -= 5;
	}
	if (bTimeout) {
	  result = MethodResult_Timeout;
	  break;
	}
	
	/* Return to READ-ARRAY mode */
	val16 = 0xf0U;
	pDevCtx->methods.flash(pDevCtx, TRUE, 0x0, &val16);
      }
      break;
    }
    
  default:
    os_assert(FALSE);
    result = MethodResult_NotSupported;
    break;
  }
  
  return result;
}

/*
** -----------------------------------------------------------------
** Exported routines
** -----------------------------------------------------------------
*/

/*
** Capabilities of supported Flash devices
**
** Device     Width  Ven:Dev  Size    Block-erase   Bootblocks
** ==================================================================
**
** 28F320     8      89:16    4MB     128k, 5s      -
** 28F640     8      89:17    8MB     128k, 5s      -
** 28F128     8      89:18    16MB    128k, 5s      -
** 28F0008B   8      89:99    4MB     128k, 14s     16k, 8k, 8k, 96k
** 28F256K3   16     89:8803  32MB    128k, 4s      -
** 39VF016Q   8      bf:d9    2MB     64k, 25ms     -
*/
void
identifyFlash(
  DeviceContext* pDevCtx)
{
  os_boolean   bFound = FALSE;
  os_boolean   res;
  os_uint16    val16;
  os_uint16    vendor = 0xffU, device = 0xffU;
  os_uint32    size = 0;
  os_uint32    eraseBlock = 0;
  os_uint8     width = 0;
  unsigned int numBootBlock = 0;
  os_uint32    bootBlock[4];
  char*        name = "unknown";

  /* Issue a reset to the Flash */
  pDevCtx->methods.flashReset(pDevCtx);

  /* Go to read array mode */
  val16 = 0xffU;
  res = pDevCtx->methods.flash(pDevCtx, TRUE, 0, &val16);
  if (res != MethodResult_Success) {
    return;
  }
  val16 = 0xf0U;
  res = pDevCtx->methods.flash(pDevCtx, TRUE, 0, &val16);
  
  if (!bFound) {
    /*
    ** Try Intel.
    **
    ** Send "Read Identifier Codes" command.
    */
    val16 = 0x90U;
    res = pDevCtx->methods.flash(pDevCtx, TRUE, 0, &val16);
    res = pDevCtx->methods.flash(pDevCtx, FALSE, 0, &val16);
    if (val16 == 0x89) {
      /* Identified Intel vendor - now get device */
      vendor = val16;
      res = pDevCtx->methods.flash(pDevCtx, FALSE, 0x2, &val16);
      device = val16;
      
      /* Go to read array mode */
      val16 = 0xffU;
      res = pDevCtx->methods.flash(pDevCtx, TRUE, 0, &val16);
      
      switch (device) {
      case 0x16:
	{
	  /* Intel 28F320J3A, 16Mbit */
	  bFound = TRUE;
	  eraseBlock = 128*1024;
	  size = 4*1024*1024;
	  name = "Intel 28F320";
	  width = 8;
	  break;
	}

      case 0x17:
	{
	  /* Intel 28F640J3A, 64Mbit */
	  bFound = TRUE;
	  eraseBlock = 128*1024;
	  size = 8*1024*1024;
	  name = "Intel 28F640";
	  width = 8;
	  break;
	}

      case 0x18:
	{
	  /* Intel 28F128J3A, 128Mbit */
	  bFound = TRUE;
	  eraseBlock = 128*1024;
	  size = 16*1024*1024;
	  name = "Intel 28F128";
	  width = 8;
	  break;
	}

      case 0x8803:
	{
	  /* Intel 28F256K3, 256Mbit */
	  bFound = TRUE;
	  eraseBlock = 128*1024;
	  size = 32*1024*1024;
	  name = "Intel 28F256K3";
	  width = 16;
	  break;
	}

      default:
	break;
      }
    }
  }

  if (!bFound) {
    /*
    ** Try Micron
    **
    ** Send "Read Identifier Codes" command
    */
    val16 = 0x90U;
    res = pDevCtx->methods.flash(pDevCtx, TRUE, 0, &val16);
    res = pDevCtx->methods.flash(pDevCtx, FALSE, 0, &val16);
    if (val16 == 0x89) {
      /* Vendor is Micron - now get device */
      vendor = 0x89;
      res = pDevCtx->methods.flash(pDevCtx, FALSE, 1, &val16);
      device = val16;
      
      /* Go to read array mode */
      val16 = 0xffU;
      res = pDevCtx->methods.flash(pDevCtx, TRUE, 0, &val16);
      
      switch (device) {
      case 0x99:
	{
	  /* Micron 28F008B, 8Mbit */
	  bFound = TRUE;
	  eraseBlock = 128*1024;
	  numBootBlock = 4;
	  bootBlock[0] = 16*1024;
	  bootBlock[1] = 8*1024;
	  bootBlock[2] = 8*1024;
	  bootBlock[3] = 96*1024;
	  size = 1*1024*1024;
	  name = "Micron 28F008B";
	  width = 8;
	  break;
	}

      default:
	break;
      }
    }
  }
  
  if (!bFound) {
    /*
    ** Try SST/ATMEL - these devices conform to an older JEDEC standard.
    **
    ** Enter "Read ID" mode.
    */
    val16 = 0xaaU;
    res = pDevCtx->methods.flash(pDevCtx, TRUE, 0x5555, &val16);
    val16 = 0x55U;
    res = pDevCtx->methods.flash(pDevCtx, TRUE, 0xaaaa, &val16);
    val16 = 0x90U;
    res = pDevCtx->methods.flash(pDevCtx, TRUE, 0x5555, &val16);
    res = pDevCtx->methods.flash(pDevCtx, FALSE, 0, &val16);
    if (val16 == 0xbfU) {
      vendor = val16;
      res = pDevCtx->methods.flash(pDevCtx, FALSE, 1, &val16);
      device = val16;

      /* Exit read identity mode */
      val16 = 0xf0U;
      res = pDevCtx->methods.flash(pDevCtx, TRUE, 0, &val16);
      
      switch (device) {
      case 0xd9U:
	{
	  /* SST 39VF016Q, 16Mbit */
	  bFound = TRUE;
	  eraseBlock = 64*1024;
	  size = 2*1024*1024;
	  name = "SST 39VF016Q";
	  width = 8;
	  break;
	}

      default:
	break;
      }
    }
    
    /* Go to read array mode */
    val16 = 0xffU;
    res = pDevCtx->methods.flash(pDevCtx, TRUE, 0, &val16);
    val16 = 0xf0U;
    res = pDevCtx->methods.flash(pDevCtx, TRUE, 0, &val16);
  }
  
  if (bFound) {
    unsigned int	i;
    
    os_dbgprint(1, ("identifyFlash: identified Flash as %s, eraseBlock = 0x%lx\n",
		    name, (unsigned long) eraseBlock));
    pDevCtx->info.flash.present = TRUE;
    pDevCtx->info.flash.identity.device = device;
    pDevCtx->info.flash.identity.vendor = vendor;
    pDevCtx->info.flash.identity.name = name;
    pDevCtx->info.flash.size = size;
    pDevCtx->info.flash.width = width;
    pDevCtx->info.flash.layout.numBootBlock = numBootBlock;
    for (i = 0; i < numBootBlock; i++) {
      pDevCtx->info.flash.layout.bootBlock[i] = bootBlock[i];
    }
    pDevCtx->info.flash.layout.eraseBlock = eraseBlock;
    
    pDevCtx->info.flash.targetArea.start = 0;
    pDevCtx->info.flash.targetArea.length = 0;
    switch (pDevCtx->devDesc.id.pci.subDId) {
    case PCI_SUBDEVICE_ADATA_ADMXRC:
    case PCI_SUBDEVICE_ADATA_ADMXRC2L:
    case PCI_SUBDEVICE_ADATA_ADMXRC2:
    case PCI_SUBDEVICE_ADATA_ADPWRC2:
    case PCI_SUBDEVICE_ADATA_ADPDRC2:
      pDevCtx->info.flash.targetArea.start = 0;
      pDevCtx->info.flash.targetArea.length = size;
      break;
      
    case PCI_SUBDEVICE_ADATA_ADMXRC2PL_PCIX:
    case PCI_SUBDEVICE_ADATA_ADMXRC2PL_PCI:
    case PCI_SUBDEVICE_ADATA_ADMXRC2P_PCIX:
    case PCI_SUBDEVICE_ADATA_ADMXRC2P_PCI:
      pDevCtx->info.flash.targetArea.start = 0x800000U;
      pDevCtx->info.flash.targetArea.length = size - 0x800000U;
      break;

    default:
      os_assert(FALSE);
      break;
    }
  } else {
    os_dbgprint(0, ("*** identifyFlash: failed to identify Flash\n"));
  }
}

MethodResult
identifyFlashBlock(
  DeviceContext* pDevCtx,
  os_uint32      location,
  os_uint32*     blockStart,
  os_uint32*     blockSize,
  os_boolean*    isBootBlock)
{
  os_uint32    eraseBlock = pDevCtx->info.flash.layout.eraseBlock;
  os_uint32    start, end, size = 0;
  os_boolean   bBoot;
  unsigned int i;
  
  if (location > pDevCtx->info.flash.size) {
    return MethodResult_InvalidLocation;
  }
  
  /* Identify which block the location belongs to */
  start = 0;
  bBoot = FALSE;
  for (i = 0; i < pDevCtx->info.flash.layout.numBootBlock; i++) {
    size = pDevCtx->info.flash.layout.bootBlock[i];
    end = start + size;
    if (location >= start && location < end) {
      bBoot = TRUE;
      break;
    }
    start += size;
  }
  
  if (!bBoot) {
    /* It's not a boot block, must be a normal block */
    start = location & ~(eraseBlock - 1);
    size = eraseBlock;
  }
  
  *blockStart = start;
  *blockSize = size;
  *isBootBlock = bBoot;
  
  return MethodResult_Success;
}

MethodResult
eraseFlash(
  DeviceContext* pDevCtx,
  os_uint32      start,
  os_uint32      length)
{
  MethodResult  result = MethodResult_Success;
  os_uint32     eraseBlock;
  os_uint32     blockStart, blockEnd;
  os_uint32     blockSize;
  os_boolean    isBootBlock;
  os_uint32     chunk;
  os_uint32     venDev;
  os_uint16     val16;
  unsigned long	timeout_ms, remaining_ms;
  os_boolean    bTimeout, bInterrupt, bStatusError;
  unsigned long interval;

  os_dbgprint(3, ("eraseFlash: entered, start=0x%lx length=0x%lx\n",
		  (unsigned long) start, (unsigned long) length));

  if (!pDevCtx->info.flash.present) {
    return MethodResult_NotPresent;
  }
  
  venDev = pDevCtx->info.flash.identity.vendor;
  venDev = (venDev << 16) | pDevCtx->info.flash.identity.device;
  eraseBlock = pDevCtx->info.flash.layout.eraseBlock;
  
  switch (venDev) {
  case 0x890099: /* Micron 28F008B */
    timeout_ms = 14000;
    break;
    
  case 0x890016: /* Intel 28320J3A */
  case 0x890017: /* Intel 28640J3A */
  case 0x890018: /* Intel 28128J3A */
    timeout_ms = 5000;
    break;
    
  case 0x898803: /* Intel 28F256K3 */
    timeout_ms = 4000;
    break;

  case 0xbf00d9: /* SST 39VF016Q */
    timeout_ms = 25;
    break;
    
  default:
    /* Unrecognised device */
    return MethodResult_NotSupported;
  }
  
  while (length) {
    /* Identify which block the range belongs to */
    identifyFlashBlock(pDevCtx, start, &blockStart, &blockSize, &isBootBlock);
    blockEnd = blockStart + eraseBlock;
    chunk = blockEnd - start;
    if (chunk > length) {
      chunk = length;
    }
    
    os_dbgprint(6, ("eraseFlash: start=0x%lx length=0x%lx blockStart=0x%lx blockEnd=0x%lx blockSize=0x%lx chunk=0x%lx\n",
		    (unsigned long) start, (unsigned long) length, (unsigned long) blockStart,
		    (unsigned long) blockEnd, (unsigned long) blockSize, (unsigned long) chunk));
    
    /* Erase the block */
    switch (venDev) {
    case 0x890099: /* Micron 28F008B */
    case 0x890016: /* Intel 28320J3A */
    case 0x890017: /* Intel 28640J3A */
    case 0x890018: /* Intel 28128J3A */
      {
	/* Send ERASE-BLOCK command */
	val16 = 0x20U;
	pDevCtx->methods.flash(pDevCtx, TRUE, 0, &val16);
	val16 = 0xd0U;
	pDevCtx->methods.flash(pDevCtx, TRUE, blockStart, &val16);
	
	/* Wait for completion */
	remaining_ms = timeout_ms;
	bTimeout = FALSE;
	bStatusError = FALSE;
	bInterrupt = FALSE;
	while (1) {
	  pDevCtx->methods.flash(pDevCtx, FALSE, 0, &val16);
	  if (val16 & 0x80U) {
	    break;
	  }
	  interval = 5;
	  bInterrupt = os_delay_thread(FALSE, &interval);
	  if (bInterrupt) {
	    os_dbgprint(1, ("*** eraseFlash: delay interrupted\n"));
	    break;
	  }
	  if (remaining_ms < 5) {
	    bTimeout = TRUE;
	    break;
	  }
	  remaining_ms -= 5;
	}
	
	if (bInterrupt) {
	  result = MethodResult_GeneralFailure;
	}

	if (bTimeout) {
	  result = MethodResult_Timeout;
	} else {
	  /* Status-check */
	  if (val16 & 0x1cU) {
	    /* Vpp Error */
	    os_dbgprint(0, ("*** eraseFlash: Error detected after ERASE-BLOCK: SR[7:0] = 0x%lx\n",
			    (unsigned long) val16));
	    result = MethodResult_VerifyFailed;
	    bStatusError = TRUE;
	  }
	}
	  
	/* If an error occurred, attempt to return to normal operation */
	if (bTimeout || bInterrupt || bStatusError) {
	  /* Send CLEAR-STATUS command */
	  val16 = 0x50U;
	  pDevCtx->methods.flash(pDevCtx, TRUE, start, &val16);
	}
	
	/* Send READ-ARRAY command */
	val16 = 0xffU;
	pDevCtx->methods.flash(pDevCtx, TRUE, blockStart, &val16);
	break;
      }
      
    case 0x898803: /* Intel 28F256K3 */
      {
	/* Send UNLOCK-BLOCK command */
	val16 = 0x60U;
	pDevCtx->methods.flash(pDevCtx, TRUE, blockStart, &val16);
	val16 = 0xd0U;
	pDevCtx->methods.flash(pDevCtx, TRUE, blockStart, &val16);
	
	/* Send ERASE-BLOCK command */
	val16 = 0x20U;
	pDevCtx->methods.flash(pDevCtx, TRUE, 0, &val16);
	val16 = 0xd0U;
	pDevCtx->methods.flash(pDevCtx, TRUE, blockStart, &val16);
	
	/* Wait for completion */
	remaining_ms = timeout_ms;
	bTimeout = FALSE;
	bStatusError = FALSE;
	bInterrupt = FALSE;
	while (1) {
	  pDevCtx->methods.flash(pDevCtx, FALSE, 0, &val16);
	  if (val16 & 0x80U) {
	    if (val16 & 0x20U) {
	      result = MethodResult_GeneralFailure;
	    }
	    break;
	  }

	  interval = 5;
	  bInterrupt = os_delay_thread(FALSE, &interval);
	  if (bInterrupt) {
	    os_dbgprint(1, ("*** eraseFlash: delay interrupted\n"));
	    break;
	  }
	  if (remaining_ms < 5) {
	    bTimeout = TRUE;
	    break;
	  }
	  remaining_ms -= 5;
	}
	
	if (bInterrupt) {
	  result = MethodResult_GeneralFailure;
	}

	if (bTimeout) {
	  result = MethodResult_Timeout;
	} else {
	  /* Status-check */
	  if (val16 & 0x1cU) {
	    /* Vpp Error */
	    os_dbgprint(0, ("*** eraseFlash: Error detected after ERASE-BLOCK: SR[7:0] = 0x%lx\n",
			    (unsigned long) val16));
	    result = MethodResult_VerifyFailed;
	    bStatusError = TRUE;
	  }
	  if (val16 & 0x20U) {
	    /* Erase Error */
	    os_dbgprint(0, ("*** eraseFlash: Error detected after ERASE-BLOCK: SR[7:0] = 0x%lx\n",
			    (unsigned long) val16));
	    result = MethodResult_VerifyFailed;
	    bStatusError = TRUE;
	  }
	}
	
	/* If an error occurred, attempt to return to normal operation */
	if (bTimeout || bInterrupt || bStatusError) {
	  /* Send CLEAR-STATUS command */
	  val16 = 0x50U;
	  pDevCtx->methods.flash(pDevCtx, TRUE, start, &val16);
	}
	
	/* Send LOCK-BLOCK command */
	val16 = 0x60U;
	pDevCtx->methods.flash(pDevCtx, TRUE, blockStart, &val16);
	val16 = 0x01U;
	pDevCtx->methods.flash(pDevCtx, TRUE, blockStart, &val16);
	
	/* Send READ-ARRAY command */
	val16 = 0xffU;
	pDevCtx->methods.flash(pDevCtx, TRUE, blockStart, &val16);
	break;
      }

    case 0xbf00d9: /* SST 39VF016Q */
      {
	os_uint32     i;
	unsigned long interval;
	
	/* Send ERASE-BLOCK command */
	val16 = 0xaaU;
	pDevCtx->methods.flash(pDevCtx, TRUE, 0x5555, &val16);
	val16 = 0x55U;
	pDevCtx->methods.flash(pDevCtx, TRUE, 0x2aaa, &val16);
	val16 = 0x80U;
	pDevCtx->methods.flash(pDevCtx, TRUE, 0x5555, &val16);
	val16 = 0xaaU;
	pDevCtx->methods.flash(pDevCtx, TRUE, 0x5555, &val16);
	val16 = 0x55U;
	pDevCtx->methods.flash(pDevCtx, TRUE, 0x2aaa, &val16);
	val16 = 0x50U;
	pDevCtx->methods.flash(pDevCtx, TRUE, blockStart, &val16);
	
	/* Wait for completion */
	interval = timeout_ms;
	bInterrupt = os_delay_thread(FALSE, &interval);
	
	/* Return to READ-ARRAY mode */
	val16 = 0xffU;
	pDevCtx->methods.flash(pDevCtx, TRUE, 0x0, &val16);
	
	/* Verify that sector has been erased */
	for (i = blockStart; i < blockEnd; i++) {
	  pDevCtx->methods.flash(pDevCtx, FALSE, i, &val16);
	  if (val16 != 0xff) {
	    result = MethodResult_VerifyFailed;
	    break;
	  }
	}
	
	break;
      }
    }
    
    if (result != MethodResult_Success) {
      return result;
    }
    
    length -= chunk;
    start += chunk;
  }
  
  return MethodResult_Success;
}

MethodResult
readFlash(
  DeviceContext* pDevCtx,
  os_uint32      start,
  os_uint32      length,
  void*          buffer)
{
  MethodResult result = MethodResult_Success;
  os_uint32    i, length16;
  os_uint16    val16;
  os_uint8*    p8;

  os_dbgprint(3, ("readFlash: entered, start=0x%lx length=0x%lx buffer=0x%p\n",
		  (unsigned long) start, (unsigned long) length, buffer));

  if (!pDevCtx->info.flash.present) {
    return MethodResult_NotPresent;
  }

  if (!length) {
    return MethodResult_Success;
  }

  p8 = (os_uint8*) buffer;
  switch (pDevCtx->info.flash.width) {
  case 8:
    /* 8-bit wide Flash */
    i = start;
    while (length) {
      result = pDevCtx->methods.flash(pDevCtx, FALSE, i++, &val16);
      if (result != MethodResult_Success) {
	break;
      }
      length--;
      *p8++ = (os_uint8) val16;
    }
    break;
    
  case 16:
    /* 16-bit wide Flash */
    i = start;
    if (i & 0x1U) {
      /* Align to even address */
      result = pDevCtx->methods.flash(pDevCtx, FALSE, i & ~0x1U, &val16);
      if (result != MethodResult_Success) {
	goto done;
      }
      i++;
      length--;
      *p8++ = (os_uint8) (val16 >> 8);
    }
    length16 = length >> 1;
    while (length16) {
      result = pDevCtx->methods.flash(pDevCtx, FALSE, i, &val16);
      if (result != MethodResult_Success) {
	goto done;
      }
      i += 2;
      length16--;
      *p8++ = (os_uint8) (val16 >> 0);
      *p8++ = (os_uint8) (val16 >> 8);
    }
    if (length & 0x1U) {
      /* Finish off odd byte */
      result = pDevCtx->methods.flash(pDevCtx, FALSE, i, &val16);
      if (result != MethodResult_Success) {
	goto done;
      }
      *p8++ = (os_uint8) (val16 >> 0);
    }
    break;
  }

 done:
  return result;
}

MethodResult
writeFlash(
  DeviceContext* pDevCtx,
  os_uint32      start,
  os_uint32      length,
  const void*    data)
{
  MethodResult result = MethodResult_Success;
  os_uint8*    saveBuff;
  os_uint32    eraseBlock;
  os_uint32    blockStart, blockEnd;
  os_uint32    blockSize;
  os_boolean   isBootBlock;

  os_dbgprint(3, ("writeFlash: entered, start=0x%lx length=0x%lx data=0x%p\n",
		  (unsigned long) start, (unsigned long) length, data));

  if (!pDevCtx->info.flash.present) {
    return MethodResult_NotPresent;
  }
  
  saveBuff = pDevCtx->info.flash.buffer;
  eraseBlock = pDevCtx->info.flash.layout.eraseBlock;

  while (length) {
    os_uint32	chunk;
    
    identifyFlashBlock(pDevCtx, start, &blockStart, &blockSize, &isBootBlock);
    blockEnd = blockStart + blockSize;
    chunk = blockEnd - start;
    if (chunk > length) {
      chunk = length;
    }
    
    os_dbgprint(6, ("writeFlash: start=0x%lx blockStart=0x%lx blockEnd=0x%lx blockSize=0x%lx chunk=0x%lx data=0x%p saveBuff=0x%p\n",
		    (unsigned long) start, (unsigned long) blockStart, (unsigned long) blockEnd,
		    (unsigned long) blockSize, (unsigned long) chunk, data, saveBuff));
    
    /* Read the block */
    result = readFlash(pDevCtx, blockStart, blockSize, saveBuff);
    if (result != MethodResult_Success) {
      break;
    }
    
    /* Merge in new data */
    memcpy(saveBuff + (start - blockStart), data, chunk);
    
    /* Erase the block */
    result = eraseFlash(pDevCtx, blockStart, blockSize);
    if (result != MethodResult_Success) {
      break;
    }
    
    /* Write the block */
    result = writeBlock(pDevCtx, blockStart, blockSize, saveBuff);
    if (result != MethodResult_Success) {
      break;
    }
    
    length -= chunk;
    start += chunk;
    data = (const void*) (((os_uint8*) data) + chunk);
  }
  
  return result;
}
