/*
** ADM-XRC Series Driver
**
** (C) Copyright 2001-2004 Alpha Data
**
** flashrw.c - IOCTL handlers for reading and writing the Flash memory that
**             is attached to the PCI bridge.
**
** ----------------------------------------------------------------------------
*/

#include "device.h"
#include "drvparam.h"
#include "flash.h"
#include "flashrw.h"
#include "ioctl.h"

static OS_STATUS
mapMethodResultToStatus(
  MethodResult res)
{
  switch (res) {
  case MethodResult_Success:
    return OS_STATUS_SUCCESS;
  case MethodResult_Timeout:
    return OS_MAKE_STATUS(ADMXRC2_TIMEOUT);
  case MethodResult_VerifyFailed:
    return OS_MAKE_STATUS(ADMXRC2_FAILED);
  case MethodResult_InvalidLocation:
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  case MethodResult_NotPresent:
    return OS_MAKE_STATUS(ADMXRC2_NOT_SUPPORTED);
  case MethodResult_NotSupported:
    return OS_MAKE_STATUS(ADMXRC2_NOT_SUPPORTED);
  case MethodResult_GeneralFailure:
    return OS_MAKE_STATUS(ADMXRC2_FAILED);
  default:
    return OS_MAKE_STATUS(ADMXRC2_UNKNOWN_ERROR);
  }
}

/*
** ---------------------------------------------------------------------
** Exported routines
** ---------------------------------------------------------------------
*/

OS_STATUS
ADMXRC_EraseFlash(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_ERASEFLASH* in;
  MethodResult              res;
  OS_STATUS                 status;
  os_uint32                 start, length, limit;
  os_uint32                 tgtStart, tgtLimit;

  in = (ADMXRC_IOCTLS_ERASEFLASH*) req->param.inBuf;
  if (in == NULL || req->param.inSize != sizeof(in->in)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  if (in->in.signature1 != ADMXRC_IOCTLS_ERASEFLASH_SIGNATURE1 ||
      in->in.signature2 != ADMXRC_IOCTLS_ERASEFLASH_SIGNATURE2) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  start = in->in.start;
  length = in->in.length;
  limit = start + length;
  tgtStart = pDevCtx->info.flash.targetArea.start;
  tgtLimit = tgtStart + pDevCtx->info.flash.targetArea.length;
  if (driverParams.model.admxrc2pl.unlockFlash || (start >= tgtStart && limit <= tgtLimit)) {
    res = eraseFlash(pDevCtx, start, length);
    status = mapMethodResultToStatus(res);
  } else {
    status = OS_MAKE_STATUS(ADMXRC_FAILED);
  }
  
  return status;
}

OS_STATUS
ADMXRC_GetFlashInfo(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* req)
{
  OS_STATUS                   status = OS_STATUS_SUCCESS;
  ADMXRC_IOCTLS_GETFLASHINFO* out;
  ADMXRC2_FLASH_INFO*         info;
  
  out = (ADMXRC_IOCTLS_GETFLASHINFO*) req->param.outBuf;
  if (out == NULL || req->param.outSize != sizeof(out->out)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  
  if (!pDevCtx->info.flash.present) {
    return OS_MAKE_STATUS(ADMXRC_NOT_SUPPORTED);
  }
  
  info = &out->out.info;
  strncpy(info->DeviceName, pDevCtx->info.flash.identity.name, 31);
  info->DeviceName[31] = '\0';
  info->DeviceSize = pDevCtx->info.flash.size;
  info->DeviceBlockSize = pDevCtx->info.flash.layout.eraseBlock;
  info->UseableStart = pDevCtx->info.flash.targetArea.start;
  info->UseableLength = pDevCtx->info.flash.targetArea.length;
  
  req->status.length = sizeof(out->out);

  return status;
}

OS_STATUS
ADMXRC_GetFlashBlockInfo(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* req)
{
  OS_STATUS                        status = OS_STATUS_SUCCESS;
  ADMXRC_IOCTLS_GETFLASHBLOCKINFO* in;
  ADMXRC_IOCTLS_GETFLASHBLOCKINFO* out;
  ADMXRC2_FLASHBLOCK_INFO*         info;
  os_uint32                        location, blockStart, blockSize;
  os_boolean                       bBootBlock;

  in = (ADMXRC_IOCTLS_GETFLASHBLOCKINFO*) req->param.inBuf;
  if (in == NULL || req->param.inSize != sizeof(in->in)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  out = (ADMXRC_IOCTLS_GETFLASHBLOCKINFO*) req->param.outBuf;
  if (out == NULL || req->param.outSize != sizeof(out->out)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  if (!pDevCtx->info.flash.present) {
    return OS_MAKE_STATUS(ADMXRC_NOT_SUPPORTED);
  }
  
  location = in->in.location;
  if (location >= pDevCtx->info.flash.size) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  identifyFlashBlock(pDevCtx, location, &blockStart, &blockSize, &bBootBlock);
  info = &out->out.info;
  info->Start = blockStart;
  info->Length = blockSize;
  info->Flags = 0;
  if (bBootBlock) {
    info->Flags |= ADMXRC2_FLASHBLOCKFLAG_BOOT;
  }

  req->status.length = sizeof(out->out);
  
  return status;
}

OS_STATUS
ADMXRC_ReadFlash(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_READFLASH* in;
  MethodResult             res;
  OS_STATUS                status;

  in = (ADMXRC_IOCTLS_READFLASH*) req->param.inBuf;
  if (in == NULL || req->param.inSize != sizeof(in->in)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  res = readFlash(pDevCtx, in->in.start, req->param.massSize, req->param.massBuf);
  status = mapMethodResultToStatus(res);

  return status;
}

OS_STATUS
ADMXRC_WriteFlash(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* req)
{
  MethodResult              res;
  OS_STATUS                 status;
  ADMXRC_IOCTLS_WRITEFLASH* in;
  unsigned long             length;
  os_uint32                 start, limit;
  os_uint32                 tgtStart, tgtLimit;

  in = (ADMXRC_IOCTLS_WRITEFLASH*) req->param.inBuf;
  if (in == NULL || req->param.inSize != sizeof(in->in)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  if (in->in.signature1 != ADMXRC_IOCTLS_WRITEFLASH_SIGNATURE1 ||
      in->in.signature2 != ADMXRC_IOCTLS_WRITEFLASH_SIGNATURE2) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  length = in->in.length;
  if (req->param.massSize != in->in.length) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  
  start = in->in.start;
  limit = start + length;
  tgtStart = pDevCtx->info.flash.targetArea.start;
  tgtLimit = tgtStart + pDevCtx->info.flash.targetArea.length;
  if (driverParams.model.admxrc2pl.unlockFlash || (start >= tgtStart && limit <= tgtLimit)) {
    res = writeFlash(pDevCtx, start, length, req->param.massBuf);
    status = mapMethodResultToStatus(res);
  } else {
    status = OS_MAKE_STATUS(ADMXRC2_FAILED);
  }

  return status;
}
