/*
** ADM-XRC Series Driver
**
** (C) Copyright 2001-2004 Alpha Data
**
** intwait.c - IOCTL handlers for waiting on FPGA interrupts.
**
** ----------------------------------------------------------------------------
*/

#include "platform.h"
#include "device.h"
#include "ioctl.h"
#include "intwait.h"

static void
intWaitCancel(
  OS_DEVICE_OBJECT* devObj,
  OS_IO_REQUEST*    req)
{
  OS_SPINLOCK_FLAGS f;
  DeviceContext*    devCtx = (DeviceContext*) os_get_device_context(devObj);
  os_boolean        complete;

  os_dbgprint(3, ("intWaitCancel: entered, req = 0x%p\n", req));

  os_acquire_spinlock(&devCtx->intwait.lock, &f);
  if (devCtx->intwait.waiting.req != NULL) {
    if (devCtx->intwait.waiting.completing) {
      /* Was about to be completed by intWaitComplete() */
      complete = TRUE;
    } else {
      /* Queued, intWaitComplete() not running */
      devCtx->intwait.waiting.req = NULL;
      complete = TRUE;
    }
  } else {
    os_dbgprint(3, ("intWaitCancel: no request queued, req = 0x%p\n", req));  
    complete = FALSE;
  }
  os_release_spinlock(&devCtx->intwait.lock, f);

  if (complete) {
    os_dbgprint(3, ("intWaitCancel: completing, req = 0x%p\n", req));  
    req->status.status = OS_MAKE_STATUS(ADMXRC_CANCELLED);
    os_complete_request(req);
  }
}

static void
intWaitComplete(
  DeviceContext* devCtx)
{
  OS_IO_REQUEST*                  req = NULL;
  unsigned long                   now;
  long                            remaining;
  OS_SPINLOCK_FLAGS               f;
  ADMXRC_IOCTLS_WAITFORINTERRUPT* optr;
  os_boolean                      complete;

  os_dbgprint(3, ("intWaitComplete: entered\n"));

  os_acquire_spinlock(&devCtx->intwait.lock, &f);
  req = devCtx->intwait.waiting.req;
  if (req == NULL) {
    /* No IO request waiting - set the signalled flag and do nothing more */
    os_dbgprint(6, ("intWaitComplete: no IO request queued, setting signalled flag\n"));
    devCtx->intwait.signalled = TRUE;
    complete = FALSE;
  } else {
    /* There is a waiting IO request */
    devCtx->intwait.waiting.req = NULL;
    devCtx->intwait.waiting.completing = TRUE;
    if (os_set_cancel_routine(req, NULL)) {
      /*
      ** Cancel routine wasn't running - so it's safe to complete
      ** the request.
      */
      complete = TRUE;
    } else {
      os_dbgprint(3, ("intWaitComplete: cancel routine running - not completing req 0x%p\n", req));
      /*
      ** Cancel routine is already running, but spinning on the lock.
      ** Let the cancel rountine complete the request.
      */
      complete = FALSE;
    }
  }
  os_release_spinlock(&devCtx->intwait.lock, f);

  if (complete) {
    os_dbgprint(3, ("intWaitComplete: completing req 0x%p\n", req));
    /*
    ** Cancel routine successfully removed - it's safe to complete
    ** the request here
    */
    now = os_get_time();
    optr = (ADMXRC_IOCTLS_WAITFORINTERRUPT*) req->param.outBuf;
    if (!devCtx->intwait.expire.never) {
#if 0
      os_uint64 val64;
#endif

      if (devCtx->intwait.expire.ticks > now) {
	remaining = devCtx->intwait.expire.ticks - now;
      } else {
	remaining = 0;
      }

#if 0
      val64 = (unsigned long) ((os_uint64) remaining) * 1000000;
      optr->out.remaining_us = (unsigned long) os_div_64by32(val64, os_ticks_per_second);
#else
      optr->out.remaining_us = 0;
#endif
    }
    req->status.length = sizeof(optr->out);
    req->status.status = OS_STATUS_SUCCESS;
    os_complete_request(req);
  }
}

static void
signalIntEvents(
  DeviceContext* pDevCtx)
{
  OS_SPINLOCK_FLAGS f;
  ADMXRC_INTEVENT* ev;

  /* Signal all registered interrupt wait events */
  os_acquire_spinlock(&pDevCtx->intwait.lock, &f);
  ev = pDevCtx->intwait.eventList;
  while (1) {
    if (ev == NULL) {
      break;
    }
    os_dbgprint(6, ("signalIntEvents: event list node = 0x%lx\n",
		 (unsigned long) ev));
    os_set_event(ev->event);
    ev = ev->next;
  }
  os_release_spinlock(&pDevCtx->intwait.lock, f);
}

/*
** The FPGA interrupt DPC
*/

void
ADMXRC_FpgaIntDpc(
  void* context,
  void* argument)
{
  DeviceContext* devCtx = (DeviceContext*) context;

  intWaitComplete(devCtx);
  signalIntEvents(devCtx);
}

/*
** IOCTL handlers
*/

OS_STATUS
ADMXRC_CancelWaitForInterrupt(
  DeviceContext* devCtx,
  OS_IO_REQUEST* req)
{
  OS_SPINLOCK_FLAGS               f;
  os_boolean                      complete;
  OS_IO_REQUEST*                  creq;
  OS_STATUS                       status;

  os_dbgprint(3, ("ADMXRC_CancelWaitForInterrupte: entered\n"));

  os_acquire_spinlock(&devCtx->intwait.lock, &f);
  creq = devCtx->intwait.waiting.req;
  if (creq == NULL) {
    /* No IO request waiting - do nothing more */
    complete = FALSE;
  } else {
    /* There is a waiting IO request */
    devCtx->intwait.waiting.req = NULL;
    devCtx->intwait.waiting.completing = TRUE;
    if (os_set_cancel_routine(creq, NULL)) {
      /*
      ** Cancel routine wasn't running - so it's safe to complete
      ** the request.
      */
      complete = TRUE;
    } else {
      os_dbgprint(3, ("intWaitComplete: cancel routine running - not completing req 0x%p\n", req));
      /*
      ** Cancel routine is already running, but spinning on the lock.
      ** Let the cancel rountine complete the request.
      */
      complete = FALSE;
    }
  }
  os_release_spinlock(&devCtx->intwait.lock, f);

  if (complete) {
    /*
    ** Cancel routine successfully removed - it's safe to complete
    ** the request here
    */
    os_dbgprint(3, ("ADMXRC_CancelWaitForInterrupt: completing req 0x%p\n", req));
    creq->status.status = OS_MAKE_STATUS(ADMXRC2_CANCELLED);
    os_complete_request(creq);
    status = OS_STATUS_SUCCESS;
  } else {
    status = OS_MAKE_STATUS(ADMXRC2_FAILED);
  }

  return status;
}

OS_STATUS
ADMXRC_WaitForInterrupt(
  DeviceContext* devCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_WAITFORINTERRUPT* iptr;
  ADMXRC_IOCTLS_WAITFORINTERRUPT* optr;
  OS_SPINLOCK_FLAGS               f;
  unsigned long                   offset;
  OS_STATUS                       status = OS_STATUS_SUCCESS;

  iptr = (ADMXRC_IOCTLS_WAITFORINTERRUPT*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  optr = (ADMXRC_IOCTLS_WAITFORINTERRUPT*) req->param.outBuf;
  if (optr == NULL || req->param.outSize != sizeof(optr->out)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  optr->out.expired = 0;
  os_acquire_spinlock(&devCtx->intwait.lock, &f);
  if (devCtx->intwait.waiting.req != NULL) {
    /* A thread is already waiting for an interrupt */
    os_release_spinlock(&devCtx->intwait.lock, f);
    status = OS_MAKE_STATUS(ADMXRC_DEVICE_BUSY);
  } else {
    if (devCtx->intwait.signalled) {
      /* An FPGA interrupt has already occurred */
      devCtx->intwait.signalled = FALSE;
      os_release_spinlock(&devCtx->intwait.lock, f);
      optr->out.remaining_us = iptr->in.timeout_us;
      req->status.length = sizeof(optr->out);
      status = OS_STATUS_SUCCESS;
    } else {
      if (os_set_cancel_routine(req, intWaitCancel)) {
	/* We must wait for an FPGA interrupt */
	if (0 && iptr->in.flags & ADMXRC_INTWAIT_TIMEOUT) {
	  os_uint64 val64 = ((os_uint64) iptr->in.timeout_us) * 1000000;

	  offset = (unsigned long) os_div_64by32(val64, os_ticks_per_second);
	  devCtx->intwait.expire.ticks = os_get_time() + offset;
	  devCtx->intwait.expire.never = FALSE;
	} else {
	  devCtx->intwait.expire.never = TRUE;
	}
	devCtx->intwait.waiting.req = req;
	devCtx->intwait.waiting.completing = FALSE;
	os_release_spinlock(&devCtx->intwait.lock, f);
	status = OS_STATUS_PENDING;
      } else {
	/* IO request already pending cancellation */
	os_release_spinlock(&devCtx->intwait.lock, f);
	status = OS_MAKE_STATUS(ADMXRC_CANCELLED);
      }
    }
  }

  return status;
}

void
ADMXRC_UnregAllInterruptEvents(
  DeviceContext* pDevCtx)
{
  OS_SPINLOCK_FLAGS f;
  ADMXRC_INTEVENT* node;
  ADMXRC_INTEVENT* next;
  int n;

  os_dbgprint(2, ("AdmXrc_UnregAllIntEvents: entered\n"));
  
  os_acquire_spinlock(&pDevCtx->intwait.lock, &f);
  node = pDevCtx->intwait.eventList;
  pDevCtx->intwait.eventList = NULL;
  os_release_spinlock(&pDevCtx->intwait.lock, f);

  n = 0;
  while (node != NULL) {
    next = node->next;
    os_kernel_free(node);
    node = next;
    n++;
  }
  
  os_dbgprint(2, ("AdmXrc_UnregAllIntEvents: %ld interrupt events unregistered\n", 
		  (unsigned long) n));
}

OS_STATUS
ADMXRC_RegInterruptEvent(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_REGINTERRUPTEVENT* iptr;
  OS_SPINLOCK_FLAGS f;
  ADMXRC_INTEVENT* node;
  OS_STATUS status = OS_STATUS_SUCCESS;
  
  iptr = req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  node = (ADMXRC_INTEVENT*) os_kernel_malloc(sizeof(ADMXRC_INTEVENT));
  if (node == NULL) {
    return OS_MAKE_STATUS(ADMXRC_NO_MEMORY);
  } 
  node->event = iptr->in.event;

  os_acquire_spinlock(&pDevCtx->intwait.lock, &f);
  node->next = pDevCtx->intwait.eventList;
  pDevCtx->intwait.eventList = node;
  os_release_spinlock(&pDevCtx->intwait.lock, f);
  
  return status;
}

OS_STATUS
ADMXRC_UnregInterruptEvent(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_UNREGINTERRUPTEVENT* iptr;
  OS_SPINLOCK_FLAGS f;
  ADMXRC_INTEVENT* node;
  ADMXRC_INTEVENT** pnode;
  os_event event;
  
  iptr = (ADMXRC_IOCTLS_UNREGINTERRUPTEVENT*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  event = iptr->in.event;
  os_acquire_spinlock(&pDevCtx->intwait.lock, &f);
  pnode = &pDevCtx->intwait.eventList;
  while (1) {
    node = *pnode;
    if (node == NULL) {
      break;
    } else {
      if (node->event == event) {
	*pnode = node->next;
	break;
      } else {
	pnode = &node->next;
      }
    }
  }
  os_release_spinlock(&pDevCtx->intwait.lock, f);
  
  if (node == NULL) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_HANDLE);
  } else {
    os_kernel_free(node);
  }
  
  return OS_STATUS_SUCCESS;
}

