/*
** ADM-XRC Series Linux Driver
**
** (C) Copyright 2001-2004 Alpha Data
**
** linux-2.4.c - Linux 2.4.x specific routines
**
** This module handles most of the Linux-specific device creation,
** initialisation and destruction issues.
**
** ----------------------------------------------------------------------------
*/

#include "platform.h"
#include "ioctl_platform.h"

#include <linux/config.h>
#ifdef CONFIG_SMP
# define __SMP__
#endif

#include <linux/version.h>
#include <asm/uaccess.h>
#include <asm/segment.h>
#include <asm/io.h>
#include <asm/pgtable.h>
#include <asm/byteorder.h>
#include <asm/errno.h>
#include <asm/semaphore.h>

#include <linux/module.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/timer.h>
#include <linux/pci.h>
#include <linux/mman.h>
#include <linux/mm.h>
#include <linux/vmalloc.h>
#include <linux/ioport.h>
#include <linux/iobuf.h>
#include <linux/delay.h>
#include <linux/ioctl.h>
#include <linux/tqueue.h>
#include <linux/sem.h>
#include <linux/param.h>
#include <linux/fs.h>

#define MAX_DEVICE          (16)

#define OS_MAX_IOCTL_LENGTH (256)

static OS_DRIVER_OBJECT       drvObj;
static os_boolean             drvObjInitted;
static struct pci_device_id*  pci_tbl;
static struct pci_driver      driver;
static os_boolean             driverReged;
static int                    major;
static os_boolean             majorValid;
static struct file_operations fops;
static OS_DEVICE_OBJECT*      devices[MAX_DEVICE];

/*
** This should probably be per-architecture in <asm/pgtable.h>
*/
#if 1
static inline pgprot_t pgprot_noncached(pgprot_t _prot)
{
  unsigned long prot = pgprot_val(_prot);
  
#if defined(__i386__) || defined(__x86_64__)
  /*
  ** On PPro and successors, PCD alone doesn't always mean 
  ** uncached because of interactions with the MTRRs. PCD | PWT
  ** means definitely uncached.
  */ 
  if (boot_cpu_data.x86 > 3)
    prot |= _PAGE_PCD | _PAGE_PWT;
#elif defined(__powerpc__)
  prot |= _PAGE_NO_CACHE | _PAGE_GUARDED;
#elif defined(__mc68000__)
#ifdef SUN3_PAGE_NOCACHE
  if (MMU_IS_SUN3)
    prot |= SUN3_PAGE_NOCACHE;
  else
#endif
    if (MMU_IS_851 || MMU_IS_030)
      prot |= _PAGE_NOCACHE030;
  /* Use no-cache mode, serialized */
    else if (MMU_IS_040 || MMU_IS_060)
      prot = (prot & _CACHEMASK040) | _PAGE_NOCACHE_S;
#endif
  
  return __pgprot(prot);
}
#endif

/*
** Map an OS_STATUS code to an IOCTL error return value.
**
** Note that the return value from a driver's IOCTL routine
** should be negative to indicate an error; the application
** receives the positive error code in 'errno'.
*/
static int
map_error(
  OS_STATUS status)
{
  if (status & OS_STATUSFLAG_PENDING) {
    os_dbgprint(0, ("map_error: shouldn't get a PENDING status of 0x%lx\n",
		    (unsigned long) status));
    return -EBUSY;
  }

  if (status & OS_STATUSFLAG_USERCODE) {
    return OS_USERCODE(status);
  } else {
    switch (status) {
    case OS_STATUS_INVALIDIOCTL:
      return -ENOTTY;
    default:
      os_dbgprint(0, ("map_error: can't map OS_STATUS of 0x%lx to an errno\n",
		      (unsigned long) status));
      return -EINVAL;
    }
  }
}

static int
init_ioreq(
  OS_IO_REQUEST*       ioreq,
  unsigned int         code,
  ADMXRC_IOCTL_PARAMS* iop)
{
  int err = 0;

  ioreq->param.massSize = iop->mass.nSize;

  if (!iop->mass.nSize) {
    /* No need to buffer anything */
    ioreq->opaque.mass.strategy = -1;
    return 0;
  } else if (ioreq->param.massSize <= PAGE_SIZE) {
    /* Small buffer strategy */
    ioreq->opaque.mass.strategy = 0;
  } else {
    /* Large buffer strategy */
    ioreq->opaque.mass.strategy = 2;
  }

  switch (ioreq->opaque.mass.strategy) {
  case 0:
    {
      /*
      ** Small buffer strategy (used when mass size is a page or less)
      **
      ** 1. Allocate a page-sized buffer using __get_free_page()
      ** 2. If user is writing data, copy the user space data into this buffer
      ** 3. Call the platform-independent IOCTL handler
      ** 4. If successful and the user is reading data, then copy the page-sized
      **    buffer back into user space
      ** 5. Free the page-sized buffer using free_page()
      */
      void* p;
      
      p = (void*) __get_free_page(GFP_KERNEL | __GFP_DMA);
      if (p == NULL) {
	return -ENOMEM;
      }
      
      if (ADMXRC_GET_IOCTLF_MASS(code) & ADMXRC_IOCTLF_MASS_WRITE) {
	/* User is writing */
	if (copy_from_user(p, iop->mass.lpBuffer, iop->mass.nSize)) {
	  free_page((unsigned long) p);
	  return -EFAULT;
	}
      }
      
      ioreq->param.massBuf = p;
      break;
    }
#if 0
  case 1:
    {
      /*
      ** Locked buffer strategy
      */
      unsigned int  offset;
      unsigned long endlen;
      void*         ps;
      void*         pe;
      
      offset = ((unsigned long) iop->mass.lpBuffer) & (PAGE_SIZE - 1);
      if (offset) {
	ps = (void*) __get_free_page(GFP_KERNEL | __GFP_DMA);
	if (ps == NULL) {
	  return -ENOMEM;
	}
      } else {
	ps = NULL;
      }
      
      endlen = (((unsigned long) iop->mass.lpBuffer) + iop->mass.nSize) & (PAGE_SIZE - 1);
      if (endlen) {
	pe = (void*) __get_free_page(GFP_KERNEL | __GFP_DMA);
	if (pe == NULL) {
	  free_page((unsigned long) ps);
	  return -ENOMEM;
	}
      } else {
	pe = NULL;
      }
      
      ioreq->opaque.mass.locked.start_buf = ps;
      ioreq->opaque.mass.locked.end_buf = pe;
      break;
    }
#endif
  case 2:
    {
      /*
      ** Large buffer strategy (used when mass size is larger than a page
      ** and the locked buffer strategy fails)
      **
      ** 1. Allocate a buffer large enough for all of the user data, using vmalloc()
      ** 2. If user is writing data, copy the user space data into this buffer
      ** 3. Call the platform-independent IOCTL handler
      ** 4. If successful and the user is reading data, then copy the page-sized
      **    buffer back into user space
      ** 5. Free the large buffer using vfree()
      */
      void* p;

      /* Use vmalloc_32() in order to get a buffer that PCI devices can access */
      p = vmalloc_32(iop->mass.nSize);
      if (p == NULL) {
	return -ENOMEM;
      }

      if (ADMXRC_GET_IOCTLF_MASS(code) & ADMXRC_IOCTLF_MASS_WRITE) {
	/* User is writing */
	if (copy_from_user(p, iop->mass.lpBuffer, iop->mass.nSize)) {
	  vfree(p);
	  return -EFAULT;
	}
      }
      
      ioreq->param.massBuf = p;
      break;
    }
  default:
    break;
  }

  return err;
}

static int
uninit_ioreq(
  OS_IO_REQUEST*       ioreq,
  unsigned int         code,
  ADMXRC_IOCTL_PARAMS* iop,
  os_boolean           success)
{
  int err = 0;

  switch (ioreq->opaque.mass.strategy) {
  case 0:
    {
      /*
      ** Small buffer strategy (memory allocated with __get_free_page())
      */
      if (ioreq->param.massBuf != NULL) {
	if (success && (ADMXRC_GET_IOCTLF_MASS(code) & ADMXRC_IOCTLF_MASS_READ)) {
	  /* User is reading */
	  if (copy_to_user(iop->mass.lpBuffer, ioreq->param.massBuf, iop->mass.nSize)) {
	    err = -EFAULT;
	  }
	}	
	free_page((unsigned long) ioreq->param.massBuf);
      }
      break;
    }
#if 0
  case 1:
    {
      /*
      ** Locked buffer strategy (up to 2 pages buffered + plus kiovec) 
      */
      if (ioreq->opaque.mass.locked.start_buf != NULL) {
	if (success && (iop->mass.mode & ADMXRC_IOCTLF_MASS_READ)) {
	  unsigned long offset = (((unsigned long) iop->mass.lpBuffer) + iop->mass.nSize) & (PAGE_SIZE - 1);

	  if (copy_to_user(iop->mass.lpBuffer, ioreq->opaque.mass.locked.start_buf, PAGE_SIZE - offset)) {
	    err = -EFAULT;
	  }
	}
	
	free_page((unsigned long) ioreq->opaque.mass.locked.start_buf);
      }
      
      /*
      ** If necessary, copy the last page buffer back to user space
      ** and free the last page buffer
      */
      if (ioreq->opaque.mass.locked.end_buf != NULL) {
	if (success && (iop->mass.mode & ADMXRC_IOCTL_MASS_READ)) {
	  unsigned long end = (((unsigned long) iop->mass.lpBuffer) + iop->mass.nSize) & ~(PAGE_SIZE - 1);
	  unsigned long n   = (((unsigned long) iop->mass.lpBuffer) + iop->mass.nSize) & (PAGE_SIZE - 1);
	  if (copy_to_user(((os_uint8*) iop->mass.lpBuffer) + end, ioreq->opaque.mass.locked.end_buf, n)) {
	    err = -EFAULT;
	  }
	}
	
	free_page((unsigned long) ioreq->opaque.mass.locked.end_buf);
      }
      break;
    }
#endif
  case 2:
    {
      /*
      ** Large buffer strategy (memory for entire mass buffer allocated with vmalloc)
      */
      if (ioreq->param.massBuf != NULL) {
	if (success && (ADMXRC_GET_IOCTLF_MASS(code) & ADMXRC_IOCTLF_MASS_READ)) {
	  if (copy_to_user(iop->mass.lpBuffer, ioreq->param.massBuf, iop->mass.nSize)) {
	    err = -EFAULT;
	  }
	}
	vfree(ioreq->param.massBuf);
      }
      break;
    }
  default:
    break;
  }

  return err;
}

static void
cancel_ioreq(
  OS_IO_REQUEST*    req)
{
  os_boolean        wait, cancel = FALSE;
  OS_SPINLOCK_FLAGS f;
  OS_DEVICE_OBJECT* devObj = req->opaque.devObj;

  os_dbgprint(10, ("cancel_ioreq: entered, req = 0x%p\n", (void*) req));

  os_acquire_spinlock(&devObj->cancelLock, &f);
  if (!req->opaque.completed) {
    /*
    ** Call the cancel routine and wait uninterruptibly for completion.
    */
    wait = TRUE;
    set_current_state(TASK_UNINTERRUPTIBLE);
    if (req->opaque.cancelFunc != NULL) {
      req->opaque.cancel = TRUE;
      cancel = TRUE;
    }
  } else {
    /* Nothing to cancel - already completed */
    wait = FALSE;
  }
  os_release_spinlock(&devObj->cancelLock, f);

  if (wait) {
    if (cancel) {
      req->opaque.cancelFunc(devObj, req);
    }
    schedule();
  }
}

static os_boolean
deliver_ioreq(
  OS_DEVICE_OBJECT* devObj,
  struct file*      filp,  
  unsigned long     code,
  OS_IO_REQUEST*    req)
{
  wait_queue_t      wait;
  wait_queue_head_t wait_q;
  os_boolean        sig;

  req->opaque.wait_q = &wait_q;
  req->opaque.completed = FALSE;
  req->opaque.cancel = FALSE;
  req->opaque.cancelFunc = NULL;
  req->opaque.devObj = devObj;

  req->status.length = 0;

  init_waitqueue_head(&wait_q);
  init_waitqueue_entry(&wait, current);
  set_current_state(TASK_INTERRUPTIBLE);
  add_wait_queue(&wait_q, &wait);
  devObj->drvObj->onIoctl(devObj, (void*) filp, req);
  schedule();

  /* Did we catch a signal while waiting? */
  if (signal_pending(current)) {
    sig = TRUE;
    cancel_ioreq(req);
  } else {
    sig = FALSE;
  }

  /* Is this necessary? */
  remove_wait_queue(&wait_q, &wait);

  return sig;
}

/*
** -----------------------------------------------------------------
** file_operations methods
** -----------------------------------------------------------------
*/

static int
fops_open(
  struct inode* inode,
  struct file*  filp)
{
  int               dev_num = MINOR(inode->i_rdev);
  OS_DEVICE_OBJECT* devObj;
  OS_STATUS         status;
  int               err = 0;

  if (dev_num > MAX_DEVICE) {
    return -EINVAL;
  }

  devObj = devices[dev_num];
  if (devObj != NULL) {
    filp->private_data = devObj;
    MOD_INC_USE_COUNT;

    if (devObj->drvObj->onOpen != NULL) {
      /* Cast the file to an opaque type */
      status = devObj->drvObj->onOpen(devObj, (void*) filp);
      if (!OS_SUCCESS(status)) {
	err = map_error(status);
      }
    }
  } else {
    err = -EINVAL;
  }

  return err;
}

static int
fops_release(
  struct inode* inode,
  struct file*  filp)
{
  OS_DEVICE_OBJECT* devObj;
  OS_STATUS         status;
  int               err = 0;

  devObj = (OS_DEVICE_OBJECT*) filp->private_data;

  if (devObj->drvObj->onClose != NULL) {
    status = devObj->drvObj->onClose(devObj, (void*) filp);
    if (!OS_SUCCESS(status)) {
      /* Do nothing */
    }
  }

  MOD_DEC_USE_COUNT;

  return err;
}

static int
fops_ioctl(
  struct inode* inode,
  struct file*  filp,
  unsigned int  code,
  unsigned long arg)
{
  ADMXRC_IOCTL_PARAMS  iop;
  OS_DEVICE_OBJECT*    devObj;
  int                  err = 0;
  OS_IO_REQUEST        ioreq;
  os_boolean           sig;
  os_uint8             inBuf[OS_MAX_IOCTL_LENGTH];
  os_uint8             outBuf[OS_MAX_IOCTL_LENGTH];
  unsigned long        n;

  devObj = (OS_DEVICE_OBJECT*) filp->private_data;

  /* If no IOCTL handler supplied, return 'unsupported IOCTL' code */
  if (devObj->drvObj->onIoctl == NULL) {
    return -ENOTTY;
  }

  if (copy_from_user(&iop, (void*) arg, sizeof(iop))) {
    os_dbgprint(6, ("fops_ioctl: fault copying iop from user space\n"));
    return -EFAULT;
  }

  /* Sanity checks */
  if (iop.in.nSize) {
    if (iop.in.lpBuffer == NULL) {
      os_dbgprint(6, ("fops_ioctl: iop.in.lpBuffer is NULL\n"));
      return -EINVAL;
    }
    if (iop.in.nSize > OS_MAX_IOCTL_LENGTH) {
      os_dbgprint(6, ("fops_ioctl: iop.in.nSize invalid %ld(0x%lx)\n",
		      (unsigned long) iop.in.nSize, (unsigned long) iop.in.nSize));
      return -EINVAL;
    }
  }
  if (iop.out.nSize) {
    if (iop.out.lpBuffer == NULL) {
      os_dbgprint(6, ("fops_ioctl: iop.out.lpBuffer is NULL\n"));
      return -EINVAL;
    }
    if (iop.out.nSize > OS_MAX_IOCTL_LENGTH) {
      os_dbgprint(6, ("fops_ioctl: iop.out.nSize invalid %ld(0x%lx)\n",
		      (unsigned long) iop.out.nSize, (unsigned long) iop.out.nSize));
      return -EINVAL;
    }
  }
  if (iop.mass.nSize) {
    if (iop.mass.lpBuffer == NULL) {
      os_dbgprint(6, ("fops_ioctl: iop.mass.lpBuffer is NULL\n"));
      return -EINVAL;
    }
  }

  /* Get IOCTL input structure from user space */
  if (iop.in.nSize && copy_from_user(inBuf, iop.in.lpBuffer, iop.in.nSize)) {
    os_dbgprint(6, ("fops_ioctl: fault copying iop.in.lpBuffer from user space\n"));
    return -EFAULT;
  }

  ioreq.param.code = code;
  ioreq.param.inBuf = inBuf;
  ioreq.param.inSize = iop.in.nSize;
  ioreq.param.outBuf = outBuf;
  ioreq.param.outSize = iop.out.nSize;
  /* Set up any mass buffer required */
  err = init_ioreq(&ioreq, code, &iop);
  if (err) {
    return err;
  }

  sig = deliver_ioreq(devObj, filp, code, &ioreq);
  if (sig) {
    err = -EINTR;
    goto failed;
  }

  if (!OS_SUCCESS(ioreq.status.status)) {
    err = map_error(ioreq.status.status);
    goto failed;
  }

  /* Copy IOCTL output structure back to user space */
  n = (ioreq.status.length > iop.out.nSize) ? iop.out.nSize : ioreq.status.length;
  if (n && copy_to_user(iop.out.lpBuffer, outBuf, n)) {
    os_dbgprint(6, ("fops_ioctl: fault copying iop.out.lpBuffer to user space\n"));
    err = -EFAULT;
    goto failed;
  }

  err = uninit_ioreq(&ioreq, code, &iop, TRUE /* copy mass buffer back to user space */);

  return err;

 failed:
  /* Get here if there was a failure after the IO request was initialised */
  uninit_ioreq(&ioreq, code, &iop, FALSE /* don't copy mass buffer back to user space */);
  return err;
}

static int
fops_mmap(
  struct file*           filp,
  struct vm_area_struct* vma)
{
  int                 err = 0;
  OS_DEVICE_OBJECT*   devObj;
  OS_MAPPABLE_REGION* region;
  os_boolean          found = FALSE;
  unsigned long       offset, length;

  devObj = (OS_DEVICE_OBJECT*) filp->private_data;

  os_dbgprint(3, ("fops_mmap: entered for devObj 0x%p\n", (void*) devObj));

#if (LINUX_VERSION_CODE >= 0x020400)
  offset = vma->vm_pgoff << PAGE_SHIFT;
#else
  offset = vma->vm_offset;
#endif
  length = vma->vm_end - vma->vm_start;

  region = devObj->maps.regions;
  while (region != NULL) {
    if (offset >= region->phys && (offset + length) <= (region->phys + region->length)) {
      /* OK to map, request is within this region's bounds */
      found = TRUE;
      /* We need to make sure that the page range is uncached on a PPC */
      vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
      vma->vm_flags |= VM_RESERVED;  /* Don't swap physical pages */
      vma->vm_flags |= VM_IO;	     /* Don't core dump IO pages */
#if (REMAP_PAGE_RANGE_WITH_VMA_ARG==1)
      if (remap_page_range(vma,
			   vma->vm_start,
			   offset,
			   length,
			   vma->vm_page_prot)) {
#else
      if (remap_page_range(vma->vm_start,
			   offset,
			   length,
			   vma->vm_page_prot)) {
#endif
	os_dbgprint(0, ("fops_mmap: remap_page_range() failed\n"));
	err = -EAGAIN;
      } else {
	if (devObj->drvObj->onMap != NULL) {
	  devObj->drvObj->onMap(devObj, region, region->context, (void*) vma->vm_start, offset, length);
	}
      }

      break;
    }
    region = region->next;
  }

  if (!found) {
    err = -EINVAL;
  }

  return err;
}

/* 
** Module initialization functions for Linux 2.4 Kernels
**
** Uses the new PCI driver interface
*/

/* PCI device init function */
static int
init_board(
  struct pci_dev*             dev,
  const struct pci_device_id* entry)
{
  OS_DEVICE_OBJECT*   devObj = NULL;
  OS_DEVICE_RESOURCES resources;
  OS_DEVICE_DESC      devId;
  int                 index = (int) entry->driver_data;
  OS_STATUS           status;
  int                 res;
  int                 i, n;
  os_uint8            rev;

  os_dbgprint(1, ("init_board: entered, index = %d\n", index));

  pci_read_config_byte(dev, PCI_REVISION_ID, &rev);

  i = entry - pci_tbl;
#if 0
  if (drvObj.hardwareIds[i].id.pci.match & OS_DEVICE_DESC_MATCH_PCI_REV) {
    if (rev != drvObj.hardwareIds[i].id.pci.revId) {
      res = -EINVAL;
      os_dbgprint(0, ("*** init_board: PCI revision ID mismatch: wanted=0x%2.2lx actual=0x%2.2lx\n",
		      (unsigned long) drvObj.hardwareIds[i].id.pci.revId, (unsigned long) rev));
      goto failed;
    }
  }
#endif

  devObj = (OS_DEVICE_OBJECT*) os_kernel_malloc(sizeof(OS_DEVICE_OBJECT));
  if (devObj == NULL) {
    res = -ENOMEM;
    os_dbgprint(0, ("*** init_board: Cannot allocate devObj\n"));
    goto failed;
  }
  memset(devObj, 0, sizeof(OS_DEVICE_OBJECT));

  memset(&resources, 0, sizeof(resources));
  n = 0;
  for (i = 0; i < DEVICE_COUNT_RESOURCE; i++) {
    if (dev->resource[i].start >= dev->resource[i].end) {
      break;
    }
    resources.raw.pci.base[n] = (os_uint64) dev->resource[i].start;
    resources.raw.pci.size[n] = (os_uint64) (dev->resource[i].end - dev->resource[i].start + 1);
    resources.raw.pci.io[n] = !(dev->resource[i].flags & IORESOURCE_MEM);
    n++;
  }
  resources.raw.pci.count = n;
  resources.raw.pci.irq_line = dev->irq;

  for (i = 0; i < n; i++) {
    resources.translated.pci.base[i] = resources.raw.pci.base[i];
    resources.translated.pci.size[i] = resources.raw.pci.size[i];
    resources.translated.pci.io[i] = resources.raw.pci.io[i];
  }
  resources.translated.pci.count = resources.raw.pci.count;
  resources.translated.pci.vector = resources.raw.pci.irq_line;

  pci_enable_device(dev);
  pci_set_master(dev);

  devId.type = OS_DEVICE_DESC_TYPE_PCI;
  devId.id.pci.vId = dev->vendor;
  devId.id.pci.dId = dev->device;
  devId.id.pci.subVId = dev->subsystem_vendor;
  devId.id.pci.subDId = dev->subsystem_device;
  devId.id.pci.revId = rev;

  dev->driver_data = devObj;
  devObj->dev.pci = dev;
  devObj->drvObj = &drvObj;  
  devObj->deviceContext = NULL;
  os_init_spinlock(&devObj->cancelLock);
  os_init_spinlock(&devObj->dpcLock);
  devObj->maps.regions = NULL;

  status = drvObj.onCreateDevice(&drvObj, devObj, index, &devId);
  if (!OS_SUCCESS(status)) {
    os_dbgprint(0, ("init_board: onCreateDevice failed\n"));
    res = map_error(status);
    goto failed;
  }
  if (devObj->deviceContext == NULL) {
    os_dbgprint(0, ("init_board: warning: deviceContext is NULL\n"));
  }

  status = drvObj.onStartDevice(devObj, &resources);
  if (!OS_SUCCESS(status)) {
    os_dbgprint(0, ("init_board: onStartDevice failed\n"));
    res = map_error(status);
    goto failed;
  }

  /*
  ** Record this device in the array of device objects,
  ** so we can look it up using the minor device number
  */
  devices[drvObj.numDev++] = devObj;

  return 0;

 failed:
  pci_disable_device(dev);
  if (devObj != NULL) {
    os_kernel_free(devObj);
  }
  return res;
}
  
static void
close_board(
  struct pci_dev *dev)
{
  OS_DEVICE_OBJECT* devObj;

  os_dbgprint(1, ("close_board: entered\n"));

  devObj = (OS_DEVICE_OBJECT*) dev->driver_data;

  /* Stop the device - unmap BARs from kernel space etc. */
  drvObj.onStopDevice(devObj);

  /* Delete the device - free memory used for data structures etc. */
  drvObj.onDeleteDevice(devObj);

  pci_disable_device(dev);

  os_kernel_free(devObj);
}

void
cleanup_module(void)
{
  os_dbgprint(1, ("cleanup_module: entered\n"));

  if (majorValid) {
    unregister_chrdev(major, drvObj.drvName);
    majorValid = FALSE;
  }
  if (driverReged) {
    pci_unregister_driver(&driver);
    driverReged = FALSE;
  }

  if (drvObjInitted) {
    driver_uninit(&drvObj);
    drvObjInitted = FALSE;
  }

  if (pci_tbl != NULL) {
    os_kernel_free(pci_tbl);
    pci_tbl = NULL;
  }
}

int
init_module(void)
{
  int       i, n, err;
  OS_STATUS status;

  _os_dbglevel = 0;

  os_dbgprint(0, ("init_module: entered\n"));

  pci_tbl = NULL;
  drvObjInitted = FALSE;
  driverReged = FALSE;
  majorValid = FALSE;
  for (i = 0; i < MAX_DEVICE; i++) {
    devices[i] = NULL;
  }
  memset(&drvObj, 0, sizeof(drvObj));
  drvObj.numDev = 0;
  drvObj.drvName = "unnamed";

  status = driver_init(&drvObj);
  if (!OS_SUCCESS(status)) {
    err = map_error(status);
    os_dbgprint(0, ("*** init_module: driver_init() returned failure status\n"));
    goto failed;
  }
  drvObjInitted = TRUE;

  if (drvObj.drvName == NULL) {
    err = -ENODEV;
    os_dbgprint(0, ("init_module: No drvName specified\n"));
    goto failed;
  }
  if (drvObj.hardwareIds == NULL) {
    err = -ENODEV;
    os_dbgprint(0, ("init_module: No hardwareIds specified\n"));
    goto failed;
  }
  if (drvObj.onCreateDevice == NULL) {
    err = -ENODEV;
    os_dbgprint(0, ("init_module: No onCreateDevice handler specified\n"));
    goto failed;
  }
  if (drvObj.onStartDevice == NULL) {
    err = -ENODEV;
    os_dbgprint(0, ("init_module: No onStartDevice handler specified\n"));
    goto failed;
  }
  if (drvObj.onStopDevice == NULL) {
    err = -ENODEV;
    os_dbgprint(0, ("init_module: No onStopDevice handler specified\n"));
    goto failed;
  }
  if (drvObj.onDeleteDevice == NULL) {
    err = -ENODEV;
    os_dbgprint(0, ("init_module: No onDeleteDevice handler specified\n"));
    goto failed;
  }
  if (drvObj.onOpen == NULL) {
    os_dbgprint(0, ("init_module: No onOpen handler specified\n"));
  }
  if (drvObj.onClose == NULL) {
    os_dbgprint(0, ("init_module: No onClose handler specified\n"));
  }
  if (drvObj.onIoctl == NULL) {
    os_dbgprint(0, ("init_module: No onIoctl handler specified\n"));
  }

  for (n = 0;; n++) {
    if (drvObj.hardwareIds[n].type != OS_DEVICE_DESC_TYPE_NONE) {
      if (drvObj.hardwareIds[n].type != OS_DEVICE_DESC_TYPE_PCI) {
	n++;
      }
    } else {
      break;
    }
  }
  pci_tbl = os_kernel_malloc((n + 1) * sizeof(struct pci_device_id));
  if (pci_tbl == NULL) {
    err = -ENOMEM;
    os_dbgprint(0, ("init_module: Can't allocate pci_tbl\n"));
    goto failed;
  }
  memset(pci_tbl, 0, (n + 1) * sizeof(struct pci_device_id));
  for (i = 0; i < n; i++) {
    if (drvObj.hardwareIds[i].id.pci.match & OS_DEVICE_DESC_MATCH_PCI_VENDOR) {
      pci_tbl[i].vendor = drvObj.hardwareIds[i].id.pci.vId;
    } else {
      pci_tbl[i].vendor = PCI_ANY_ID;
    }
    if (drvObj.hardwareIds[i].id.pci.match & OS_DEVICE_DESC_MATCH_PCI_DEVICE) {
      pci_tbl[i].device = drvObj.hardwareIds[i].id.pci.dId;
    } else {
      pci_tbl[i].device = PCI_ANY_ID;
    }
    if (drvObj.hardwareIds[i].id.pci.match & OS_DEVICE_DESC_MATCH_PCI_SUBVENDOR) {
      pci_tbl[i].subvendor = drvObj.hardwareIds[i].id.pci.subVId;
    } else {
      pci_tbl[i].subvendor = PCI_ANY_ID;
    }
    if (drvObj.hardwareIds[i].id.pci.match & OS_DEVICE_DESC_MATCH_PCI_SUBDEVICE) {
      pci_tbl[i].subdevice = drvObj.hardwareIds[i].id.pci.subDId;
    } else {
      pci_tbl[i].subdevice = PCI_ANY_ID;
    }
    pci_tbl[i].driver_data = (unsigned long) i;
  }

  memset(&driver, 0, sizeof(driver));
  driver.name = drvObj.drvName;
  driver.id_table = pci_tbl;
  driver.probe = init_board;
  driver.remove = close_board;

  n = pci_register_driver(&driver);
  if (n == 0) {
    os_dbgprint(0, ("init_module: No devices found\n"));
    err = -ENODEV;
    goto failed;
  }
  driverReged = TRUE;

  memset(&fops, 0, sizeof(fops));
  fops.open = fops_open;
  fops.release = fops_release;
  fops.ioctl = fops_ioctl;
  fops.mmap = fops_mmap;

  SET_MODULE_OWNER(&fops);
  major = register_chrdev(0, drvObj.drvName, &fops);
  if (major < 0) {
    err = major;
    os_dbgprint(0, ("init_module: failed to get major\n"));
    goto failed;
  }
  majorValid = TRUE;
  os_dbgprint(1, ("init_module: major = %d\n", major));

  return 0;

 failed:
  cleanup_module();
  return err;
}
