/*
** ADM-XRC Series Driver
**
** (C) Copyright 2001-2004 Alpha Data
**
** loadfpga.c - IOCTL handlers for configuring/unconfiguring the FPGA.
**
** ----------------------------------------------------------------------------
*/

#include "platform.h"
#include "device.h"
#include "drvparam.h"
#include "dmacore.h"
#include "dmadesc.h"
#include "ioctl.h"
#include "loadfpga.h"

static OS_STATUS
clearFPGA(
  DeviceContext* pDevCtx)
{
  OS_STATUS     status;
  unsigned long timeout;
  unsigned long now;
  os_boolean    success;
  os_boolean    bInterrupt;
  unsigned long delay;
  os_boolean    bVal;

  /*
  ** Pulse PROG for 20 usec - this will cause DONE to
  ** become deasserted and INIT to (eventually) become
  ** deasserted.
  */
  bVal = TRUE;
  pDevCtx->methods.fpga(pDevCtx, MethodFpgaOp_SetProg, &bVal);
  os_delay_busy_us(20);
  bVal = FALSE;
  pDevCtx->methods.fpga(pDevCtx, MethodFpgaOp_SetProg, &bVal);
  
  /*
  ** Delay until DONE is deasserted.
  **
  ** We allow a timeout of 1s.
  */
  timeout = os_get_time() + os_ticks_per_second;
  success = FALSE;
  while (1) {
    pDevCtx->methods.fpga(pDevCtx, MethodFpgaOp_GetDone, &bVal);
    if (!bVal) {
      success = TRUE;
      break;
    }
    
    now = os_get_time();
    if (now >= timeout) {
      /* Timeout occurred */
      break;
    }

    delay = 10;
    bInterrupt = os_delay_thread(TRUE, &delay);
    if (bInterrupt) {
      os_dbgprint(3, ("clearFPGA: interrupt while waiting for DONE deasserted\n"));
      status = OS_MAKE_STATUS(ADMXRC_CANCELLED);
      goto done;
    }
  }
  
  if (!success) {
    status = OS_MAKE_STATUS(ADMXRC_FAILED);
    goto done;
  }

  /*
  ** Delay until INIT is deasserted.
  **
  ** We allow a timeout of 1s.
  */
  timeout = os_get_time() + os_ticks_per_second;
  success = FALSE;
  while (1) {
    pDevCtx->methods.fpga(pDevCtx, MethodFpgaOp_GetInit, &bVal);
    if (!bVal) {
      success = TRUE;
      break;
    }
    
    now = os_get_time();
    if (now >= timeout) {
      /* Timeout occurred */
      break;
    }

    delay = 10;
    bInterrupt = os_delay_thread(TRUE, &delay);
    if (bInterrupt) {
      os_dbgprint(3, ("clearFPGA: interrupt while waiting for INIT deasserted\n"));
      status = OS_MAKE_STATUS(ADMXRC_CANCELLED);
      goto done;
    }
  }
  
  if (!success) {
    status = OS_MAKE_STATUS(ADMXRC_FAILED);
    goto done;
  }
  
  status = OS_STATUS_SUCCESS;
  
 done:
  return status;
}

static OS_STATUS
checkConfigured(
  DeviceContext* pDevCtx)
{
  OS_STATUS     status = OS_STATUS_SUCCESS;
  os_boolean    success;
  unsigned long timeout;
  unsigned long now;
  os_boolean    bInterrupt;
  unsigned long delay;
  os_boolean    bVal;

  /*
  ** Wait for device to assert DONE, which may rely on DLL's/DCM's.
  **
  ** We allow a timeout of 1s.
  */
  timeout = os_get_time() + os_ticks_per_second;
  success = FALSE;
  while (1)	{
    pDevCtx->methods.fpga(pDevCtx, MethodFpgaOp_GetDone, &bVal);
    if (bVal) {
      success = TRUE;
      break;
    }
    
    now = os_get_time();
    if (now >= timeout) {
      /* Timeout occurred */
      break;
    }

    delay = 10;
    bInterrupt = os_delay_thread(TRUE, &delay);
    if (bInterrupt) {
      os_dbgprint(3, ("checkConfigured: interrupt while waiting for DONE asserted\n"));
      status = OS_MAKE_STATUS(ADMXRC_CANCELLED);
      goto done;
    }
  }
  
  if (!success) {
    pDevCtx->methods.fpga(pDevCtx, MethodFpgaOp_GetInit, &bVal);
    if (bVal) {
      os_dbgprint(3, ("checkConfigured: Init indicates failure to load\n"));
    } else {
      os_dbgprint(3, ("checkConfigured: Done did not go high\n"));
    }
    
    /*
    ** Something bad happened, so clear the FPGA in case
    ** it has been loaded with a bad bitstream.
    */
    bVal = TRUE;
    pDevCtx->methods.fpga(pDevCtx, MethodFpgaOp_SetProg, &bVal);
    os_delay_busy_us(20);
    bVal = FALSE;
    pDevCtx->methods.fpga(pDevCtx, MethodFpgaOp_SetProg, &bVal);
    status = OS_MAKE_STATUS(ADMXRC_FAILED);
  }
  
 done:
  return status;
}

/*
** IOCTL handlers
*/

OS_STATUS
ADMXRC_LoadFPGA(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* req)
{
  os_uint8*               data;
  unsigned long           length;
  unsigned long           count;
  OS_STATUS               status;
  OS_SPINLOCK_FLAGS       f;

  os_dbgprint(3, ("ADMXRC_LoadFPGA: entered\n"));

  data = (os_uint8*) req->param.massBuf;
  length = req->param.massSize;

  os_dbgprint(3, ("ADMXRC_LoadFPGA: data = 0x%p length = %ld\n", (void*) data, length));

  os_acquire_spinlock(&pDevCtx->fpga.lock, &f);
  if (pDevCtx->fpga.busy) {
    os_release_spinlock(&pDevCtx->fpga.lock, f);
    return OS_MAKE_STATUS(ADMXRC_DEVICE_BUSY);
  } else {
    pDevCtx->fpga.busy = TRUE;
    os_release_spinlock(&pDevCtx->fpga.lock, f);
  }
  
  status = clearFPGA(pDevCtx);
  if (!OS_SUCCESS(status)) {
    os_dbgprint(3, ("ADMXRC_LoadFPGA: failed to clear FPGA\n"));
    goto done;
  }
  
  os_dbgprint(3, ("ADMXRC_LoadFPGA: cleared FPGA\n"));

  /* Send the bitstream to the FPGA */
  count = length;
  for (count = 0; count < length; count++) {
    pDevCtx->methods.selectMap(pDevCtx, TRUE, data++);
  }
  
  status = checkConfigured(pDevCtx);
  if (!OS_SUCCESS(status)) {
    os_dbgprint(3, ("ADMXRC_LoadFPGA: configuration failed\n"));
    goto done;
  }
  
  os_dbgprint(3, ("ADMXRC_LoadFPGA: success\n"));

  status = OS_STATUS_SUCCESS;
  
 done:
  os_acquire_spinlock(&pDevCtx->fpga.lock, &f);
  pDevCtx->fpga.busy = FALSE;
  os_release_spinlock(&pDevCtx->fpga.lock, f);
  
  return status;
}

static void
loadFPGADMACallback(
  DeviceContext*  devCtx,
  ADMXRC_DMANODE* node)
{
  OS_EVENT* event;
  
  os_dbgprint(6, ("loadFPGADMACallback\n"));
  
  event = (OS_EVENT*) node->client.context;
  os_signal_event(event);
}

OS_STATUS
ADMXRC_LoadFPGADMA(
  DeviceContext* devCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_LOADFPGADMA* iptr;
  void*                      buffer;
  unsigned long              length;
  OS_STATUS                  status;
  OS_SPINLOCK_FLAGS          f;
  ADMXRC_LOADFPGADMA*        ctl = NULL;
  ADMXRC_DMANODE*            dmanode = NULL;
  os_uint32                  mode;
  os_uint32                  local;
  unsigned long              flags;
  unsigned char              channel, queue;
  os_boolean                 locked = FALSE;
  OS_EVENT                   event;
  os_boolean                 must_start;

  os_dbgprint(3, ("ADMXRC_LoadFPGADMA: entered\n"));

  iptr = (ADMXRC_IOCTLS_LOADFPGADMA*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(*iptr)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  flags = iptr->in.flags;
  if (flags & ~(ADMXRC_DMAFLAG_TIMEOUT | ADMXRC_DMAFLAG_DONOTQUEUE)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  flags &= ~ADMXRC_DMAFLAG_TIMEOUT;

  channel = iptr->in.channel;
  if (channel != ADMXRC_DMACHAN_ANY && channel >= devCtx->info.cardInfo.NumDMAChan) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  length = iptr->in.length;
  buffer = iptr->in.buffer;
  if (buffer == NULL) {
    return OS_MAKE_STATUS(ADMXRC_NULL_POINTER);
  }

  ctl = os_alloc_lcache(devCtx->fpga.ctlPool, 0);
  if (ctl == NULL) {
    status = OS_MAKE_STATUS(ADMXRC_NO_MEMORY);
    goto completed;
  }

  os_acquire_spinlock(&devCtx->fpga.lock, &f);
  if (devCtx->fpga.busy) {
    os_release_spinlock(&devCtx->fpga.lock, f);
    status = OS_MAKE_STATUS(ADMXRC_DEVICE_BUSY);
    goto completed;
  } else {
    locked = TRUE;
    devCtx->fpga.busy = TRUE;
    os_release_spinlock(&devCtx->fpga.lock, f);
  }

  status = clearFPGA(devCtx);
  if (!OS_SUCCESS(status)) {
    goto completed;
  }

  /* Determine which DMA queue to add this node to */
  queue = (channel == ADMXRC_DMACHAN_ANY) ? devCtx->info.cardInfo.NumDMAChan : (unsigned char) channel;
  
  /* Calculate local bus address of SelectMap port and DMA mode */
  if (devCtx->devDesc.id.pci.subDId == PCI_SUBDEVICE_ADATA_ADMXRC2PL_PCIX ||
      devCtx->devDesc.id.pci.subDId == PCI_SUBDEVICE_ADATA_ADMXRC2PL_PCI ||
      devCtx->devDesc.id.pci.subDId == PCI_SUBDEVICE_ADATA_ADMXRC2P_PCIX ||
      devCtx->devDesc.id.pci.subDId == PCI_SUBDEVICE_ADATA_ADMXRC2P_PCI) {
    local = devCtx->info.spaceInfo[1].LocalBase + 0x10;
    mode = 0x000009c2U;
  } else {
    local = devCtx->info.spaceInfo[1].LocalBase + 0x8;
    mode = 0x000009c0U;
  }
  
  dmanode = &ctl->dmanode;
  dmanode->client.context = (void*) &event;
  dmanode->client.callback = loadFPGADMACallback;
  dmanode->client.direction = (unsigned char) ADMXRC2_PCITOLOCAL;
  dmanode->client.length = length;
  dmanode->client.pos = 0;
  dmanode->client.mode = mode;
  dmanode->client.queue = queue;
  
  os_dbgprint(6, ("ADMXRC_LoadFPGADMA: enqueuing DMANODE 0x%p buffer=0x%p len=0x%ld loc=0x%lx dir=%ld chan=%ld mode=0x%lx flags=0x%lx q=%ld\n",
		  (void*) dmanode,
		  (void*) buffer,
		  (unsigned long) length,
		  (unsigned long) local,
		  (unsigned long) ADMXRC2_PCITOLOCAL,
		  (unsigned long) channel,
		  (unsigned long) mode,
		  (unsigned long) flags,
		  (unsigned long) queue));

  status = OS_STATUS_SUCCESS;
  req->user.context = (void*) ctl;
  os_init_event(&event);

  must_start = (flags & ADMXRC_DMAFLAG_DONOTQUEUE) != 0;
  if (ADMXRC_EnqueueDma(devCtx, dmanode, must_start)) {
    goto completed;
  }

  while (1) {
    os_boolean    signalled;
    unsigned long spanned;

    signalled = os_wait_event(&event, TRUE);
    if (signalled) {
      if (ADMXRC_CancelDma(devCtx, dmanode)) {
	goto completed;
      }
      os_wait_event(&event, FALSE);
    }

    switch (dmanode->status.reason) {
    case ADMXRC_DMACALLBACK_REASON_SUBMIT:
      {
	ADMXRC_SubmitDmaUser(devCtx, dmanode, buffer, length, local, &spanned);
	length -= spanned;
	buffer = ((os_uint8*) buffer) + spanned;
	break;
      }

    case ADMXRC_DMACALLBACK_REASON_COLLECT:
      {
	ADMXRC_CollectDma(devCtx, dmanode);
	break;
      }

    case ADMXRC_DMACALLBACK_REASON_COMPLETED:
      goto completed;
      break;

    default:
      os_assert(FALSE);
      break;
    }

    os_reset_event(&event);
    if (ADMXRC_ContinueDma(devCtx, dmanode)) {
      goto completed;
    }
  }

 completed:
  status = ctl->dmanode.status.status;
  if (OS_SUCCESS(status)) {
    status = checkConfigured(devCtx);
  }

  if (locked) {
    os_acquire_spinlock(&devCtx->fpga.lock, &f);
    devCtx->fpga.busy = FALSE;
    os_release_spinlock(&devCtx->fpga.lock, f);
  }
  
  if (ctl != NULL) {
    os_free_lcache(devCtx->fpga.ctlPool, ctl);
  }

  return status;
}
