/*
** ADM-XRC Series Linux Driver
**
** (C) Copyright 2001-2004 Alpha Data
**
** platform-2.4.c - Platform abstraction layer routines for Linux 2.4.x
**
** This module contains routines that hide the underlying operating system
** and platform from the generic section of the driver, in order to make the
** driver more portable across various operating systems and platforms.
**
** ----------------------------------------------------------------------------
*/

#include "platform.h"

#if !defined(OS_ALLOC_KIOVEC_SZ)
#   define alloc_kiovec_sz(x, y, z) alloc_kiovec(x, y)
#   define free_kiovec_sz(x, y, z) free_kiovec(x, y)
#endif

#if (LINUX_VERSION_CODE >= 0x020400)
#   define _os_page_is_high(k, i) PageHighMem((k)->maplist[i]);
#else
#   define _os_page_is_high(k, i) PageHighMem((k)->pagelist[i]);
#endif

int _os_dbglevel = 0;          /* no debugging messages by default          */

static void
_os_isr(
  int             irq,
  void*           context,
  struct pt_regs* regs)
{
  OS_ATTACHED_IRQ* irqObj = (OS_ATTACHED_IRQ*) context;

  irqObj->isr(irqObj->deviceContext);
}

static void
_os_dpc_handler(
  unsigned long context)
{
  OS_DEVICE_OBJECT* devObj;
  OS_DPC*           dpc = (OS_DPC*) context;
  OS_SPINLOCK_FLAGS f;
  void*             arg;
  void*             dpc_context;
  OS_DPC_FUNC*      handler;
  os_boolean        destroy;

  devObj = dpc->devObj;
  os_acquire_spinlock(&devObj->dpcLock, &f);
  dpc->scheduled = FALSE;
  handler = dpc->handler;
  dpc_context = dpc->context;
  arg = dpc->argument;
  destroy = dpc->destroy;
  os_release_spinlock(&devObj->dpcLock, f);

  handler(dpc_context, arg);

  if (destroy) {
    wake_up(dpc->wake);
  }
}

/*
** -------------------------------------------------------------------
** Exported to client code
** -------------------------------------------------------------------
*/

OS_STATUS
os_allow_mapping(
  OS_DEVICE_OBJECT*    devObj,
  os_uint64            phys,
  os_uint64            length,
  void*                context,
  OS_MAPPABLE_REGION** handle)
{
  OS_MAPPABLE_REGION* region;

  os_dbgprint(3, ("os_allow_mapping: phys = 0x%8.8lx%8.8lx length = 0x%8.8lx%8.8lx context = 0x%p\n",
		  (unsigned long) os_uint64_hi(phys),
		  (unsigned long) os_uint64_lo(phys),
		  (unsigned long) os_uint64_hi(length),
		  (unsigned long) os_uint64_lo(length),
		  context));

  region = (OS_MAPPABLE_REGION*) os_kernel_malloc(sizeof(OS_MAPPABLE_REGION));
  if (region == NULL) {
    return OS_STATUS_NOMEMORY;
  }

  region->devObj = devObj;
  region->phys = phys;
  region->length = length;
  region->context = context;
  region->next = devObj->maps.regions;
  region->prev = &devObj->maps.regions;
  devObj->maps.regions = region;

  os_dbgprint(3, ("os_allow_mapping: created region 0x%p\n", region));

  *handle = region;

  return OS_STATUS_SUCCESS;
}

void
os_complete_request(
  OS_IO_REQUEST* req)
{
  OS_DEVICE_OBJECT* devObj;
  OS_SPINLOCK_FLAGS f;

  os_dbgprint(4, ("os_complete_request: req = 0x%p, status = %lx, length = %ld\n",
		  (void*) req, (unsigned long) req->status.status, (unsigned long) req->status.length));

  devObj = req->opaque.devObj;
  os_acquire_spinlock(&devObj->cancelLock, &f);
  req->opaque.completed = TRUE;
  wake_up(req->opaque.wait_q);
  os_release_spinlock(&devObj->cancelLock, f);
}

OS_STATUS
os_create_usbd(
  void*         user,
  unsigned long length,
  OS_USBD**     usbd)
{
  OS_STATUS      status;
  OS_USBD*       u = NULL;
  int            err;
  unsigned int   nmapped = 0, i, nkiobuf;
  os_boolean     kiobufsAlloced = FALSE;
  unsigned long  userl, chunk, page_addr, page_limit_addr;

  u = (OS_USBD*) os_kernel_malloc(sizeof(OS_USBD));
  if (u == NULL) {
    return OS_STATUS_NOMEMORY;
  }

  userl = (unsigned long) user;
  page_addr = userl & os_page_mask_high;
  page_limit_addr = (userl + length + os_page_size - 1) & os_page_mask_high;
  nkiobuf = (page_limit_addr - page_addr + OS_KIOBUFCHUNK - 1) >> OS_KIOBUFSHIFT;

  u->user = user;
  u->length = length;
  u->nkiobuf = nkiobuf;
  u->kio = NULL;
  u->nbhs = NULL;

  u->kio = os_kernel_malloc(nkiobuf * sizeof(struct kiobuf*));
  u->nbhs = os_kernel_malloc(nkiobuf * sizeof(int));
  for (i = 0; i < nkiobuf; i++) {
    u->nbhs[i] = KIO_MAX_SECTORS;
  }

  err = alloc_kiovec_sz(nkiobuf, u->kio, u->nbhs);
  if (err) {
    status = OS_STATUS_NOMEMORY;
    goto failed;
  }
  kiobufsAlloced = TRUE;

  for (i = 0; i < nkiobuf; i++) {
    page_addr = userl & os_page_mask_high;
    page_limit_addr = page_addr + OS_KIOBUFCHUNK;
    chunk = page_limit_addr - userl;
    if (chunk > length) {
      chunk = length;
    }
    err = map_user_kiobuf(READ, u->kio[i], userl, chunk);
    if (err) {
      status = OS_STATUS_UNSUCCESSFUL;
      goto failed;
    }
    userl += chunk;
    nmapped++;
  }

  *usbd = u;

  return OS_STATUS_SUCCESS;

 failed:
  if (u != NULL) {
    for (i = 0; i < nmapped; i++) {
      unmap_kiobuf(u->kio[i]);
    }
    if (kiobufsAlloced) {
      free_kiovec_sz(nkiobuf, u->kio, u->nbhs);
    }
    if (u->kio != NULL) {
      os_kernel_free(u->kio);
    }
    if (u->nbhs != NULL) {
      os_kernel_free(u->nbhs);
    }
    os_kernel_free(u);
  }
  return status;
}

os_boolean
os_delay_thread(
  os_boolean     interruptible,
  unsigned long* ms)
{
  long          start;
  long          elapsed;
  unsigned long n = (*ms) * os_ticks_per_second / 1000;

  if (interruptible) {
    set_current_state(TASK_INTERRUPTIBLE);
  } else {
    set_current_state(TASK_UNINTERRUPTIBLE);
  }
  if (!n) {
    /* Don't allow a wait of zero ticks */
    n = 1;
  }

  start = jiffies;
  schedule_timeout(n);

  /* Calculate remaining time */
  elapsed = jiffies - start;
  if (elapsed >= n) {
    *ms = 0;
  } else {
    *ms = n - (elapsed * 1000 / os_ticks_per_second);
  }

  /* return TRUE if interrupted by a signal, otherwise FALSE */
  return (os_boolean) signal_pending(current);
}

void
os_destroy_usbd(
  OS_USBD* usbd)
{
  unsigned int i, n;

  n = usbd->nkiobuf;
  for (i = 0; i < n; i++) {
    unmap_kiobuf(usbd->kio[i]);
  }
  free_kiovec_sz(n, usbd->kio, usbd->nbhs);
  os_kernel_free(usbd->nbhs);
  os_kernel_free(usbd->kio);
  os_kernel_free(usbd);
}

void
os_disallow_mapping(
  OS_DEVICE_OBJECT*   devObj,
  OS_MAPPABLE_REGION* region)
{
  os_dbgprint(3, ("os_disallow_mapping: region = 0x%p\n", region));

  *region->prev = region->next;

  os_kernel_free(region);
}

void
os_flush_io_buffer(
  OS_USBD*      usbd,
  unsigned long offset,
  unsigned long length,
  os_boolean    write)
{
#ifdef CONFIG_NONCOHERENT_IO
  /*
  ** This is necessary because MIPS D-cache uses partial virtual address
  ** to generate cache line index; hence possibility of cache line aliasing
  ** where two or more cache lines may reference the same physical memory
  ** locations. To get around this problem we just flush the entire D-cache.
  */
  flush_cache_all();

#ifdef CONFIG_BOARD_SCACHE
  {
    unsigned long chunk;
    unsigned long p_idx, p_offs;
    os_uint8* kAddr;
    struct page* p;
    struct kiobuf* kio;
    struct kiobuf** kiovp;
    unsigned int kio_idx;

    /*
    ** Make secondary cache coherent.
    **
    ** On MIPS, secondary cache is physically indexed; hence cache line
    ** aliasing is not a problem and we can selectively write-back & invalidate
    ** just the cache lines that are part of the buffer.
    */
    offset += usbd->kio[0]->offset;
    kio_idx = offset >> OS_KIOBUFSHIFT;
    kiovp = &usbd->kio[kio_idx];
    kio = *kiovp++;
    offset &= OS_KIOBUFMASKLO;
    p_idx = offset >> os_page_shift;
    p_offs = offset & os_page_mask_low;
    while (length) {   
      if (p_idx == OS_KIOBUFMAXPAGE) {
	p_idx = 0;
	kio = *kiovp++;
      }
      p = kio->maplist[p_idx];
      chunk = os_page_size - p_offs;
      if (chunk > length) {
	chunk = length;
      }
      kAddr = ((os_uint8*) page_address(p)) + p_offs;
      bc_wback_inv(kAddr, chunk);
      length -= chunk;
      p_idx++;
      p_offs = 0;
    }
  }
#endif /* def CONFIG_BOARD_SCACHE */
#endif /* def CONFIG_NONCOHERENT_IO */
}

void*
os_invoke_atomic(
  OS_ATTACHED_IRQ* obj,
  OS_INVOKE_FUNC*  function,
  void*            context)
{
  void*             ret;

  disable_irq(obj->vector);
  ret = function(context);
  enable_irq(obj->vector);

  return ret;
}

os_uint64
os_kernel_to_bus(
  OS_DEVICE_OBJECT* devObj,
  void*             kern)
{
  return (os_uint64) ((unsigned long) virt_to_bus(kern));
}

void*
os_map_kernel(
  os_uint64 phys,
  os_uint64 length)
{
  void* ptr;

  if (!length) {
    os_dbgprint(0, ("os_map_kernel: attempt to map zero-length region\n"));
  }

  ptr = ioremap_nocache(phys, length);
  
  return ptr;
}

void
os_unmap_kernel(
  void*     kern,
  os_uint64 phys,
  os_uint64 length)
{
  if (kern == NULL) {
    os_dbgprint(0, ("os_unmap_kernel: NULL pointer\n"));
  }

  iounmap(kern);
}

os_boolean
os_set_cancel_routine(
  OS_IO_REQUEST*    req,
  OS_IO_CANCELFUNC* func)
{
  OS_SPINLOCK_FLAGS f;

  os_acquire_spinlock(&req->opaque.devObj->cancelLock, &f);
  if (req->opaque.cancel) {
    /*
    ** Cannot change the cancel routine, as if set, it may be already
    ** running or about to run
    */
    os_release_spinlock(&req->opaque.devObj->cancelLock, f);
    return FALSE;
  } else {
    /*
    ** Can successfully change the cancel routine as if set, it's not
    ** running yet.
    */
    req->opaque.cancelFunc = func;
    os_release_spinlock(&req->opaque.devObj->cancelLock, f);
    return TRUE;
  }
}

/*
** --------------------------------------------------------------
** Scatter-gather functions
** --------------------------------------------------------------
*/

/*
** These flags indicate whether the DMA targets the buffer directly (1), or
** targets a bounce buffer (0)
*/
#define OS_SG_BUFF_DIRECT (0x1U << 0)

/*
** These flags only make sense if DMA targets a bounce buffer, and indicate
** whether the DMA transfer is targetting a kernel-space buffer (0) or a
** user-space buffer (1). The distinction is required because there may be
** a memory copy operation before or after the DMA transfer.
*/
#define OS_SG_BUFF_USER   (0x1U << 1)

OS_SG_MAP*
os_sg_alloc(
  OS_DEVICE_OBJECT* pDevObj,
  unsigned int nEntry)
{
  OS_SG_MAP* map;
  unsigned int i;

  map = os_kernel_malloc(sizeof(OS_SG_MAP));
  if (map == NULL) {
    return NULL;
  }
  map->pDevObj = pDevObj;
  map->maxBuff = nEntry;
  map->bounce = NULL;
  map->buffInfo = NULL;

  map->bounce = os_kernel_malloc(nEntry * sizeof(void*));
  if (map->bounce == NULL) {
    os_sg_free(map);
    return NULL;
  }
  for (i = 0; i < nEntry; i++) {
    map->bounce[i] = NULL;
  }
  for (i = 0; i < nEntry; i++) {
    map->bounce[i] = os_kernel_alloc_page_dma();
    if (map->bounce[i] == NULL) {
      os_sg_free(map);
      return NULL;
    }
  }

  map->buffInfo = os_kernel_malloc(nEntry * sizeof(map->buffInfo[0]));
  if (map->buffInfo == NULL) {
    os_sg_free(map);
    return NULL;
  }

  return map;
}

void
os_sg_free(
  OS_SG_MAP* map)
{
  int i;

  if (map->bounce != NULL) {
    for (i = 0; i < map->maxBuff; i++) {
      if (map->bounce[i] != NULL) {
	os_kernel_free_page_dma(map->bounce[i]);
      }
    }
    os_kernel_free(map->bounce);
  }
  if (map->buffInfo != NULL) {
    os_kernel_free(map->buffInfo);
  }
  os_kernel_free(map);
}

extern unsigned long
os_sg_add_kernel_buffer(
  OS_SG_MAP* map,
  os_uint32 local,
  os_boolean bFixedLocal,
  void* buffer,
  unsigned int length)
{
  int nBuff = map->nBuff;
  int maxBuff = map->maxBuff;
  unsigned long bytesMapped = 0;
  unsigned long p_offs;

  p_offs = ((unsigned long) buffer) & os_page_mask_low;
  while (length && nBuff < maxBuff) {
    unsigned long chunk = os_page_size - p_offs;

    if (chunk > length) {
      chunk = length;
    }
    map->buffInfo[nBuff].kernel = buffer;
    map->buffInfo[nBuff].local = local;
    map->buffInfo[nBuff].length = chunk;
    map->buffInfo[nBuff].type = OS_SG_BUFF_DIRECT;
    p_offs = 0;
    buffer = (void*) (((os_uint8*) buffer) + chunk);
    nBuff++;
    length -= chunk;
    bytesMapped += chunk;
    if (!bFixedLocal) {
      local += chunk;
    }
  }
  map->nBuff = nBuff;

  return bytesMapped;
}

unsigned long
os_sg_add_usbd_buffer(
  OS_SG_MAP* map,
  os_uint32 local,
  os_boolean bFixedLocal,
  OS_USBD* usbd,
  unsigned long offset,
  unsigned long length)
{
  struct page* p;
  unsigned int kio_idx, p_idx, p_offs;
  os_uint8* user;
  int nBuff = map->nBuff;
  int nBounceUsed = map->nBounceUsed;
  int maxBuff = map->maxBuff;
  os_boolean bDmaWrite = map->bDmaWrite;
  unsigned long nBytesMapped = 0;
  struct kiobuf* kio;
  struct kiobuf** kiovp;

  user = ((os_uint8*) usbd->user) + offset;
  offset += usbd->kio[0]->offset;
  kio_idx = offset >> OS_KIOBUFSHIFT;
  kiovp = &usbd->kio[kio_idx];
  kio = *kiovp++;
  offset &= OS_KIOBUFMASKLO;
  p_idx = offset >> os_page_shift;
  p_offs = offset & os_page_mask_low;
  while (length && nBuff < maxBuff) {
    unsigned long chunk;
    os_boolean bDirect;
    void* kAddr;

    if (p_idx == OS_KIOBUFMAXPAGE) {
      p_idx = 0;
      kio = *kiovp++;
    }
    chunk = os_page_size - p_offs;
    if (chunk > length) {
      chunk = length;
    }
    p = kio->maplist[p_idx];

#ifdef CONFIG_MIPS
    /* Force use of bounce buffer */
    bDirect = FALSE;
#else /* def CONFIG_MIPS */
    bDirect = !PageHighMem(p);
#endif /* ndef CONFIG_MIPS */

    if (bDirect) {
      /* No bounce buffer required for this page */
      map->buffInfo[nBuff].kernel = ((os_uint8*) page_address(p)) + p_offs;
      map->buffInfo[nBuff].type = OS_SG_BUFF_DIRECT;
    } else {
      /* Bounce buffer required for this page */
      kAddr = map->bounce[nBounceUsed++];
      map->buffInfo[nBuff].kernel = kAddr;
      map->buffInfo[nBuff].actual = user;
      map->buffInfo[nBuff].type = OS_SG_BUFF_USER;
      if (!bDmaWrite) {
	copy_from_user(kAddr, user, chunk);
      }
    }
    map->buffInfo[nBuff].local = local;
    map->buffInfo[nBuff].length = chunk;

    nBuff++;
    p_offs = 0;
    p_idx++;
    length -= chunk;
    nBytesMapped += chunk;
    user += chunk;
    if (!bFixedLocal) {
      local += chunk;
    }
 }

  map->nBuff = nBuff;
  map->nBounceUsed = nBounceUsed;

  return nBytesMapped;
}

extern unsigned long
os_sg_add_user_buffer(
  OS_SG_MAP* map,
  os_uint32 local,
  os_boolean bFixedLocal,
  void* uBuff,
  unsigned int length)
{
  int nBuff = map->nBuff;
  int nBounceUsed = map->nBounceUsed;
  int maxBuff = map->maxBuff;
  unsigned long bytesMapped = 0;
  unsigned long p_offs;
  os_boolean bDmaWrite = map->bDmaWrite;

  p_offs = ((unsigned long) uBuff) & os_page_mask_low;
  while (length && nBuff < maxBuff) {
    unsigned long chunk = os_page_size - p_offs;
    void* kAddr;
    
    if (chunk > length) {
      chunk = length;
    }
    kAddr = map->bounce[nBounceUsed++];
    map->buffInfo[nBuff].kernel = kAddr;
    map->buffInfo[nBuff].local = local;
    map->buffInfo[nBuff].length = chunk;
    map->buffInfo[nBuff].type = OS_SG_BUFF_USER;
    if (!bDmaWrite) {
      copy_from_user(kAddr, uBuff, chunk);
    }
    p_offs = 0;
    uBuff = (void*) (((os_uint8*) uBuff) + chunk);
    nBuff++;
    length -= chunk;
    bytesMapped += chunk;
    if (!bFixedLocal) {
      local += chunk;
    }
  }
  map->nBuff = nBuff;
  map->nBounceUsed = nBounceUsed;

  return bytesMapped;
}

unsigned int
os_sg_map(
  OS_SG_MAP* map)
{
  unsigned int i, n;
  os_boolean bDmaWrite = map->bDmaWrite;

  n = map->nBuff;
  for (i = 0; i < n; i++) {
    map->buffInfo[i].logical =
      pci_map_single(map->pDevObj->dev.pci,
		     map->buffInfo[i].kernel,
		     map->buffInfo[i].length,
		     bDmaWrite ? PCI_DMA_FROMDEVICE : PCI_DMA_TODEVICE);
  }

  return n;
}

void
os_sg_unmap(
  OS_SG_MAP* map)
{
  int i, n = map->nBuff;
  os_boolean bDmaWrite = map->bDmaWrite;

  for (i = 0; i < n; i++) {
    pci_unmap_single(map->pDevObj->dev.pci,
		     (dma_addr_t) map->buffInfo[i].logical,
		     map->buffInfo[i].length,
		     bDmaWrite ? PCI_DMA_FROMDEVICE : PCI_DMA_TODEVICE);
    if (bDmaWrite) {
      int type =  map->buffInfo[i].type;
      
      if (!(type & OS_SG_BUFF_DIRECT)) {
	if (type & OS_SG_BUFF_USER) {
	  copy_to_user(map->buffInfo[i].actual,
		       map->buffInfo[i].kernel,
		       map->buffInfo[i].length);
	} else {
	  memcpy(map->buffInfo[i].kernel,
		 map->buffInfo[i].actual,
		 map->buffInfo[i].length);
	}
      }
    }
  }
}

/*
** --------------------------------------------------------------
** Event functions
** --------------------------------------------------------------
*/

void
os_init_event(
  OS_EVENT* event)
{
  os_init_spinlock(&event->lock);
  init_waitqueue_head(&event->q);
  event->bRaised = FALSE;
}

void
os_signal_event(
  OS_EVENT* event)
{
  OS_SPINLOCK_FLAGS f;

  os_acquire_spinlock(&event->lock, &f);
  os_assert(!event->bRaised);
  event->bRaised = TRUE;
  wake_up(&event->q);
  os_release_spinlock(&event->lock, f);
}

void
os_reset_event(
  OS_EVENT* event)
{
  OS_SPINLOCK_FLAGS f;

  os_acquire_spinlock(&event->lock, &f);
  event->bRaised = FALSE;
  os_release_spinlock(&event->lock, f);
}

os_boolean
os_wait_event(
  OS_EVENT*  event,
  os_boolean interruptible)
{
  OS_SPINLOCK_FLAGS f;
  wait_queue_t      wait;

  os_acquire_spinlock(&event->lock, &f);
  if (event->bRaised) {
    os_release_spinlock(&event->lock, f);
    return FALSE;
  }
  init_waitqueue_entry(&wait, current);
  set_current_state(interruptible ? TASK_INTERRUPTIBLE : TASK_UNINTERRUPTIBLE);
  add_wait_queue(&event->q, &wait);
  os_release_spinlock(&event->lock, f);

  schedule();
  remove_wait_queue(&event->q, &wait);

#if 0
  os_acquire_spinlock(&event->lock, &f);
  remove_wait_queue(&event->q, &wait);
  os_release_spinlock(&event->lock, f);
#endif

  if (interruptible) {
    if (signal_pending(current)) {
      /* Caught a signal */
      return TRUE;
    }
  }

  /* No signal caught */
  return FALSE;
}

/*
** --------------------------------------------------------------
** Interrupt functions
** --------------------------------------------------------------
*/

OS_STATUS
os_attach_interrupt(
  OS_DEVICE_OBJECT* devObj,
  OS_ISR*           isr,
  long              vector,
  const char*       devName,
  OS_ATTACHED_IRQ** o)
{
  OS_ATTACHED_IRQ* obj;
  int              err;

  obj = os_kernel_malloc(sizeof(OS_ATTACHED_IRQ));
  if (obj == NULL) {
    return OS_STATUS_NOMEMORY;
  }
  obj->vector = vector;
  obj->isr = isr;
  obj->devObj = devObj;
  obj->deviceContext = devObj->deviceContext;

  /* SA_INTERRUPT => fast handler, so interrupts are disabled in ISR */
  err = request_irq(obj->vector, _os_isr, SA_SHIRQ | SA_INTERRUPT, devName, (void*) obj);
  if (err == 0) {
    *o = obj;
    return OS_STATUS_SUCCESS;
  } else {
    os_kernel_free(obj);
    if (err == -EBUSY) {
      return OS_STATUS_CANNOTSHARE;
    } else {
      return OS_STATUS_UNSUCCESSFUL;
    }
  }
}

void
os_detach_interrupt(
  OS_ATTACHED_IRQ* obj)
{
  free_irq(obj->vector, obj);
  os_kernel_free(obj);
}

/*
** --------------------------------------------------------------
** DPC functions
** --------------------------------------------------------------
*/

OS_STATUS
os_create_dpc(
  OS_DEVICE_OBJECT* devObj,
  OS_DPC_FUNC*      handler,
  void*             context,
  OS_DPC**          dpc)
{
  OS_DPC* p;

  p = (OS_DPC*) os_kernel_malloc(sizeof(OS_DPC));
  if (p == NULL) {
    return OS_STATUS_NOMEMORY;
  }

  p->devObj = devObj;
  p->handler = handler;
  p->context = context;
  p->destroy = FALSE;
  p->scheduled = FALSE;
  *dpc = p;

  return OS_STATUS_SUCCESS;
}

void
os_destroy_dpc(
  OS_DPC* dpc)
{
  OS_DEVICE_OBJECT* devObj;
  OS_SPINLOCK_FLAGS f;
  os_boolean        wait_destroy;
  wait_queue_head_t wait_q;
  wait_queue_t      wait;

  if (dpc == NULL) {
    os_dbgprint(0, ("os_destroy_dpc: NULL pointer\n"));
    return;
  }

  devObj = dpc->devObj;
  os_acquire_spinlock(&devObj->dpcLock, &f);
  if (dpc->scheduled) {
    /*
    ** DPC is either scheduled or running - we must wait for
    ** it to finish before freeing the control structure
    */

    wait_destroy = TRUE;
    set_current_state(TASK_UNINTERRUPTIBLE);
    init_waitqueue_head(&wait_q);
    init_waitqueue_entry(&wait, current);
    add_wait_queue(&wait_q, &wait);
    dpc->wake = &wait_q;
    dpc->destroy = TRUE;
  } else {
    wait_destroy = FALSE;
  }
  os_release_spinlock(&devObj->dpcLock, f);

  if (wait_destroy) {
    schedule();
    remove_wait_queue(&wait_q, &wait);
  }

  os_kernel_free(dpc);
}

OS_STATUS
os_schedule_dpc(
  OS_DPC* dpc,
  void*   argument)
{
  OS_DEVICE_OBJECT* devObj;
  OS_SPINLOCK_FLAGS f;

  devObj = dpc->devObj;

  os_acquire_spinlock(&devObj->dpcLock, &f);
  if (dpc->scheduled) {
    /* Already scheduled, don't schedule again */
    os_release_spinlock(&devObj->dpcLock, f);
    return OS_STATUS_SUCCESS;
  }
  if (dpc->destroy) {
    /* About to be destroyed - this indicates a programming error */
    os_release_spinlock(&devObj->dpcLock, f);
    os_dbgprint(0, ("os_schedule_dpc: DPC scheduled when being destroyed\n"));
    return OS_STATUS_UNSUCCESSFUL;
  }
  dpc->scheduled = TRUE;
  dpc->argument = argument;
  os_release_spinlock(&devObj->dpcLock, f);

#if (NO_TASKLETS)
  /* Just call the function inline */
  _os_dpc_handler((unsigned long) dpc);
#else
  tasklet_init(&dpc->tasklet, _os_dpc_handler, (unsigned long) dpc);
  tasklet_schedule(&dpc->tasklet);
  mark_bh(IMMEDIATE_BH);
#endif

  return OS_STATUS_SUCCESS;
}

/*
** --------------------------------------------------------------
** Lookaside cache functions
** --------------------------------------------------------------
*/

OS_STATUS
os_create_lcache(
  unsigned long item_size,
  int           flags,
  OS_LCACHE**   cache)
{
  static int    n = 0;
  char          name[32];
  char          x[32];
  unsigned long f;
  OS_LCACHE*    c;

  name[0] = '\0';
  if (driver_name != NULL) {
    strncpy(name, driver_name, 10);
  }
  sprintf(x, "_%ld", (unsigned long) n);
  strcat(name, x);
  n++;
  f = SLAB_HWCACHE_ALIGN;
  c = kmem_cache_create(name, item_size, 0, f, NULL, NULL);
  if (c == NULL) {
    return OS_STATUS_NOMEMORY;
  }

  *cache = c;
  return OS_STATUS_SUCCESS;
}

void
os_destroy_lcache(
  OS_LCACHE* cache)
{
  if (cache == NULL) {
    os_dbgprint(0, ("*** os_destroy_lcache: NULL pointer\n"));
    return;
  }

  if (kmem_cache_destroy(cache)) {
    os_dbgprint(0, ("*** os_destroy_lcache: memory leak detected\n"));
  }
}

/*
** --------------------------------------------------------------
** Memory allocation/freeing
** --------------------------------------------------------------
*/

void* os_kernel_malloc_dma(
  OS_DEVICE_OBJECT* devObj,
  unsigned long size,
  os_uint64* logicalAddr)
{
  os_uint64 log;
  dma_addr_t da;
  void* p;

  p = pci_alloc_consistent(devObj->dev.pci, size, &da);
  if (p == NULL) {
    return NULL;
  }
  log = (os_uint64) da;
  *logicalAddr = log;
  os_dbgprint(9, ("os_kernel_malloc_dma: size=%ld kern=0x%p log=0x%8.8lx%8.8lx\n",
                  size, p, (unsigned long) (log >> 32), (unsigned long) log));
  return p;
}

void os_kernel_free_dma(
  OS_DEVICE_OBJECT* devObj,
  unsigned long size,
  void* kernelAddr,
  os_uint64 logicalAddr)
{
  dma_addr_t da;

  da = (dma_addr_t) logicalAddr;
  pci_free_consistent(devObj->dev.pci, size, kernelAddr, da);
}

