/*
** platform-2.6.h - platform abstraction layer interface for Linux 2.6.x
*/

#ifndef _ADATA_PLATFORM_2_6_H
#define _ADATA_PLATFORM_2_6_H

#include <linux/module.h>
#include <linux/spinlock.h>
#include <asm/io.h>
#include <asm/byteorder.h>
#include <asm/irq.h>
#include <asm/div64.h>
#include <linux/config.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/ioctl.h>
#include <linux/kernel.h> 
#include <linux/pci.h>
#include <linux/types.h>
#include <linux/highmem.h>
#include <linux/mm.h>
#include <asm/uaccess.h>
#include <linux/slab.h>
#include <linux/workqueue.h>
#include <linux/pagemap.h>

/*
** This is used to ensure that structures that are overlaid
** on top of registers are packed and not padded by the compiler.
*/
#define os_declare_struct_packed __attribute__((packed,aligned(1)))

/*
** Status values from os_*() functions
**
** OS_SUCCESS(x) is TRUE if x indicates completed successfully
** OS_PENDING(x) is TRUE if x indicates not completed yet
** OS_FAILURE(x) is TRUE if x indicates completed unsuccessfully
*/

typedef unsigned long OS_STATUS;

#define OS_SUCCESS(x)          ((x) == OS_STATUS_SUCCESS)
#define OS_PENDING(x)          ((x) & OS_STATUSFLAG_PENDING)
#define OS_FAILURE(x)          (!OS_SUCCESS(x) && !OS_PENDING(x))

#define OS_MAKE_STATUS(x)      (OS_STATUSFLAG_USERCODE | (x))

#define OS_USERCODE(x)         ((x) & 0xffffffU)

#define OS_STATUSFLAG_PENDING  (0x80000000)
#define OS_STATUSFLAG_USERCODE (0x40000000)

/* Succesfull completion */
#define OS_STATUS_SUCCESS      (0x00000000)

/* Failure completion */
#define OS_STATUS_NOMEMORY     (0x00000001)
#define OS_STATUS_CANNOTSHARE  (0x00000002)
#define OS_STATUS_UNSUCCESSFUL (0x00000003)
#define OS_STATUS_CONFIGERROR  (0x00000004)
#define OS_STATUS_INVALIDPARAM (0x00000005)
#define OS_STATUS_INVALIDIOCTL (0x00000006)

/* Not completed yet */
#define OS_STATUS_PENDING      (OS_STATUSFLAG_PENDING)

/*
** Atomic datatypes, defined for portability
*/
typedef int os_boolean;

typedef u8  os_uint8;
typedef u16 os_uint16;
typedef u32 os_uint32;
typedef u64 os_uint64;

typedef s8  os_int8;
typedef s16 os_int16;
typedef s32 os_int32;
typedef s64 os_int64;

typedef void* os_event;

#define os_uint64_lo(x) ((os_uint32) (((x) >>  0) & 0xffffffffU))
#define os_uint64_hi(x) ((os_uint32) (((x) >> 32) & 0xffffffffU))
#define os_int64_lo(x)  ((os_int32) (((x) >>  0) & 0xffffffffU))
#define os_int64_hi(x)  ((os_int32) (((x) >> 32) & 0xffffffffU))

/*
** Get the endian conversion macros
*/
#include "platform_endian.h"

/*
** Define TRUE and FALSE if necessary
*/
#if !defined(TRUE)
#   define TRUE  (1)
#endif
#if !defined(FALSE)
#   define FALSE (!TRUE)
#endif

/*
** Event management
*/
#define os_set_event(X) /* No op on Linux */
			
/*
** Spinlock and context
*/
typedef unsigned long OS_SPINLOCK_FLAGS;
typedef struct _OS_SPINLOCK {
  spinlock_t spinlock;
} OS_SPINLOCK;

/*
** Allow forward references
*/
struct _OS_DRIVER_OBJECT;
typedef struct _OS_DRIVER_OBJECT OS_DRIVER_OBJECT;
struct _OS_IO_REQUEST;
typedef struct _OS_IO_REQUEST OS_IO_REQUEST;
struct _OS_MAPPABLE_REGION;
typedef struct _OS_MAPPABLE_REGION OS_MAPPABLE_REGION;

typedef struct _OS_DEVICE_OBJECT {
  union {
    struct pci_dev*     pci;
  }                     dev;
  OS_DRIVER_OBJECT*     drvObj;
  int                   openCount;
  void*                 deviceContext;
  OS_SPINLOCK           cancelLock;
  OS_SPINLOCK           dpcLock;
  struct {
    OS_MAPPABLE_REGION* regions;
  } maps;
} OS_DEVICE_OBJECT;

#define OS_DEVICE_DESC_TYPE_NONE           (0)
#define OS_DEVICE_DESC_TYPE_PCI            (1)
#define OS_DEVICE_DESC_MATCH_PCI_VENDOR    (0x1 << 0)
#define OS_DEVICE_DESC_MATCH_PCI_DEVICE    (0x1 << 1)
#define OS_DEVICE_DESC_MATCH_PCI_SUBVENDOR (0x1 << 2)
#define OS_DEVICE_DESC_MATCH_PCI_SUBDEVICE (0x1 << 3)
#define OS_DEVICE_DESC_MATCH_PCI_REV       (0x1 << 4)
typedef struct _OS_DEVICE_DESC {
  union {
    struct {
      os_uint16    dId, vId;
      os_uint16    subDId, subVId;
      os_uint8     revId;
      os_uint8     match;
    } pci;
  } id;
  int type;
} OS_DEVICE_DESC;

typedef struct _OS_DEVICE_RESOURCES {
  union {
    struct {
      os_uint64    base[6];
      os_uint64    size[6];
      os_boolean   io[6];
      int          count;
      os_uint8     irq_line;
    } pci;
  } raw;
  union {
    struct {
      os_uint64    base[6];
      os_uint64    size[6];
      os_boolean   io[6];
      int          count;
      long         vector;
    } pci;
  } translated;
} OS_DEVICE_RESOURCES;

typedef struct _OS_MEMORY_REGION {
  struct kiobuf* kiobuf;
} OS_MEMORY_REGION;

/*
** An I/O request. The client must complete it to unblock the user thread.
*/

typedef void (OS_IO_CANCELFUNC)(OS_DEVICE_OBJECT* devObj, OS_IO_REQUEST* req);

struct _OS_IO_REQUEST {
  /* Client must not touch */
  struct {
    wait_queue_head_t*   wait_q;
    os_boolean           completed;
    os_boolean           cancel;
    OS_IO_CANCELFUNC*    cancelFunc;
    OS_DEVICE_OBJECT*    devObj;
    struct {
      int                strategy; /* mass buffering strategy */
    } mass;
  } opaque;

  struct {
    unsigned long        code;
    void*                inBuf;
    unsigned long        inSize;
    void*                outBuf;
    unsigned long        outSize;
    void*                massBuf;
    unsigned long        massSize;
  } param;

  struct {
    OS_STATUS            status;
    unsigned long        length;
  } status;

  struct {
    void*                context;
  } user;
};

typedef struct _OS_USBD {
  void*           user;     /* User space address of buffer */
  unsigned long   length;
  unsigned long   first_page_offset;
  unsigned int    npage;
  struct page**   pages;
} OS_USBD;

typedef struct _OS_EVENT {
  OS_SPINLOCK       lock;
  wait_queue_head_t q;
  os_boolean        bRaised; /* event has been signalled */
} OS_EVENT;

/*
** Ensure that any cancellation routine that was previously set
** has been removed before calling!
*/
extern void
os_complete_request(
  OS_IO_REQUEST* req);

/*
** Returns FALSE if the request was already pending cancellation
**         TRUE if the cancellation routine was successfully set
*/
extern os_boolean
os_set_cancel_routine(
  OS_IO_REQUEST*    req,
  OS_IO_CANCELFUNC* func);

/* Returns TRUE if the request is pending cancellation */
#define os_cancel_pending(req) ((req)->opaque.cancel)

/*
** Functions/macros for reading hardware registers
**
** The following perform conversion to little-endian:
**
**     os_read_register_le8
**     os_read_register_le16
**     os_read_register_le32
**     os_write_register_le8
**     os_write_register_le16
**     os_write_register_le32
**     os_read_buffer_le8
**     os_read_buffer_le16
**     os_read_buffer_le32
**     os_write_buffer_le8
**     os_write_buffer_le16
**     os_write_buffer_le32
**
** The following do not perform endian conversion:
**
**     os_read_register_8
**     os_read_register_16
**     os_read_register_32
**     os_write_register_8
**     os_write_register_16
**     os_write_register_32
**     os_read_buffer_8
**     os_read_buffer_16
**     os_read_buffer_32
**     os_write_buffer_8
**     os_write_buffer_16
**     os_write_buffer_32
**
** os_write_register_*endian perform conversion to little endian before writing data
** os_read_register_* & os_write_register* do no endian conversion
*/

#define os_read_register_le8(x)  (readb((os_uint8*)  (x)))
#define os_read_register_le16(x) (readw((os_uint16*) (x)))
#define os_read_register_le32(x) (readl((os_uint32*) (x)))

#define os_write_register_le8(x, y)  { writeb(y, x); wmb(); }
#define os_write_register_le16(x, y) { writew(y, x); wmb(); }
#define os_write_register_le32(x, y) { writel(y, x); wmb(); }

static inline void os_read_buffer_le8(void* d, void* s, unsigned long n)
{
  os_uint8* d8;
  os_uint8* s8;

  d8 = (os_uint8*) d;
  s8 = (os_uint8*) s;
  while (n--) {
    *d8++ = readb(s8++);
  }
}

static inline void os_read_buffer_le16(void* d, void* s, unsigned long n)
{
  os_uint16* d16;
  os_uint16* s16;

  n >>= 1;
  d16 = (os_uint16*) d;
  s16 = (os_uint16*) s;
  while (n--) {
    *d16++ = readb(s16++);
  }
}

static inline void os_read_buffer_le32(void* d, void* s, unsigned long n)
{
  os_uint32* d32;
  os_uint32* s32;

  n >>= 2;
  d32 = (os_uint32*) d;
  s32 = (os_uint32*) s;
  while (n--) {
    *d32++ = readb(s32++);
  }
}

static inline void os_write_buffer_le8(void* d, void* s, unsigned long n)
{
  os_uint8* d8;
  os_uint8* s8;

  d8 = (os_uint8*) d;
  s8 = (os_uint8*) s;
  while (n--) {
    *d8++  =  readb(s8++);
  }
  wmb();
}

static inline void os_write_buffer_le16(void* d, void* s, unsigned long n)
{
  os_uint16* d16;
  os_uint16* s16;

  n >>= 1;
  d16 = (os_uint16*) d;
  s16 = (os_uint16*) s;
  while (n--) {
    *d16++  =  readb(s16++);
  }
  wmb();
}

static inline void os_write_buffer_le32(void* d, void* s, unsigned long n)
{
  os_uint32* d32;
  os_uint32* s32;

  n >>= 2;
  d32 = (os_uint32*) d;
  s32 = (os_uint32*) s;
  while (n--) {
    *d32++  =  readb(s32++);
  }
  wmb();
}

#define os_read_register_8(x)  (__raw_readb(x))
#define os_read_register_16(x) (__raw_readw(x))
#define os_read_register_32(x) (__raw_readl(x))

#define os_write_register_8(x, y)  { __raw_writeb(y, x); wmb(); }
#define os_write_register_16(x, y) { __raw_writew(y, x); wmb(); }
#define os_write_register_32(x, y) { __raw_writel(y, x); wmb(); }

#define os_read_buffer_8(d, s, n)  (memcpy_fromio(d, s, n))
#define os_read_buffer_16(d, s, n) (memcpy_fromio(d, s, n))
#define os_read_buffer_32(d, s, n) (memcpy_fromio(d, s, n))

#define os_write_buffer_8(d, s, n)  { memcpy_toio(d, s, n); wmb(); }
#define os_write_buffer_16(d, s, n) { memcpy_toio(d, s, n); wmb(); }
#define os_write_buffer_32(d, s, n) { memcpy_toio(d, s, n); wmb(); }

/*
** Debug message output
*/
extern int _os_dbglevel;

#if 1
#   define os_dbglevel(l) (((l) == 0) || (l) <= _os_dbglevel)
#   define os_assert(cond) do { if (!(cond)) { if (driver_name != NULL) { printk("*** %s: ", driver_name); } printk("assertion failed @ %s:%d, %s\n", __FILE__, (int) __LINE__, #cond); } } while (0);
#   define os_dbgprint(l, x) do { if (os_dbglevel(l)) { if (driver_name != NULL) { printk("%s: ", driver_name); } printk x; } } while (0);
#else
#   define os_dbglevel(l) (0)
#   define os_assert(cond)
#   define os_dbgprint(l, x)
#endif

/*
** Kernel memory allocation/freeing. Memory allocated this way must
** not be used as a DMA target.
**
** Must be called only in thread context.
*/
#define os_kernel_malloc(s)        (kmalloc((s), GFP_KERNEL))
#define os_kernel_free(s)          (kfree(s))

/*
** Kernel DMA memory allocation/freeing. Memory allocated this way
** may be used as a DMA target.
**
** Must be called only in thread context.
*/
extern void* os_kernel_malloc_dma(
  OS_DEVICE_OBJECT* devObj,
  unsigned long size,
  os_uint64* logicalAddr);

extern void os_kernel_free_dma(
  OS_DEVICE_OBJECT* devObj,
  unsigned long size,
  void* kernelAddr,
  os_uint64 logicalAddr);

#define os_kernel_alloc_page_dma() ((void*) __get_free_page(__GFP_DMA | GFP_KERNEL))
#define os_kernel_free_page_dma(p) (free_page((unsigned long) (p)))

#define os_memcpy(d, s, n) memcpy(d, s, n)

/*
** Lookaside caches
**
** os_create_lcache() and os_destroy_lcache() must be running in a
** thread context.
*/

typedef struct _OS_LCACHE
{
  kmem_cache_t* cache;
  char name[32];
} OS_LCACHE;

extern OS_STATUS
os_create_lcache(
  unsigned long item_size,
  int           flags,
  OS_LCACHE**   cache);

extern void
os_destroy_lcache(
  OS_LCACHE* cache);

#define os_alloc_lcache(CACHE, F) kmem_cache_alloc((CACHE)->cache, SLAB_KERNEL);

#define os_free_lcache(CACHE, ITEM)   kmem_cache_free((CACHE)->cache, (ITEM));

/*
** Mapping/unmapping of I/O memory into/out of kernel space
*/
extern void*
os_map_kernel(
  os_uint64 phys,
  os_uint64 length);

extern void
os_unmap_kernel(
  void*     kern,
  os_uint64 phys,
  os_uint64 length);

/*
** Convert a kernel address to a bus address
*/
extern os_uint64
os_kernel_to_bus(
  OS_DEVICE_OBJECT* devObj,
  void*             kern);

/*
** Scatter-gather map functions
**
** os_sg_alloc             - Allocates a scatter-gather map object.
**
** os_sg_free              - Frees a scatter-gather map object.
**
** os_sg_init              - Initializes a scatter-gather map object in
**                           preparation for a DMA transfer, removing any
**                           buffers that were added to it and setting the
**                           direction of the DMA transfer to be performed.
**
** os_sg_add_kernel_buffer - Adds a kernel-space buffer to a scatter-gather
**                           map object; must be called after os_sg_init.
**
** os_sg_add_usbd_buffer   - Adds a user-space buffer that is described
**                           by a USBD to a scatter-gather object; must be
**                           called after os_sg_init.
**
** os_sg_add_user_buffer   - Adds a user-space buffer to a scatter-gather
**                           map object; must be called after os_sg_init.
**
** os_sg_map               - Maps the buffers that have been added to the
**                           scatter-gather map object so that the
**                           os_sg_get_buffer function may be called. Must
**                           be called before a DMA transfer. Returns the
**                           number of discontiguous buffers.
**
** os_sg_unmap             - Unmaps the scatter-gather map object; must be
**                           called after a DMA transfer.
**
** os_sg_get_buffer        - Returns the bus address and length of a given
**                           buffer. May only be called in between
**                           os_sg_map and os_sg_unmap.
*/

typedef struct _OS_SG_MAP {
  OS_DEVICE_OBJECT* pDevObj;
  void** bounce;
  unsigned int nBuff, maxBuff;
  unsigned int nBounceUsed;
  os_boolean bDmaWrite;
  struct {
    void* kernel;
    void* actual;
    os_uint64 logical;
    unsigned long length;
    os_uint32 local;
    int type;
  }* buffInfo;
} OS_SG_MAP;

extern OS_SG_MAP*
os_sg_alloc(
  OS_DEVICE_OBJECT* devObj,
  unsigned int nEntry);

extern void
os_sg_free(
  OS_SG_MAP* map);

static inline void
os_sg_init(
  OS_SG_MAP* map,
  os_boolean bDmaWrite)
{
  map->nBuff = 0;
  map->nBounceUsed = 0;
  map->bDmaWrite = bDmaWrite;
}

extern unsigned long
os_sg_add_kernel_buffer(
  OS_SG_MAP* map,
  os_uint32 local,
  os_boolean bFixedLocal,
  void* kBuffer,
  unsigned int length);

extern unsigned long
os_sg_add_usbd_buffer(
  OS_SG_MAP* map,
  os_uint32 local,
  os_boolean bFixedLocal,
  OS_USBD* usbd,
  unsigned long offset,
  unsigned long length);

extern unsigned long
os_sg_add_user_buffer(
  OS_SG_MAP* map,
  os_uint32 local,
  os_boolean bFixedLocal,
  void* uBuffer,
  unsigned int length);

extern unsigned int
os_sg_map(
  OS_SG_MAP* map);

extern void
os_sg_unmap(
  OS_SG_MAP* map);

static inline void
os_sg_get_buffer(
  OS_SG_MAP* map,
  unsigned int index,
  os_uint32* localAddr,
  os_uint64* bufferLogical,
  unsigned long* bufferLength)
{
  *localAddr = map->buffInfo[index].local;
  *bufferLogical = map->buffInfo[index].logical;
  *bufferLength = map->buffInfo[index].length;
}

/*
** User-space buffer descriptions
**
** os_create_usbd()       Creates a buffer description for the user buffer
**                        specified by 'user' and 'length'. If necessary, it
**                        is also locked in physical memory.
**
**                        Must be called in a thread context.
**
** os_destroy_usbd()      Undoes the effect of os_create_usbd(). Unlocks
**                        the user buffer and destroys the buffer description.
**
** os_flush_io_buffer()   Call to ensure CPU cache coherency with a buffer
**                        that has been used for DMA.
**
** os_copy_from_usbd()    Copy data from a USBD to kernel space
**
** os_copy_to_usbd()      Copy data from kernel space to a USBD
**
** os_get_usbd_bus_addr() Get bus address of a byte at a specifed offset in a
**                        USBD, and also its page offset.
*/

#define os_page_size      (PAGE_SIZE)
#define os_page_mask_low  ((unsigned long) (PAGE_SIZE - 1))
#define os_page_mask_high (~(os_page_mask_low))
#define os_page_shift     (PAGE_SHIFT)

#if 0
#define OS_KIOBUFSHIFT   (26)
#define OS_KIOBUFCHUNK   (1 << OS_KIOBUFSHIFT)
#define OS_KIOBUFMAXPAGE (1 << (OS_KIOBUFSHIFT - os_page_shift))
#define OS_KIOBUFMASKLO  (OS_KIOBUFCHUNK - 1)
#define OS_KIOBUFMASKHI  (~OS_KIOBUFMASKLO)
#endif

extern OS_STATUS
os_create_usbd(
  void*         user,
  unsigned long length,
  OS_USBD**     usbd);

extern void
os_destroy_usbd(
  OS_USBD*      usbd);

extern void
os_flush_io_buffer(
  OS_USBD*      usbd,
  unsigned long offset,
  unsigned long length,
  os_boolean    dma_write);

#define os_get_usbd_page_offset(x) ((x)->first_page_offset)

static inline os_boolean
os_get_usbd_bus_addr(
  OS_USBD*       usbd,
  unsigned long  offset,
  os_uint32*     busAddr,
  unsigned long* pg_offs)
{
  struct page* p;
  unsigned int p_idx, p_offs;

  offset += usbd->first_page_offset;
  p_idx = offset >> os_page_shift;
  p_offs = offset & os_page_mask_low;
  p = usbd->pages[p_idx];
  *pg_offs = p_offs;
  if (PageHighMem(p)) {
    return FALSE;
  }
  *busAddr = virt_to_bus(((os_uint8*) page_address(p)) + p_offs);
  return TRUE;
}

static inline void
os_copy_from_usbd(
  void*         buffer,
  OS_USBD*      usbd,
  unsigned long offset,
  unsigned long n)
{
  void* pkernel;
  struct page* p;
  unsigned int p_idx, p_offs;

  offset += usbd->first_page_offset;
  p_idx = offset >> os_page_shift;
  p_offs = offset & os_page_mask_low;
  p = usbd->pages[p_idx];
  pkernel = kmap(p);
  memcpy(buffer, pkernel + p_offs, n);
  kunmap(p);
}

static inline void
os_copy_to_usbd(
  const void*   buffer,
  OS_USBD*      usbd,
  unsigned long offset,
  unsigned long n)
{
  void* pkernel;
  struct page* p;
  unsigned int p_idx, p_offs;

  offset += usbd->first_page_offset;
  p_idx = offset >> os_page_shift;
  p_offs = offset & os_page_mask_low;
  p = usbd->pages[p_idx];
  pkernel = kmap(p);
  memcpy(pkernel + p_offs, buffer, n);
  kunmap(p);
}

/*
** os_copy_from_user()    Copy data from user space to kernel space
**
** os_copy_to_user()      Copy data from kernel space to user space
*/

#define os_copy_from_user(kernel, user, length) copy_from_user(kernel, user, length);
#define os_copy_to_user(kernel, user, length)   copy_to_user(kernel, user, length);

/*
** Spinlock functions
**
** 'l' is of type OS_SPINLOCK*
** 'f' is of type OS_SPINLOCK_FLAGS*
*/

#define os_init_spinlock(l)       spin_lock_init(&(l)->spinlock)
#define os_acquire_spinlock(l, f) spin_lock_irqsave(&(l)->spinlock, *(f))
#define os_release_spinlock(l, f) spin_unlock_irqrestore(&(l)->spinlock, f)

/*
** Interrupt capture.
**
** Following must be called in a thread context:
**
**     os_attach_interrupt()
**     os_detach_interrupt()
*/

/*
** An interrupt service routine for a device.
**
** Returns TRUE if at least one interrupt was handled by the routine,
** otherwise FALSE.
*/
typedef os_boolean (OS_ISR)(void* deviceContext);

typedef struct _OS_ATTACHED_IRQ {
  long              vector;
  OS_ISR*           isr;
  OS_DEVICE_OBJECT* devObj;
  void*             deviceContext;
} OS_ATTACHED_IRQ;

extern OS_STATUS
os_attach_interrupt(
  OS_DEVICE_OBJECT* devObj,
  OS_ISR*           isr,
  long              vector,
  const char*       devName,
  OS_ATTACHED_IRQ** o);

extern void
os_detach_interrupt(
  OS_ATTACHED_IRQ*  irqObj);

/*
** Kernel mode event handling:
**
**   os_init_event           call in any context
**   os_reset_event          call in any context
**   os_signal_event         call in any context
**   os_wait_event           call in a thread context
*/

extern void
os_init_event(
  OS_EVENT* event);

/*
** Waits for the specified event to be signalled. The wait mode may be
** interruptible (signals can be caught) or noninterruptible (signals are
** ignored during the wait).
**
** Returns TRUE if a signal/alert was caught while waiting for the event,
** otherwise FALSE (implies that the event was signalled).
**
** If the event is already in the 'signalled' state when os_wait_event()
** is called, it returns FALSE immediately.
*/
extern os_boolean
os_wait_event(
  OS_EVENT*  event,
  os_boolean interruptible);

/*
** Put an event into the 'nonsignalled' state.
*/
extern void
os_reset_event(
  OS_EVENT* event);


/*
** Put an event into the 'signalled' state. Will wake up any thread
** waiting on the signal.
*/
extern void
os_signal_event(
  OS_EVENT* event);

/*
** Deferred procedure calls
**
** Following must be called in a thread context:
**
**     os_destroy_dpc()
*/

typedef void (OS_DPC_FUNC)(void* context, void* argument);

typedef struct _OS_DPC {
  OS_DEVICE_OBJECT*     devObj;
  OS_DPC_FUNC*          handler;
  void*                 context;
  void*                 argument;
  os_boolean            destroy;
  os_boolean            scheduled;
  wait_queue_head_t*    wake;
  struct work_struct    work;
} OS_DPC;

/*
** Create and initialise a deferred procedure call control structure.
**
** 'context' is a value that is fixed throughout the life of the DPC.
*/
extern OS_STATUS
os_create_dpc(
  OS_DEVICE_OBJECT* devObj,
  OS_DPC_FUNC*      func,
  void*             context,
  OS_DPC**          dpc);

/*
** If the DPC is scheduled when this function is called, it
** unschedules the DPC before destroying the DPC.
**
** If the DPC handler is running when this function is called, it busy-waits
** for the DPC handler to finish before destroying the DPC.
**
** MUST be called in a thread context.
*/
extern void
os_destroy_dpc(
  OS_DPC*           dpc);

/*
** Schedules a DPC for execution. Note that:
**
** (a) On an SMP system, the DPC's handler could begin executing
**     immediately.
**
** (b) On a platform where DPC's don't really exist, os_schedule_dpc()
**     will actually invoke the DPC's handler synchronously, before
**     returning.
**
** 'argument' is passed to the DPC handler, and may be different on
** each call to os_schedule_dpc().
*/
extern OS_STATUS
os_schedule_dpc(
  OS_DPC*           dpc,
  void*             argument); /* May be different on each invocation */

/*
** Allow a function to be invoked with no possibility of receiving
** an interrupt while the function executes.
**
** Note: does NOT disable other interrupt vectors; only disables the
**       interrupt vector that was used to create 'irqObj'.
*/

typedef void* (OS_INVOKE_FUNC)(void* context);

void*
os_invoke_atomic(
  OS_ATTACHED_IRQ*  irqObj,
  OS_INVOKE_FUNC*   function,
  void*             context);

/*
** Mapping memory (must be called in a thread context)
**
** os_set_region_mappable()     Marks a region of memory as mappable to user space.
**                              Mapping may occur immediately (Windows) or when a
**                              process requests a mapping (UNIX)
**
** os_set_region_nonmappable()  Marks a region of memory as not mappable to user
**                              space. If the region has been mapped, it is unmapped.
*/

struct _OS_MAPPABLE_REGION {
  OS_DEVICE_OBJECT*    devObj;  
  os_uint64            phys;
  void*                sysAddr;
  unsigned long        length;
  void*                context;
  OS_MAPPABLE_REGION** prev;
  OS_MAPPABLE_REGION*  next;
};

OS_STATUS
os_allow_mapping(
  OS_DEVICE_OBJECT*    devObj,
  os_uint64            phys,
  os_uint64            length,
  void*                context,
  OS_MAPPABLE_REGION** region);

void
os_disallow_mapping(
  OS_DEVICE_OBJECT*    devObj,
  OS_MAPPABLE_REGION*  region);

/*
** Delaying execution
**
** os_get_time()         Return system time since boot in ticks.
**
** os_delay_busy_us()    Busy-delay for specified microseconds.
**                       Ties up the CPU it runs on.
**
** os_delay_thread()     Delay for specified number of ticks. Must
**                       be running in a thread context and may be
**                       interrupted by signals.
**
**                       Returns TRUE if interrupted by a signal,
**                       otherwise FALSE.
*/

#define os_delay_busy_us(x) (udelay((unsigned long) (x)))

#define os_ticks_per_second (HZ)

static inline unsigned long /* ticks */
os_get_time(void)
{
  return (unsigned long) jiffies;
}

extern os_boolean
os_delay_thread(
  os_boolean     interruptible,
  unsigned long* ticks);

/*
** Driver initialisation
**
** Data members:
**
**   drvName                     Client should initialise to the name of the device
**
**   hardwareIds                 Client must set to point to an array of
**                               OS_DEVICE_DESC structures.
**
** Methods (client must initialise):
**
**   onCreateDevice              Called by framework when a new device is found
**
**   onStartDevice               Called by framework to start the device (map
**                               BARs into kernel space, attach interrupt handler
**                               etc.)
**
**   onStopDevice                Called by framework to stop the device (unmap
**                               BARs from kernel space, detach interrupt handler
**                               etc.)
**
**   onDeleteDevice              Called by framework when a device is deleted
**  
**   onOpen                      Called by framework when a user-mode thread
**                               opens the device.
**
**   onClose                     Called by framework when a user-mode thread
**                               closes the device.
**
**   onIoctl                     Called as a result of an ioctl() call in application
**
** All of the above methods are called in a thread context, so they may
** call blocking OS functions.
**
** Of these methods, the following are not allowed to return a pending status:
**
**   onCreateDevice
**   onStartDevice
**   onStopDevice
**   onDeleteDevice
**   onOpen
**   onClose
**
** The following methods are allowed to return a pending status:
**
**   onIoctl
*/

/*
** Represents an open handle to the device, from the driver's point of view
*/
typedef void* OS_USER_CLIENT;

struct _OS_DRIVER_OBJECT {
  int                         numDev;
  char*                       drvName;
  OS_DEVICE_DESC*             hardwareIds;
  OS_STATUS (*onCreateDevice)(
    OS_DRIVER_OBJECT*         drvObj,
    OS_DEVICE_OBJECT*         devObj,
    int                       matchIndex,
    OS_DEVICE_DESC*           devDesc);
  OS_STATUS (*onStartDevice)(
    OS_DEVICE_OBJECT*         devObj,
    OS_DEVICE_RESOURCES*      resources);
  OS_STATUS (*onStopDevice)(
    OS_DEVICE_OBJECT*         devObj);
  void (*onDeleteDevice)(
    OS_DEVICE_OBJECT*         devObj);
  OS_STATUS (*onOpen)(
    OS_DEVICE_OBJECT*         devObj,
    OS_USER_CLIENT            client);
  OS_STATUS (*onClose)(
    OS_DEVICE_OBJECT*         devObj,
    OS_USER_CLIENT            client);
  OS_STATUS (*onMap)(
    OS_DEVICE_OBJECT*         devObj,
    OS_MAPPABLE_REGION*       region,
    void*                     context,
    void*                     virtual,
    os_uint64                 phys,
    os_uint64                 length);
  OS_STATUS (*onIoctl)(
    OS_DEVICE_OBJECT*         devObj,
    OS_USER_CLIENT            client,
    OS_IO_REQUEST*            req);
};

/*
** Device initialisation
**
** os_set_device_context()    call in the onCreateDevice handler to get a device object's context
**
** os_get_device_context()    call any time to get a device object's context
**
** os_get_driver_param()      call any time to get driver's parameters
*/

static inline void
os_set_device_context(
  OS_DEVICE_OBJECT* devObj,
  void* deviceContext)
{
  devObj->deviceContext = deviceContext;
}

static inline void*
os_get_device_context(
  OS_DEVICE_OBJECT* devObj)
{
  return devObj->deviceContext;
}

typedef enum DriverParamType {
  DriverParamType_uint32 = 0,
  DriverParamType_int32  = 1
} DriverParamType;

extern os_boolean
os_get_driver_param(
  OS_DRIVER_OBJECT* pDrvObj,
  const char*       name,
  DriverParamType   type,
  void*             value,
  unsigned long     size);

/*
** Arithmetic
*/

static inline os_uint32 os_div_64by32(os_uint64 dividend, os_uint32 divisor)
{
  do_div(dividend, divisor);
  return (os_uint32) dividend;
}

/*
** The client code MUST define these functions and variables.
*/

extern OS_STATUS
driver_init(
  OS_DRIVER_OBJECT* drvObj);

extern void
driver_uninit(
  OS_DRIVER_OBJECT* drvObj);

extern const char* driver_name;

#endif
