/*
** ADM-XRC Series Driver
**
** (C) Copyright 2001-2004 Alpha Data
**
** rw.c - IOCTL handlers for reading and writing the local bus spaces.
**
** ----------------------------------------------------------------------------
*/

#include "platform.h"
#include "device.h"
#include "ioctl.h"
#include "rw.h"

/*
** Validate the parameters of an IoRead or IoWrite call and
** return a pointer to the kernel address where the transfer
** is to start.
*/
static OS_STATUS
validate_rw(
  DeviceContext* pDevCtx,
  unsigned long  flags,
  unsigned long  width,
  unsigned long  length,
  os_uint32      local,
  void**         devicemem)
{
  unsigned long span;
  os_uint32     maxlocal;
  os_boolean    fixed = (flags & ADMXRC2_IOFIXED) != 0;
  os_uint8*     mem;
	
  span = fixed ? width : length;
	
  if (flags & ADMXRC2_IOADAPTER) {
    if (local >= pDevCtx->info.adapterLimit) {
      return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
    }
    mem = (os_uint8*) pDevCtx->hw.adapter.untyped;
    maxlocal = pDevCtx->info.adapterLimit;
  } else {
    if ((local >= pDevCtx->info.spaceInfo[0].LocalBase) &&
	(local < (pDevCtx->info.spaceInfo[0].LocalBase + pDevCtx->info.spaceInfo[0].LocalSize))) {
      /* Space 0 */
      local -= pDevCtx->info.spaceInfo[0].LocalBase;
      mem = (os_uint8*) pDevCtx->hw.fpga;
      maxlocal = pDevCtx->info.spaceInfo[0].LocalSize;
    } else if ((local >= pDevCtx->info.spaceInfo[1].LocalBase) &&
	       (local < (pDevCtx->info.spaceInfo[1].LocalBase + pDevCtx->info.spaceInfo[1].LocalSize))) {
      /* Space 1 */
      local -= pDevCtx->info.spaceInfo[1].LocalBase;
      mem = (os_uint8*) pDevCtx->hw.space1.untyped;
      maxlocal = pDevCtx->info.spaceInfo[1].LocalSize;
    } else {
      /* Invalid local bus address */
      return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
    }
  }
	
  /* Don't permit the space boundary to be crossed */
  if (local + span > maxlocal) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
	
  /* Get the system address at which to begin writing */
  mem += local;
  
  *devicemem = mem;
  
  return OS_STATUS_SUCCESS;
}

/*
** ADMXRC_ProgramBytes() interprets a simple program and is intended
** to enable flash programming algorithms to be implemented at
** application level with a reduced number of IOCTLs.
**
** The program format is an array of pairs of ULONGs:
**
**       +----+--------------------------------------------+
**       | op |                 flash_addr                 |
**       +----+--------------------------------------------+
**       |                    operands                     |
**       +-------------------------------------------------+
**
** where
**       flash_addr = bits [27:0] of word 0
**       op = bits [31:28] of word 0
**
** There are currently 3 ops defined:
**
**       op = 0  => write to flash
**       op = 1  => wait for flash location to read back value
**       op = 2  => read flash
**
** Since flash_addr is 28 bits wide, Flash up to 256MB in size is supported.
**
** OP 0 (write value)
** ------------------
**
**       +----+--------------------------------------------+
**       |  0 |                 flash_addr                 |
**       +----+------------------+------------+------------+
**       |                       |    mask    |    data    |
**       +-----------------------+------------+------------+
**
** where
**       data = bits[7:0] of word 1
**       mask = bits[15:8] of word 1
**
** the operation performed is:
**
**    FLASH_WRITE(flash_addr, (FLASH_READ(flash_addr) & mask) | data);
**
**
** OP 1 (wait for value)
** ---------------------
**
**       +----+--------------------------------------------+
**       |  1 |                 flash_addr                 |
**       +-+--+------------------+------------+------------+
**       |m|      timeout        |    mask    |    data    |
**       +-+---------------------+------------+------------+
**
** where
**       data = bits[7:0] of word 1
**       mask = bits[15:8] of word 1
**		 timeout = bits[30:16] of word 1
**		 m = bits[31] of word 1
**
** the operation performed is:
**
**    while ((FLASH_READ(flash_addr) & mask) != data) { };
**
** There is a timeout on this wait; the timeout is measured in
** milliseconds if m == 1 and in microseconds if m == 0. If the
** timeout period elapses without the specified value being read,
** the return value from AdmXrcProgramBytes() is ADMXRC_TIMEOUT.
**
** OP 2 (read value)
** -----------------
**
**       +----+--------------------------------------------+
**       |  2 |                 flash_addr                 |
**       +----+-------------------------------+------------+
**       |                                    |    data    |
**       +------------------------------------+------------+
**
** where
**       data = bits[7:0]
**
** the operation performed is:
**
**    data = FLASH_READ(flash_addr);
**
*/
OS_STATUS
ADMXRC_ProgramBytes(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* req)
{
  unsigned long               i, length;
  os_uint32*                  program;
  os_uint32                   op1, op2, addr;
  os_uint8                    data, mask;
  os_uint16                   val16;
  os_uint8                    page;
  os_boolean                  bInterrupt;

  os_dbgprint(3, ("ADMXRC_ProgramBytes: entered\n"));

  length = req->param.massSize;
  program = req->param.massBuf;
  if ((length & (2 * sizeof(os_uint32) - 1)) ||
      (length && program == NULL)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  switch (pDevCtx->info.cardInfo.BoardType) {
  case ADMXRC2_BOARD_ADMXRC:
  case ADMXRC2_BOARD_ADMXRC_P:
  case ADMXRC2_BOARD_ADMXRC2_LITE:
  case ADMXRC2_BOARD_ADMXRC2:
  case ADMXRC2_BOARD_ADPWRC2:
  case ADMXRC2_BOARD_ADPDRC2:
    break;

  default:
    return OS_MAKE_STATUS(ADMXRC_NOT_SUPPORTED);
  }

  /* Interpret the programming codes supplied */
  for (i = 0; i < length; i += 2 * sizeof(os_uint32)) {
    op1 = program[0];
    op2 = program[1];
    addr = op1 & 0x001fffffU;
    page = (os_uint8) ((op1 >> 21) & 0x7fU);
    data = (os_uint8) (op2 >> 0);
    mask = (os_uint8) (op2 >> 8);
    
    switch ((op1 >> 28) & 0xfU) {
    case 0x0:
      {
	/* Write a value to flash location with mask */
	os_dbgprint(3, ("ADMXRC_ProgramBytes: op 0 mask = 0x%2.2lx data = 0x%2.2lx\n",
			(unsigned long) mask, (unsigned long) data));

	if (mask) {
	  pDevCtx->methods.flash(pDevCtx, FALSE, addr, &val16);
	  data |= ((os_uint8) val16) & mask;
	}
	val16 = data;
	pDevCtx->methods.flash(pDevCtx, TRUE, addr, &val16);
	break;
      }

    case 0x1:
      {
	/* Wait for a certain flash location to read a certain value */
	os_boolean    done = FALSE;
	unsigned long timeout;
	unsigned long expires, dummy;

	os_dbgprint(3, ("ADMXRC_ProgramBytes: op 1 mask = 0x%2.2lx data = 0x%2.2lx\n",
			(unsigned long) mask, (unsigned long) data));
	
	/* Get timeout value */
	timeout = (op2 >> 16) & 0x7fffU;

	if (op2 & (1 << 31)) {
	  /* Timeout is in milliseconds */
	  expires = timeout * os_ticks_per_second / 1000;
	  if (!expires) {
	    expires++;
	  }
	  expires += os_get_time();
	  
	  done = FALSE;
	  while (1) {
	    pDevCtx->methods.flash(pDevCtx, FALSE, addr, &val16);
	    if ((val16 & mask) == data) {
	      done = TRUE;
	      break;
	    }

	    dummy = 1;
	    bInterrupt = os_delay_thread(TRUE, &dummy);
	    if (bInterrupt) {
	      return OS_MAKE_STATUS(ADMXRC_CANCELLED);
	    }
	    
	    if (os_get_time() >= expires) {
	      break;
	    }
	  }
	} else {
	  /* Timeout is in microseconds */
	  while (timeout) {
	    unsigned long chunk = (timeout > 5) ? 5 : timeout;
	    
	    pDevCtx->methods.flash(pDevCtx, FALSE, addr, &val16);
	    if ((val16 & mask) == data) {
	      done = TRUE;
	      break;
	    }

	    os_delay_busy_us(chunk);
	    timeout -= chunk;
	  }
	}

	if (!done) {
	  return OS_MAKE_STATUS(ADMXRC_TIMEOUT);
	}
      }
      break;
    case 0x2:
      {
	/* Read value from flash location */
	pDevCtx->methods.flash(pDevCtx, FALSE, addr, &val16);
	program[1] &= ~0xffU;
	program[1] |= val16 & 0xffU;

	os_dbgprint(3, ("ADMXRC_ProgramBytes: op 2 addr = 0x%lx data = 0x%2.2lx\n",
			(unsigned long) addr, (unsigned long) val16));
      }
      break;
    }
    program += 2;
  }

  return OS_STATUS_SUCCESS;
}

OS_STATUS
ADMXRC_Read(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_READ* iptr;
  OS_STATUS           status;
  void*               devicemem;
  os_boolean          fixed, endian;
  unsigned long       length;
  void*               buffer;

  iptr = (ADMXRC_IOCTLS_READ*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  buffer = (void*) req->param.massBuf;
  length = req->param.massSize;

  /* Only byte (0), word (1) or longword (2) accesses are allowed */
  if (iptr->in.width > ADMXRC_IOLONG) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  /* Check the local bus address, transfer size and flags parameters */
  status = validate_rw(pDevCtx,
		       iptr->in.flags,
		       1 << iptr->in.width,
		       length,
		       iptr->in.local,
		       &devicemem);
  if (status != OS_STATUS_SUCCESS) {
    return status;
  }

  /*
  ** Note that 'length' is always in bytes.
  ** Transfers of 16 bits and 32 bits will be truncated if not correct multiple
  */

  fixed = (iptr->in.flags & ADMXRC2_IOFIXED) != 0;
  endian = (iptr->in.flags & ADMXRC2_IOADAPTER) != 0;

  if (endian) {
    if (!fixed) {
      switch (iptr->in.width) {
      case ADMXRC_IOBYTE:
	os_read_buffer_le8(buffer, (void*) devicemem, length);
	break;

      case ADMXRC_IOWORD:
	os_read_buffer_le16(buffer, (void*) devicemem, length);
	break;

      case ADMXRC_IOLONG:
	os_read_buffer_le32(buffer, (void*) devicemem, length);
	break;

      default:
	os_assert(FALSE);
	break;
      }
    } else {		
      switch (iptr->in.width) {
      case ADMXRC_IOBYTE:
	{
	  os_uint8* puc;
	  
	  puc = (os_uint8*) buffer;
	  while (length > 0) {
	    *puc = os_read_register_le8((os_uint8*) devicemem);
	    puc++;
	    length--;
	  }
	  break;
	}
      case ADMXRC_IOWORD:
	{
	  os_uint16* pus;
	  
	  pus = (os_uint16*) buffer;
	  length >>= 1;
	  while (length > 0) {
	    *pus = os_read_register_le16((os_uint16*) devicemem);
	    pus++;
	    length--;
	  }
	  break;
	}
      case ADMXRC_IOLONG:
	{
	  os_uint32* pul;
	  
	  pul = (os_uint32*) buffer;
	  length >>= 2;
	  while (length > 0) {
	    *pul = os_read_register_le32((os_uint32*) devicemem);
	    pul++;
	    length--;
	  }
	  break;
	}

      default:
	os_assert(FALSE);
	break;
      }
    }
  } else {
    if (!fixed) {
      switch (iptr->in.width) {
      case ADMXRC_IOBYTE:
	os_read_buffer_8(buffer, (void*) devicemem, length);
	break;

      case ADMXRC_IOWORD:
	os_read_buffer_16(buffer, (void*) devicemem, length);
	break;

      case ADMXRC_IOLONG:
	os_read_buffer_32(buffer, (void*) devicemem, length);
	break;

      default:
	os_assert(FALSE);
	break;
      }
    } else {		
      switch (iptr->in.width) {
      case ADMXRC_IOBYTE:
	{
	  os_uint8* puc;
	  
	  puc = (os_uint8*) buffer;
	  while (length > 0) {
	    *puc = os_read_register_8((os_uint8*) devicemem);
	    puc++;
	    length--;
	  }
	  break;
	}
      case ADMXRC_IOWORD:
	{
	  os_uint16* pus;
	  
	  pus = (os_uint16*) buffer;
	  length >>= 1;
	  while (length > 0) {
	    *pus = os_read_register_16((os_uint16*) devicemem);
	    pus++;
	    length--;
	  }
	  break;
	}
      case ADMXRC_IOLONG:
	{
	  os_uint32* pul;
	  
	  pul = (os_uint32*) buffer;
	  length >>= 2;
	  while (length > 0) {
	    *pul = os_read_register_32((os_uint32*) devicemem);
	    pul++;
	    length--;
	  }
	  break;
	}

      default:
	os_assert(FALSE);
	break;
      }
    }
  }
  
  return OS_STATUS_SUCCESS;
}

OS_STATUS
ADMXRC_ReadReg(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_READREG* iptr;
  ADMXRC_IOCTLS_READREG* optr;
  os_uint8               index;
  os_uint8*              base;

  iptr = (ADMXRC_IOCTLS_READREG*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  optr = (ADMXRC_IOCTLS_READREG*) req->param.outBuf;
  if (optr == NULL || req->param.outSize != sizeof(optr->out)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  base = (os_uint8*) pDevCtx->hw.space1.untyped;
  index = iptr->in.index;
  if (index > pDevCtx->info.spaceInfo[1].LocalSize) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  optr->out.value = os_read_register_8(base + index);

  req->status.length = sizeof(optr->out);

  return OS_STATUS_SUCCESS;
}

OS_STATUS
ADMXRC_Write(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_WRITE* iptr;
  OS_STATUS            status;
  void*                devicemem;
  os_boolean           fixed, endian;
  unsigned long        length;
  void*                data;

  iptr = (ADMXRC_IOCTLS_WRITE*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  data = (void*) req->param.massBuf;
  length = req->param.massSize;

  /* Only byte (0), word (1) or longword (2) accesses are allowed */
  if (iptr->in.width > ADMXRC_IOLONG) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  /* Check the local bus address, transfer size and flags parameters */
  status = validate_rw(pDevCtx,
		       iptr->in.flags,
		       1 << iptr->in.width,
		       length,
		       iptr->in.local,
		       &devicemem);
  if (status != OS_STATUS_SUCCESS) {
    return status;
  }

  /*
  ** Note that 'length' is always in bytes.
  ** Transfers of SHORT and LONG will be truncated if not correct multiple
  */
  fixed = (iptr->in.flags & ADMXRC_IOFIXED) != 0;
  endian = (iptr->in.flags & ADMXRC2_IOADAPTER) != 0;

  if (endian) {
    if (!fixed) {
      switch (iptr->in.width) {
      case ADMXRC_IOBYTE:
	os_write_buffer_le8(devicemem, data, length);
	break;

      case ADMXRC_IOWORD:
	os_write_buffer_le16(devicemem, data, length);
	break;

      case ADMXRC_IOLONG:
	os_write_buffer_le32(devicemem, data, length);
	break;

      default:
	os_assert(FALSE);
	break;
      }
    } else {
      
      switch (iptr->in.width) {
      case ADMXRC_IOBYTE:
	{
	  os_uint8*  puc;
	  
	  puc = (os_uint8*) data;
	  while (length > 0) {
	    os_write_register_le8((os_uint8*) devicemem, *puc);
	    puc++;
	    length--;
	  }
	  break;
	}

      case ADMXRC_IOWORD:
	{
	  os_uint16* pus;
	  
	  pus = (os_uint16*) data;
	  length >>= 1;
	  while (length > 0) {
	    os_write_register_le16((os_uint16*) devicemem, *pus);
	    pus++;
	    length--;
	  }
	  break;
	}

      case ADMXRC_IOLONG:
	{
	  os_uint32* pul;
	  
	  pul = (os_uint32*) data;
	  length >>= 2;
	  while (length > 0) {
	    os_write_register_le32((os_uint32*) devicemem, *pul);
	    pul++;
	    length--;
	  }
	  break;
	}

      default:
	os_assert(FALSE);
	break;
      }
    }
  } else {
    if (!fixed) {
      switch (iptr->in.width) {
      case ADMXRC_IOBYTE:
	os_write_buffer_8(devicemem, data, length);
	break;

      case ADMXRC_IOWORD:
	os_write_buffer_16(devicemem, data, length);
	break;

      case ADMXRC_IOLONG:
	os_write_buffer_32(devicemem, data, length);
	break;

      default:
	os_assert(FALSE);
	break;
      }
    } else {
      
      switch (iptr->in.width) {
      case ADMXRC_IOBYTE:
	{
	  os_uint8*  puc;
	  
	  puc = (os_uint8*) data;
	  while (length > 0) {
	    os_write_register_8((os_uint8*) devicemem, *puc);
	    puc++;
	    length--;
	  }
	  break;
	}

      case ADMXRC_IOWORD:
	{
	  os_uint16* pus;
	  
	  pus = (os_uint16*) data;
	  length >>= 1;
	  while (length > 0) {
	    os_write_register_16((os_uint16*) devicemem, *pus);
	    pus++;
	    length--;
	  }
	  break;
	}

      case ADMXRC_IOLONG:
	{
	  os_uint32* pul;
	  
	  pul = (os_uint32*) data;
	  length >>= 2;
	  while (length > 0) {
	    os_write_register_32((os_uint32*) devicemem, *pul);
	    pul++;
	    length--;
	  }
	  break;
	}

      default:
	os_assert(FALSE);
	break;
      }
    }
  }
  
  return OS_STATUS_SUCCESS;
}

OS_STATUS
ADMXRC_WriteReg(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_WRITEREG* iptr;
  os_uint8                index, value;
  os_uint8*               base;
  
  iptr = (ADMXRC_IOCTLS_WRITEREG*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  
  base = (os_uint8*) pDevCtx->hw.space1.untyped;
  index = iptr->in.index;
  if (index > pDevCtx->info.spaceInfo[1].LocalSize) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  value = iptr->in.value;
  os_write_register_8(base + index, value);
  
  return OS_STATUS_SUCCESS;
}
