/*
** ADM-XRC Series Driver
**
** (C) Copyright 2001-2004 Alpha Data
**
** setclock.c - IOCTL handlers for programming the clock generators.
**
** ----------------------------------------------------------------------------
*/

#include "platform.h"
#include "device.h"
#include "ioctl.h"
#include "setclock.h"

static OS_STATUS
doSetClock(
  DeviceContext* pDevCtx,
  os_uint32 freq,
  os_uint32 clockWord)
{
  OS_STATUS status = OS_STATUS_SUCCESS;
  os_boolean bSuccess;
  int i;
  unsigned long delay;
  MethodResult res;

  bSuccess = FALSE;
  for (i = 0; i < 10; i++) {
    res = pDevCtx->methods.clockProg(pDevCtx, freq, clockWord, TRUE);
    if (res != MethodResult_Success) {
      if (res == MethodResult_GeneralFailure) {
	continue;
      } else if (res == MethodResult_Timeout) {
	continue;
      } else {
	status = OS_MAKE_STATUS(ADMXRC_UNKNOWN_ERROR);
	goto done;
      }
    }
    delay = 10;
    if (os_delay_thread(TRUE, &delay)) {
      status = OS_MAKE_STATUS(ADMXRC_CANCELLED);
      goto done;
    }
    res = pDevCtx->methods.clockProg(pDevCtx, freq, clockWord, FALSE);
    if (res != MethodResult_Success) {
      if (res == MethodResult_GeneralFailure) {
	continue;
      } else if (res == MethodResult_Timeout) {
	continue;
      } else {
	status = OS_MAKE_STATUS(ADMXRC_UNKNOWN_ERROR);
	goto done;
      }
    }

    bSuccess = TRUE;
    break;
  }

  if (!bSuccess) {
    return OS_MAKE_STATUS(ADMXRC_FAILED);
  }

 done:
  return status;
}

/*
** ---------------------------------------------------------------------------
** Exported functions
** ---------------------------------------------------------------------------
*/

OS_STATUS
ADMXRC_SetClock(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* req)
{
  OS_SPINLOCK_FLAGS f;
  ADMXRC_IOCTLS_SETCLOCK* iptr;
  os_uint32 clockWord, reg;
  OS_STATUS status;

  iptr = (ADMXRC_IOCTLS_SETCLOCK*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  if (pDevCtx->info.cardInfo.BoardType != ADMXRC2_BOARD_ADMXRC &&
      pDevCtx->info.cardInfo.BoardType != ADMXRC2_BOARD_ADMXRC_P) {
    /* Allow this IOCTL to succeed only on ADM-XRC/ADM-XRC-P */
    return OS_MAKE_STATUS(ADMXRC_NOT_SUPPORTED);
  }
  
  clockWord = (os_uint32) iptr->in.clockword;
  if (!(clockWord & 0x800000U)) {
    /*
    ** Writing to one of the VCLKx or MCLK VCO registers -
    ** reject attempts to change any register other than
    ** MCLK or VCLK1
    */
    reg = (clockWord >> 21) & 0x7U;
    if (reg != ADMXRC_MCLK && reg != ADMXRC_VCLK1) {
      return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
    }
    
    if (pDevCtx->info.modifications & ADMXRC_MOD_CLOCKSWAP) {
      /*
      ** The VCLK1 and MCLK swap modification is present
      ** on this board, so we have to redirect writes to
      ** MCLK and VCLK1
      */
      switch (reg) {
      case ADMXRC_VCLK1:
	reg = ADMXRC_MCLK;
	break;
      case ADMXRC_MCLK:
	reg = ADMXRC_VCLK1;
	break;
      }
      clockWord = (clockWord & 0x1fffffU) | (reg << 21);
    }
  } else {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  os_acquire_spinlock(&pDevCtx->setclock.lock, &f);
  if (pDevCtx->setclock.busy) {
    pDevCtx->setclock.busy = TRUE;
    os_release_spinlock(&pDevCtx->setclock.lock, f);
    return OS_MAKE_STATUS(ADMXRC_DEVICE_BUSY);
  }
  os_release_spinlock(&pDevCtx->setclock.lock, f);

  status = doSetClock(pDevCtx, 0 /* luckily ignored for Plxbased */, clockWord);

  os_acquire_spinlock(&pDevCtx->setclock.lock, &f);
  pDevCtx->setclock.busy = FALSE;
  os_release_spinlock(&pDevCtx->setclock.lock, f);

  return status;
}

OS_STATUS
ADMXRC_SetClock2(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* req)
{
  OS_SPINLOCK_FLAGS f;
  ADMXRC_IOCTLS_SETCLOCK2* iptr;
  ADMXRC_IOCTLS_SETCLOCK2* optr;
  OS_STATUS status;
  int index;
  os_uint32 fref, freq, actual;
  os_uint32 clockWord;
  MethodResult res;

  iptr = (ADMXRC_IOCTLS_SETCLOCK2*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  optr = (ADMXRC_IOCTLS_SETCLOCK2*) req->param.outBuf;
  if (optr == NULL || req->param.outSize != sizeof(optr->out)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  index = iptr->in.clock_index;
  freq = iptr->in.freq;

  if (index >= pDevCtx->info.cardInfo.NumClock) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  switch (pDevCtx->info.cardInfo.BoardType) {
  case ADMXRC2_BOARD_ADMXRC_P:
  case ADMXRC2_BOARD_ADMXRC:
    {
      if (!(pDevCtx->info.modifications & ADMXRC_MOD_CLOCKSWAP)) {
	/*
	** The VCLK1 and MCLK swap modification is not present
	** on this board.
	*/
	index = index ^ 1;
      }

      switch (pDevCtx->info.clockType) {
      case ADMXRC_CLOCKTYPE_16:
	fref = 16666667;
	break;
      case ADMXRC_CLOCKTYPE_14:
	fref = 14318181;
	break;
      default:
	os_assert(FALSE);
	return OS_MAKE_STATUS(ADMXRC_UNKNOWN_ERROR);
      }

      break;
    }

  case ADMXRC2_BOARD_ADMXRC2:
  case ADMXRC2_BOARD_ADPWRC2:
  case ADMXRC2_BOARD_ADPDRC2:
  case ADMXRC2_BOARD_ADMXRC2_LITE:
    {
      switch (pDevCtx->info.clockType) {
      case ADMXRC_CLOCKTYPE_16:
	fref = 16666667;
	break;
      case ADMXRC_CLOCKTYPE_14:
	fref = 14318181;
	break;
      default:
	os_assert(FALSE);
	return OS_MAKE_STATUS(ADMXRC_UNKNOWN_ERROR);
      }

      break;

    case ADMXRC2_BOARD_ADMXRC2_PRO_LITE:
    case ADMXRC2_BOARD_ADMXRC2_PRO:
      {
	fref = 14318181;
	break;
      }
    }
  default:
    os_assert(FALSE);
    return OS_MAKE_STATUS(ADMXRC_UNKNOWN_ERROR);
  }

  res = pDevCtx->methods.clockWord(pDevCtx, index, fref, freq, &actual, &clockWord);
  if (res != MethodResult_Success) {
    if (res == MethodResult_InvalidParameter) {
      return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
    } else {
      return OS_MAKE_STATUS(ADMXRC_UNKNOWN_ERROR);
    }
  }

  os_acquire_spinlock(&pDevCtx->setclock.lock, &f);
  if (pDevCtx->setclock.busy) {
    os_release_spinlock(&pDevCtx->setclock.lock, f);
    return OS_MAKE_STATUS(ADMXRC_DEVICE_BUSY);
  } else {
    pDevCtx->setclock.busy = TRUE;
    os_release_spinlock(&pDevCtx->setclock.lock, f);
  }

  status = doSetClock(pDevCtx, actual, clockWord);

  os_acquire_spinlock(&pDevCtx->setclock.lock, &f);
  pDevCtx->setclock.busy = FALSE;
  os_release_spinlock(&pDevCtx->setclock.lock, f);

  if (OS_SUCCESS(status)) {
    optr->out.actual = actual;
    req->status.length = sizeof(optr->out);
  }

  return status;
}
