/*
** linux.c
**
** Source code for Linux Specific functions for API Library
**
** (C) Copyright 2002-2004 Alpha Data Parallel Systems Ltd.
*/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <malloc.h>

#include "platform2.h"
#include <admxrc2.h>

static ADMXRC2_STATUS
map_error(
  int err)
{
  if (err >= ADMXRC2_STATUS_ERROR_CODE) {
    /* Is an ADMXRC2_STATUS value */
    return err;
  }

  switch (err) {
  case EBADF:
    return ADMXRC2_INVALID_HANDLE;
  case EINVAL:
    return ADMXRC2_INVALID_PARAMETER;
  case ENOTTY:
    return ADMXRC2_BAD_DRIVER;
  case EINTR:
    return ADMXRC2_FAILED;
  default:
    fprintf(stderr, "map_error: unknown ioctl result %d\n", err);
    return ADMXRC2_UNKNOWN_ERROR;
  }
}

os_boolean
_ADMXRC2_IoCtlNoWait(
  ADMXRC2_HANDLE  hDevice,        /* device handle (file pointer)             */
  unsigned long   IoControlCode,  /* IOCTL code                               */
  void*	          lpInBuffer,     /* pointer to buffer to supply input data   */
  unsigned long	  nInBufferSize,  /* size of input buffer                     */
  void*           lpOutBuffer,    /* pointer to buffer to receive output data */
  unsigned long	  nOutBufferSize, /* size of output buffer                    */
  ADMXRC2_STATUS* status)
{
  ADMXRC_IOCTL_PARAMS ibp;
  ADMXRC2_STATUS      st;
  int                 err;

  ibp.in.lpBuffer = lpInBuffer;
  ibp.in.nSize = nInBufferSize;
  ibp.out.lpBuffer = lpOutBuffer;
  ibp.out.nSize = nOutBufferSize;
  ibp.mass.nSize = 0;

  err = ioctl(hDevice, IoControlCode, &ibp);
  if (err) {
    if (err < 0) {
      st = map_error(errno);
    } else {
      st = (ADMXRC2_STATUS) err;
    }
  } else {
    st = ADMXRC_SUCCESS;
  }

  *status = st;

  return (st == ADMXRC2_SUCCESS);
}

os_boolean
_ADMXRC2_IoCtlNoWaitMass(
  ADMXRC2_HANDLE  hDevice,        /* device handle (file pointer)             */
  unsigned long   IoControlCode,  /* IOCTL code                               */
  void*	          lpInBuffer,     /* pointer to buffer to supply input data   */
  unsigned long	  nInBufferSize,  /* size of input buffer                     */
  void*           lpOutBuffer,    /* pointer to buffer to receive output data */
  unsigned long	  nOutBufferSize, /* size of output buffer                    */
  void*           lpMassBuffer,    /* pointer to buffer to receive output data */
  unsigned long	  nMassBufferSize, /* size of output buffer                    */
  int             nMassMode,
  ADMXRC2_STATUS* status)
{
  ADMXRC_IOCTL_PARAMS ibp;
  ADMXRC2_STATUS      st;
  int                 err;

  ibp.in.lpBuffer = lpInBuffer;
  ibp.in.nSize = nInBufferSize;
  ibp.out.lpBuffer = lpOutBuffer;
  ibp.out.nSize = nOutBufferSize;
  ibp.mass.lpBuffer = lpMassBuffer;
  ibp.mass.nSize = nMassBufferSize;

  err = ioctl(hDevice, IoControlCode, &ibp);
  if (err) {
    if (err < 0) {
      st = map_error(errno);
    } else {
      st = (ADMXRC2_STATUS) err;
    }
  } else {
    st = ADMXRC_SUCCESS;
  }

  *status = st;

  return (st == ADMXRC2_SUCCESS);
}

void _ADMXRC2_strlwr(char *a) {
  char c;
  
  while ((c = *a) != '\0') {
    *a++ = tolower(c);
  }
}

/* Allocate and free memory to page boundaries */
void*
_ADMXRC2_PageMalloc(
  unsigned long nbytes)
{
  return memalign(PAGE_SIZE, nbytes);
}

void
_ADMXRC2_PageFree(
  void* p)
{
  free(p);
}         

ADMXRC2_STATUS
_ADMXRC2_OpenByIndex(
  int             index,
  ADMXRC2_HANDLE* handle)
{
  ADMXRC2_HANDLE driver;
  char           driverName[] = "/dev/admxrciixxxx";
  ADMXRC2_STATUS status;

  sprintf(driverName, "/dev/admxrcii%d", index);
  driver = open(driverName, O_RDWR);
  if (driver == -1) {
    status = ADMXRC2_CARD_NOT_FOUND;
  } else {
    status = ADMXRC2_SUCCESS;
    *handle = driver;
  }

  return status;
}

ADMXRC2_STATUS
_ADMXRC2_Close(
  ADMXRC2_HANDLE card)
{
  int                res;
  unsigned int       i;
  ADMXRC2_SPACE_INFO spInfo;
  ADMXRC2_STATUS     status;

  for (i = 0;; i++) {
    status = ADMXRC2_GetSpaceInfo(card, i, &spInfo);
    if (status == ADMXRC2_INVALID_HANDLE) {
      goto done;
    } else if (status == ADMXRC2_SUCCESS) {
      if (spInfo.VirtualBase != NULL) {
	munmap(spInfo.VirtualBase, spInfo.VirtualSize);
      }
    } else {
      break;
    }
  }
  status = ADMXRC2_SUCCESS;
  res = close(card);

done:
  return status;
}

ADMXRC2_STATUS
_ADMXRC2_MapToUserSpace(
  ADMXRC2_HANDLE card)
{
  ADMXRC2_SPACE_INFO spInfo;
  ADMXRC2_STATUS     status;
  unsigned int       space;
  int                res;

  for (space = 0;; space++) {
    status = ADMXRC2_GetSpaceInfo(card, space, &spInfo);
    if (status == ADMXRC2_INVALID_HANDLE) {
      goto done;
    } else if (status == ADMXRC2_SUCCESS) {
      if (spInfo.LocalSize) {
	res = (int) mmap(0, spInfo.LocalSize, PROT_READ | PROT_WRITE, MAP_SHARED, card, spInfo.PhysicalBase);
	if (res == -1) {
	  fprintf(stderr, "Device memory could not be mapped to the user process\n");
	  fprintf(stderr,
		  "Parameters %lx, %lx, %lx, %lx, %lx, %lx\n",
		  (unsigned long) 0,
		  (unsigned long) spInfo.LocalSize,
		  (unsigned long) (PROT_READ | PROT_WRITE),
		  (unsigned long) MAP_SHARED,
		  (unsigned long) card,
		  (unsigned long) spInfo.PhysicalBase);
	  status = ADMXRC2_NO_MEMORY;
	  goto done;
	}
      }
    } else {
      break;
    }
  }
  status = ADMXRC2_SUCCESS;

done:
  return status;
}

void
_ADMXRC2_Enumerate(
  _ADMXRC2_enum_fn* fn,
  void*             context)
{
  ADMXRC2_HANDLE card;
  ADMXRC2_STATUS status;
  int            i;

  for (i = 0; i < 16; i++) {
    status = _ADMXRC2_OpenByIndex(i, &card);
    if (status == ADMXRC2_SUCCESS) {
      if (fn(i, card, context) < 0) {
	break;
      }
    }
  }
}

/*
** --------------------------------------------------------------------
** Exported functions start here
** --------------------------------------------------------------------
*/
